//  CHOPPER.H 
//  A general parsing class, based on the famous "chop" function 
//    used in G and Interdyme, but by concentrating the functionality
//    of chop into a Chopper object, each object can have its own
//    set of global variables, such as feedchar, feedbuf, etc. 
//    Each Chopper object can be chopping from a different source,
//    and store the knowledge about where they are in their internal
//    variables.


/* Explanation of members:
   feedchar - Next character read using feed().
   feedbuf - 1 character buffer to hold "unfeeded" character.
   feedbufflag - true if buffer is full, false if empty
   pinbuf - Pointer to input buffer.
   inbuf - Input buffer.
   paddargs - Pointers to addfile arguments.
	ichopbuf - Flag to signal if anything is "unchopped.
   chopbuf - Buffer for storing strings that have been "unchopped",
      i.e., sent back to chop for storage.  If something is on this
      buffer, it will be popped of on the next chop.  Note that this
      concept could be generalized by creating an entire stack of
      unchopped parse strings.
   chopretrn - Next return value of chop.
   linebuf - Multiple command line buffer.
   linebpos - Position of lines in the buffer.
   linebfil - Number of lines stored in buffer
   ChopSource - Current file source of Chopper.
   console - 'y' if reading from the screen, 'n' if reading from a file.
   
*/

#if !defined(CHOPPER_H)
#define CHOPPER_H  1
#include <conio.h>

#include "message.h"

#if !defined(ERR)
#define ERR (-1)
#endif
#if !defined(TRUE)
#define TRUE 1
#endif
#if !defined(FALSE)
#define FALSE 0
#endif
#if !defined(YES)
#define YES 1
#endif

class DataBank;
class GDate;

class Chopper {
  protected:
	// enumeration constants:
	//  LLINE is the maximum length of line to read.
   //  LINEBLEN is the maximum length of the memory buffer of commands.
   //  NLINEBUF is the maximum number of lines stored.
   enum Params {LLINE=1200, LINEBLEN=600, 
					NLINEBUF=30, CHOPBUFLEN=512,
               STRINGLEN=90};
   short feedchar,feedbufflag;
	char ichopbuf,feedbuf;
	char inbuf[LLINE];
	char chopbuf[CHOPBUFLEN], chopretrn;
	char linebuf[LINEBLEN];
	char FileOrString;    // 'f' if from file, 's' if from string
	short linebpos[NLINEBUF];
	short linebfil;
	FILE *ChopSource;
	GMessage* MsgObject;
	struct text_info ta;  // to save text information

   // Databanking variables, related to rhs() and find();
   DataBank *JustFound;
   short npy;
  public:
	enum constants {OK=0,
ESC=27,
STOP=1,
GO=0,
BS=8,
CTRL_C=3};
	// Various options derived from G. (must be set in constructor):
	FILE *fpCatch, *fpsave, *fpannsave;
	char *pinbuf,**paddargs;
	char console;
	char insert, saving, annsaving, catching;
	char save_flag, annsave_flag;
   char SaveDataCommand[12], save_data_flag;
	short addprint;

   // constructors:
   Chopper();
   Chopper(FILE *iin);
   // destructor
   ~Chopper();

   // Function methods:
	short SetSource(FILE *iin);
	short SetSource(char* buf);
	short FromFile(void) {return (FileOrString=='f')?1:0;};
	FILE *GetSource(void) {return (FileOrString=='f')?ChopSource:0;};
	short close(void) {if(ChopSource!=NULL) fclose(ChopSource); 
					ChopSource=NULL; return 0;};
	short SetCatch(FILE *fp) {catching='y';fpCatch=fp; return 0;};
	short SetSave(FILE *fp)  {saving='y'; fpsave=fp; return 0;};
	short SetAnnSave(FILE *fp) {annsaving='y';fpannsave=fp;return 0;};
	short chop(char *s);
	short feed(void);     // feed from current stream
	short unchop(void) {ichopbuf=1; return(1); };
	short unfeed(char c) {feedbuf=c; feedbufflag=1; return(1); };
	short endline(void);
	short geteos(void);
	short geteol(void);
	double getd(void);
	short geti(void);
	short getfl(float &x);

	short initcommandbuf(void) {linebfil=-1;linebpos[0]=0;return 0;};
	short wrinbuf(short n, char which);
   short getopt(char *, short length=0);
	short getstring(char *);
   short fills(short members[], short& shift);
   short fills(short members[]);
//  Functions related to the fabulous rhs():

   short find(char *nam, char *pend, float *ss, short quiet=0);
   short Realign(float *f, GDate& bd, GDate& BaseDate);
   short check_lag(float *ss, char* pend);


	short rhs(float *,char *, short quiet=0);
	short rhssyn(short lt);
	short muldiv(float *a, char *op_stack, short *posp, float *b, short *pnopya, short nopyb);
	short addsub(float *f, char *op_stack, short *posp, float *a, short *pnopyf, short nopya);
	short functions(float *f, char *pend);
	short accum(float *f, char *pend);
	short csum(float *f, char *pend);
	short fill(short members[]);
	short peak(float *f, char *pend);
	short round(float *f, char *pend);
	short expo(float *f, char *pend);
	short sine(float *f, char *pend);
	short ifpos(float *f, char *pend);
	short abso(float *f, char *pend);
	short logarithm(float *f, char *pend);
	short square(float *f, char *pend);
	short square_root(float *f, char *pend);
	short zero(float *f, char *pend);
	short miss(float *f, char *pend);
	short positive(float *f, char *pend);
	short power(float *f, char *pend);
   short percent2level(float *f, char *pend);
   short pctcmd(float *f, char *pend);
   short unirand(float *f); 


};
#endif
