// DATABANK.H
// All other data bank classes will inherit from DATABANK, so that base
//   class pointers can be used to point to any databank type.

#if !defined(DATABANK_H)
#define DATABANK_H 1

#include "message.h"

// Constants used in databank classes:
const float MISSING=-.000001;

class DataBank {   // Abstract data bank class.

protected:
   char *classname;
   char *bnkname, *indname, *bnktitle;
   char banktype;
   short opened;
   FILE *pbank, *ipbank;
   unsigned short jahra,nseries,nopy,nobs,stper;

   GMessage* MsgObject;  // Handles messages that used to be external.
	// (Must be initialized in constructor!)
public:
   enum {BASE_YR=1925, 
   NOMAX=999};

   DataBank(); // default
	DataBank(const char *);
   DataBank(DataBank& db);
   // This is an abstract base class, and therefore *must* have a
   //              virtual destructor, or we get into serious trouble!
   //    (Item 14 - Meyers Effective C++)  Note that we also 
   //    have to provide a *definition* of the pure virtual 
   //    destructor.
   virtual ~DataBank(){}
   unsigned short Getnseries(void) {return nseries;}
   short Setnseries(short n) {return nseries=n;};
   unsigned short Getnopy(void)   {return nopy;}
   short Setnopy(unsigned short npy) {return(nopy = npy);}
   unsigned short Getnobs(void)   {return nobs;}
   virtual unsigned short GetBaseYr(void) {return jahra;}
   short SetBaseYr(unsigned short yr)  {jahra=yr;return 0;}
   short Setstper(short n) {stper=n;return 0;};
   short Getstper(void)  {return stper;}

   virtual short getser(char *nam,float *f,short SeriesLength=-1)=0; 
   virtual short ListNames(FILE*,char*,short wide=1)=0;
   virtual short Announce(short bankno=0);

   short IsOpen(void);
   short IsErr(void);
   char *Getclassname(void) {return classname;}
   char *Getbnkname(void)  {return bnkname;}
   char *Getindname(void) {return indname;}
   char *Getbnktitle(void) {return bnktitle;}
   virtual char *Setbnktitle(const char *s);
   char Getbanktype(void) {return banktype;}
};

#endif  // end DATABANK_H definitions

