/* DYMEUTIL -- The functions in this file are pure utility functions,
    not associated with any particular class, and not Interdyme specific.
    Some of these same functions can be found in the MISCTOOLS.CPP file 
    in the \CLASSES directory.
*/

#include "dymesys.h"   // Interdyme system files.
const float SMALL = 1.0e-7;
/* Included in this file:
    AggVec
    arith
    breakname ( 2 versions)                   
    depad
    getclopt
    getint
    getfloat
    getname
    getopt
    interp
    isnum
    _matherr (struct exception *a);
    max -> should be made into template
    min
    matchwild
    MemoryLeft
    outofmem
    safelog
    safepow
    sgetint
    sgets
    sgetfl
    splitvec
   
    TackExt
    tap
    dymetap
    ti

    vector
    matrix
    free_vector
    free_matrix
    ivector
    free_ivector
    lvector
    free_lvector
    imatrix
    free_imatrix

*/
// DymeTAP  -- Tap a key to continue, and if exiting, ask about saving
//   data.
void dymetap()
{
   int c;
    cerr << "\nTap a key to continue.  ESC to exit.";
    if(getch() == 27) {
      cerr << "\nWould you like to save vector and macro data?";
      c = getch();
      if(c=='y') {
         store(t);
         #if defined(INCL_MACRO)
         storets();
         #endif
         }
        exit(1);
      }
    }
//========================================================================
long MemoryLeft(void) {
#ifdef DJGPP
__dpmi_free_mem_info info;
__dpmi_get_free_memory_information(&info);
return ( (long) info.largest_available_free_block_in_bytes);
#elif !defined INCL_CORELEFT
return 0; // no info available?
#else
return farcoreleft();
#endif
}

#ifdef INCL_IVECTOR
// ====================================================================
short AggVec(Vector& a, Vector& b, IVector& iag) {
short nra,nrb,nri,i,loc;
nra=a.rows();
nri=iag.rows();
if(nra!=nri) {
    printf("Error in AggVec, aggregation key must be same length as source!\n\r");
    return -1;
    }
nrb=b.rows();
b.set(0);
for(i=1;i<=nra;i++) {
    loc=iag[i];
    if(loc<0 || loc>nrb) {
        printf("In AggVec, element of aggregation vector is too large!\n\r");
        return -1;
        }
    if(loc==0)
        continue;

    b[loc] += a[i];
    }
return 0;   
}
//========================================================================
// This function splits the Vector a to the Vector b, where iag would
//   be the key to aggregate b to a.  c is the vector used where splits
//   are necessary.
const short MAXSPLIT=50;
short SplitVec(Vector& a, Vector& b, Vector& c, IVector& iag) {
short na,nb,nc,ni,i,j,loc;
short list[MAXSPLIT], nlist;
float ratio, total;
na=a.numelm();
nb=b.numelm();
nc=c.numelm();
ni=iag.numelm();
if(nb!=nc) {
    printf("In SplitVec the size of the splitter vector is not correct.\r\n");
    printf("  It is %d, but should be %d.\r\n",nc,na);
    return -1;
    }
if(nb!=ni) {
    printf("In SplitVec the size of the aggregation vector is not correct.\r\n");
    printf("  It is %d, but should be %d.\r\n",ni,na);
    return -1;
    }
b.set(0.);
for(i=1;i<=na;i++) {
    nlist=0;
    // Get list of more detailed sectors that corresponds to this sector.
    for(j=1;j<=nb;j++) {
        if(iag[j]==i) {
            list[nlist] = j;
            nlist++;
            if(nlist>MAXSPLIT) {
                printf("Too many splits.  Maximum of %d allowed.\r\n",
                    MAXSPLIT);
                }
            }
        }
    if(nlist==0)
        continue;
    else if(nlist==1) {
        loc=list[0];
        b[loc]=a[i];
        }
    else {
        total=0.;
        for(j=0;j<nlist;j++) {
            loc=list[j];
            total += c[loc];
            }
        if(fabs(total)<1.0e-7)
            continue;
        for(j=0;j<nlist;j++) {
            loc=list[j];
            ratio=c[loc]/total;
            b[loc] = ratio * a[i];
            }
        }
    }
return 0;
}
#endif
// ====================================================================
/* arith()
    check for arithmetic error.
      pgm can provide some tracer text
     where can be used for a sector number, index, etc.
*/
static char pgmlast[120];
static short wherelast;

void arith(char *pgm, short where)
{
    #ifdef CONTROL87
    unsigned status;
    status = _status87();
    if(status % 32 != 0){
        if( (status & SW_INVALID) == SW_INVALID) {
            printf("<-> Invalid operation at %s (%d).\n",pgm,where);
            _clear87();
            strcpy(pgmlast,pgm);  wherelast=where;
            return;
            }
        else if( (status & SW_DENORMAL) == SW_DENORMAL) {
            #ifndef __BCB__
            printf("<-> Denormalized operand at (or before) %s (%d).\n",
                pgm,where);
            #else
            _clear87();
            strcpy(pgmlast,pgm);  wherelast=where;
            return;
            #endif
            }
        else if( (status & SW_INEXACT) == SW_INEXACT) {
            //#ifndef __BCB__
            //printf("\n  -> Loss of precision in result.\n");
            //#else
            _clear87();
            strcpy(pgmlast,pgm);  wherelast=where;
            return;
            //#endif
            }
        else if( (status & SW_ZERODIVIDE) == SW_ZERODIVIDE)
            printf("<-> Divide by zero at (or before) %s (%d).\n",pgm,where);
        else if( (status & SW_OVERFLOW) == SW_OVERFLOW)
            printf("<-> Floating point overflow at (or before) %s (%d).\n",
                pgm,where);
        else if( (status & SW_UNDERFLOW) == SW_UNDERFLOW)
            printf("<-> Floating point underflow at or before %s (%d).\n",
                pgm,where);
        printf("\nAn arithmetic exception has occured %s",pgm);
        if(where >= 0) printf(" at %d.\n",where);
        else printf(".\n");
        printf("Do not trust answers.\n");
        _clear87();
        strcpy(pgmlast,pgm);  wherelast=where;
        tap();  // exits if ESC is tapped.
        }
    #endif
    }
#if !defined __BCB__ && !defined __LINUX__
/* =================================================================== */
/* _matherr - checks for bad argument passed to math function          */
int _matherr (struct exception *a) {
char type[50];
if (a->type == DOMAIN)
    strcpy(type,"domain");
else if(a->type == SING)
    strcpy(type,"singularity");
else if(a->type == OVERFLOW)
    strcpy(type,"overflow");
else if(a->type == UNDERFLOW)
    strcpy(type,"underflow");
else if(a->type == TLOSS)
    strcpy(type,"total loss of significant digits");
printf("%s error with %s() function.\r\n",type,a->name);
dymetap();
a->retval = 0;
return 1;
}
#endif
/* =================================================================== */
// BREAKNAME is overloaded: a simple version and a more complicated.
/*** breakname divides a vector-sector name into a vector name and a sector.
      (The sector is the return value.  */

short breakname(char *nam, char *vnam) {
short len,i,j, tag, sec;
char s[10];

len = strlen(nam);
i = len-1;
while(isdigit(nam[i])) i--;
tag = i;
i++;
j = 0;
while((s[j++]=nam[i++]) != '\0');
if((len = strlen(s)) == 0)
    sec = 1;
else sec = atoi(s);
i = 0;
while(i <= tag){
    vnam[i] = nam[i];
    i++;
    }
vnam[i] = '\0';
return(sec);
}
/*=========================================================================*/
/* breakname() devides a vector-sector name into a vector name and a sector 
    For a name of the form pce27, "pce" will be put into vnam and the return
    value will be 27.  For a name of the form am3.7, (am row 3, col 7),
    vname will be returned as "am", *prow as 3, and the return value is 7.
*/ 
short breakname(char *nam, char *vnam, short &prow) {

short len,i,j, tag, sec;
char s[10];

len = strlen(nam);
i = len-1;
while(isdigit(nam[i])) i--;
tag = i;
i++;
j = 0;
while((s[j++]=nam[i++]) != '\0');
if((len = strlen(s)) == 0) sec = 0; // Was sec = 1, changed 23 Nov 92 for indexcmd
else sec = atoi(s);

if(nam[tag] == '.'){
    i = tag-1;
    while(isdigit(nam[i])) i--;
    tag = i;
    i++;
    j = 0;
    while((s[j++] = nam[i++]) != '.');
    s[j-1] = '\0';
    if((len = strlen(s)) == 0) prow = -1;
    else prow = atoi(s);
    }

i = 0;
while(i <= tag){
    vnam[i] = nam[i];
    i++;
    }
vnam[i] = '\0';
return(sec);
}
/*======================================================================*/
/* depad() -- remove trailing blanks from a string */
void depad(char *s) {
short i,len;

len = strlen(s);
i = len-1;
while(s[i] == ' ' && i >= 0)
    i--;
if(i >= 0)
    s[i+1] = '\0';
else s[0] = '\0';
}
//=====================================================================
/* getclopt returns in the second argument the string found on the
     command line following first argument preceeded by a -.  ***/

short getclopt(char opt, char *s, short argc, char **argvec)
{
    short i,option;

    if(argc <= 1) return(ERR);
    option = ERR;
    for(i = 1;i < argc; i++){
        if(*argvec[i] == '-' && *(argvec[i]+1) == opt){
            strcpy(s,argvec[i]+2);
            option = opt;
            break;
            }
        }
    return(option);
    }
//=====================================================================
/* GETINT
    Get an integer, n, from the character array s.  Start looking in
    position j of s.  If none is found, read another line from the
    file fp and search it from position 0.  The value of j returned is
    the value of the next position in s after the integer was found.  The s
    argument should be initialized with a call to fgets() before first call
    of getint.  Note that j is changed by the call, so do NOT write the call
    as getint(...,0)!  If the file pointer, fp, is 0, do not attempt
    to read another line.
*/
int getint(int& f, FILE *fp, char *s, short& j)
{
    short k = 0;
    char field[30], sj = s[j];
    short eflag = 0;

    // skip over non-numerical characters
    while( isdigit(s[j]) == 0 && s[j] != '\n'){
        sj = s[j];
        if(s[j] == '\0') break;
        if(s[j] == '+' || s[j] == '-'){         // " +1 " or " -1 "
            if(isdigit(s[j+1]) != 0){
                field[0] = s[j];
                k=1;
                break;
                }
            }
            if(s[j] == 'e'){
                if(j>0 && (s[j-1] != ' ' && s[j-1] != '\t')){
                    j++;        // ignore e in " the123"
                    continue;
                    }
                if(s[j+1] == '+' || s[j+1] == '-'){
                    if(isdigit(s[j+2])!= 0){        // " e+1 "
                    field[0] = s[j];
                    field[1] = s[j+1];
                    k=2;
                    eflag = 1;
                    break;
                    }
                }
            if(isdigit(s[j+1]) != 0){  // " e1 "
                field[0] = s[j];
                printf("Warning: values in scientific notation should be signed.\n");
                field[1] = '+';         // atof seems to require a sign
                k=2;
                j--;
                eflag = 1;
                break;
                }
            }
        j++;
        }
    if(sj == '\n' || sj == '\0'){
        if(fp == 0) return(ERR);
        if(fgets(s,240,fp) == NULL){
            return(ERR); // Hit end of file.
            }
        j = 0;
        if(getint(f, fp, s, j)) return(1);
        else return(ERR);
        }

    while(1){
        if(isdigit(s[j+k]) != 0){
            field[k] = s[j+k];
            k++;
            continue;
            }
        if(s[j+k] == 'e'){
            if(eflag > 0){
                printf("\nWarning: found 2 e's in %s\n",s);
                printf("Skipping to next integer\n");
                j += k;           // for  " e2e3 ", will return "3"
                if(getint(f, fp, s, j)) return(1);
                else return(ERR);
                }
            eflag = 1;
            if(s[j+k+1] == '+' || s[j+k+1] == '-'){
                if(isdigit(s[j+k+2])!= 0){        // " e+1 "
                    field[k] = s[j+k];
                    field[k+1] = s[j+k+1];
                    k += 2;
                    continue;
                    }
                }
            if(isdigit(s[j+k+1]) != 0){                 // " e1 "
                field[k] = s[j+k];
                printf("Warning: values in scientific notation should be signed.\n");
                field[k+1] = '+';        // atof seems to require a sign
                k += 2;
                j--;
                continue;
                }
            break;
            }
        break;
        }
    field[k] = '\0';
    if(field[0] == 'e'){       // allows proper interpretation of " e2"
        for(int m=k;m>=0; m--){ // since atof seems to require " 1e2"
            field[m+1] = field[m];
            }
        field[0] = '1';
        }
    j += k;
    if(eflag > 0){             // atoi does not allow scientific notation;
        f = (int)atof(field);   // this is a way around that
        return(1);
        }
    f = atoi(field);
    return(1);
}
/*
{
    short k;
    char field[30];

    // skip over non-numerical characters
    search: while(isdigit(s[j])==0 && s[j]!='-' && s[j]!='+' && s[j] != '\n') j++;
    if(s[j] == '\n'){
        if(fgets(s,240,fp) == NULL){
            return(ERR); // Hit end of file.
            }
        j = 0;
        goto search;
        }
    k = 0;
    while(isdigit(s[j+k]) != 0 ){
        field[k] = s[j+k];
        k++;
        }
    field[k] = '\0';
    j += k;
    f = atoi(field);
    return(OK);
    }
*/
//=====================================================================
/*  GETFLOAT
    Get a float, f, from the character array s.  Start looking in
    position j of s.  If none is found, read another line from the
    file fp and search it from position 0.  The value of j returned is
    the value of the next position in s after the float was found.  The s
    argument should be initialized with a call to fgets() before first call
    of getfloat.  Note that j is changed by the call, so do NOT write the call
    as getint(...,0)!  If the file pointer, fp, is 0, do not attempt
    to read another line.

    */
short getfloat(float& f, FILE *fp, char *s, short& j)
{
    short k = 0;
    char field[30], sj = s[j];
    short decflag = 0, eflag = 0;

    // skip over non-numerical characters
    while( isdigit(s[j]) == 0 && s[j] != '\n'){
        sj = s[j];
        if(s[j] == '\0') break;
        if(s[j] == '.' && isdigit(s[j+1]) != 0){  // " .1"
            field[0] = s[j];
            k=1;
            decflag = 1;
            break;
            }
        if(s[j] == '+' || s[j] == '-'){         // " +1 " or " -1 "
            if(isdigit(s[j+1]) != 0){
                field[0] = s[j];
                k=1;
                break;
                }
            if(s[j+1] == '.' && isdigit(s[j+2]) != 0){  // " +.1 "
                field[0] = s[j];
                field[1] = s[j+1];
                k = 2;
                decflag = 1;
                break;
                }
            }
        if(s[j] == 'e'){
            if(j>0 && (s[j-1] != ' ' && s[j-1] != '\t')){
                j++;        // ignore e in " the123"
                continue;
                }
            if(s[j+1] == '+' || s[j+1] == '-'){
                if(isdigit(s[j+2])!= 0){        // " e+1 "
                    field[0] = s[j];
                    field[1] = s[j+1];
                    k=2;
                    eflag = 1;
                    break;
                    }
                }
            if(isdigit(s[j+1]) != 0){                 // " e1 "
                field[0] = s[j];
                printf("Warning: values in scientific notation should be signed.\n");
                field[1] = '+';     // atof seems to require a sign
                k=2;
                j--;
                eflag = 1;
                break;
                }
            }
        j++;
        }

    if(sj == '\n' || sj == '\0'){
        if(fp == 0) return(ERR);
        if(fgets(s,240,fp) == NULL){
            return(ERR); // Hit end of file.
            }
        j = 0;
        if(getfloat(f, fp, s, j)) return(1);
        else return(ERR);
        }

    while(1){
        if(isdigit(s[j+k]) != 0){
            field[k] = s[j+k];
            k++;
            continue;
            }
        if(s[j+k] == '.'){
            if(decflag > 0){
                printf("\nWarning: found 2 decimal points in %s\n",s);
                printf("Skipping to next float\n");
                j += k+1;      // for " .2.3 " will return "3"
                if(getfloat(f, fp, s, j)) return(1);
                else return(ERR);
                }
            else{
                field[k] = s[j+k];
                decflag = 1;
                k++;
                continue;
                }
            }
        if(s[j+k] == 'e'){
            if(eflag > 0){
                printf("\nWarning: found 2 e's in %s\n",s);
                printf("Skipping to next float\n");
                j += k;           // for  " e2e3 ", will return "3"
                if(getfloat(f, fp, s, j)) return(1);
                else return(ERR);
                }
            eflag = 1;
            if(s[j+k+1] == '+' || s[j+k+1] == '-'){
                if(isdigit(s[j+k+2])!= 0){        // " e+1 "
                    field[k] = s[j+k];
                    field[k+1] = s[j+k+1];
                    k += 2;
                    continue;
                    }
                }
            if(isdigit(s[j+k+1]) != 0){                 // " e1 "
                field[k] = s[j+k];
                printf("Warning: values in scientific notation should be signed.\n");
                field[k+1] = '+';        // atof seems to require a sign
                k += 2;
                j--;
                continue;
                }
            break;
            }
        break;
        }
    field[k] = '\0';
    if(field[0] == 'e'){            // allows proper interpretation of " e2"
        for(int m=k;m>=0; m--){ // since atof seems to require " 1e2"
            field[m+1] = field[m];
            }
        field[0] = '1';
        }
    j += k;
    f = atof(field);
    return(1);
}


//=====================================================================
/* GETNAME
    Get a name, field, from the character array s.  Start looking in
    position j of s.  If none is found, read another line from the
    file fp and search it from position 0.  The value of j returned is
    the value of the next position in s after the name was found.  The s
    argument should be initialized with a call to fgets() before first call
    of getint.  Note that j is changed by the call, so do NOT write the call
    as getint(...,0)!  If the file pointer, fp, is 0, do not attempt
    to read another line.

    */
short getname(char *field, FILE *fp, char *s, short& j)
{
    short k;

    // skip over non-numerical characters
    search: while(s[j] == ' ' && s[j] != '\n') j++;
    if(s[j] == '\n'){
        if(fp == 0) return(ERR);
        if(fgets(s,240,fp) == NULL){
            return(ERR); // Hit end of file.
            }
        j = 0;
        goto search;
        }
    k = 0;
    while(s[j+k] != ' ' && s[j+k] != '\n' ){
        field[k] = s[j+k];
        k++;
        }
    field[k] = '\0';
    j += k;
    return(OK);
    }
//=====================================================================
/*  getopt() -- used by configure() to put into "line" what follows the ';'
    in a line of the dyme.cfg file. */ 
short getopt(FILE *fp, char* line, short maxchar)
{
    short i,n;
    char buf[240];

    if(fgets(buf,240,fp) == 0) return(ERR);
    i = 0; n = 240;
    for(i = 0; i < n; i++){
        if(buf[i] == ';') goto found;
        if(buf[i] == '\0') break;
        }
    printf("Config line missing ;: %s\n",buf);
    tap();
    return(ERR);
    found:
    i++;
    while(buf[i] == ' ') i++;
    n = i;
    for(i = 0;i <= maxchar;i++){
        line[i] = buf[i+n];
        if(line[i] == '\n') goto normal;
        }
    printf("Config line too long: %s\n",buf);
    line[maxchar] = '\0';
    return(ERR);
    normal:
    /* pad for cxl */
    n = i;
    for(i = n;i < maxchar;i++)
        line[i] = ' ';
    line[maxchar] = '\0';
    return(OK);
    }
//=====================================================================
/* interp() -- linear interpolation
    n is the number of points
    dates, the dates of the interpolation points
    values, the values at those points.
    */
float interp(short n, short *dates, float *values)
{
    short i;
    float gap;
    for (i = 0; i < n; i++){ 
        if(t >= dates[i] && t <= dates[i+1] ) break;
        }
    gap = t - dates[i];
    gap = gap/(dates[i+1] - dates[i]);
    gap = values[i] + gap*(values[i+1] - values[i]);
    return(gap);
    }
//=====================================================================
// isnum returns 1 if c is  a digit, '.', '-', '+'.
short isnum(char c)
{

printf("\nWarning: function 'isnum' is being called\n");
    if (isdigit(c)) return 1;
        // doesn't work right with 'e':  when searching through text, returns
        // when it comes across e.
    if(c == '.' || c == '-' || c == '+' || c == 'e') return 1;
    return 0;
    }
#if !defined(OS2COMPILE)
#if defined (MAXMINSHORT)
//=====================================================================
short max(short a, short b)
{
    return(a < b ? b : a);
    }
//=====================================================================
short min(short a, short b)
{
    return(a < b ? a : b);
    }
#endif

//=====================================================================
float max(float a, float b)
{
    return(a < b ? b : a);
    }
//=====================================================================
float min(float a, float b)
{
    return(a < b ? a : b);
    }
#endif
// ====================================================================
/* MATCHWILD 
    This wildcard matching function matches any number of characters
    with a '*', and any single character with a '^'.  Note that '? '
    couldn't be used, because chop takes this to mean: "look at the
    buffer" */

short matchwild(char *name, char *wildcard) {
char *cp=wildcard, *dp=name, nextchar, prevchar;
short matches=TRUE;

if(strlen(name)==0) return(FALSE); /* Null name has no match. */
if(strlen(wildcard)==0) return(TRUE);  /* Null wildcard matches everything */

while(*cp) {
    if(*cp=='*'){ /* This pattern matches any number of characters */
        /* If it matches so far, then it will now match to the end: */
        if((nextchar= *(cp+1))=='\0') 
            return(matches);
        else if (nextchar=='^'||nextchar=='*') 
            return(-1); /* Illegal format */
        else {  /* Non-wildcard character follows. */
            /* Match all characters in name up to this character.  If we
                don't find a match, return false. */
            while(*dp){
                if(*dp != nextchar) {
                    matches=FALSE;
                    dp++;
                    }
                else {
                    matches=TRUE;
                    cp++; cp++; dp++;  /* Move both characters to next position */
                    break;
                    }
                }  /* end while */
            if(!matches) return(FALSE);
            }
        } /* end if '*'   From here go back to main loop. */
    else if(*cp=='^') { /* Any single character matches this, but it most be there. */
        if(*dp=='\0') return(FALSE);
        else {
            matches=TRUE;
            dp++;  cp++;
            }
        } /* end if '^'   */
    else {  /* All other characters */
        if(*cp!=*dp) return(FALSE);
        else {
            cp++; dp++;
            }
        } /* end all other */
    }
/* If wildcard has been "used up", are there still characters in name?
        Then there is no match. */
if(!*cp && *dp) matches=FALSE;
return(matches);   /* If we get down to here. */
}
/* =================================================================== */
void outofmem(char *s)
{
    printf("Out of memory");
    if(strlen(s) != 0) printf(" at %s.\n",s);
    else printf(".\n");
    exit(1);
    }
/*======================================================================*/
/* safelog - This is a Kernighan and Pike style wrapper around the log
    function.  It lets us trap a domain error, as well as keep the program
    going, if so desired.  For that, we use dymetap().
*/
float safelog(float x, char* text) {
if(x<=SMALL) {
    printf("Bad value sent to safelog(): %9.3f\r\n",x);
    dymetap();
    x = .01; // to allow tracing back to the calling routine when debugging.
    }
return(log(x));
}
/*======================================================================*/
/* safepow - same for pow() function
*/
float safepow(float x, float y, char* text) {
if(x<=SMALL) {
    if(text)
        printf("Bad value sent to safepow(): %9.3f from %s\r\n",x,text);
    else
        printf("Bad value sent to safepow(): %9.3f\r\n",x);
    dymetap();
    }
return(pow(x,y));
}
/*======================================================================*/
/* safesqrt - same for sqrt() function
*/
float safesqrt(float x, char* text) {
if(x<=SMALL) {
    printf("Bad value sent to safesqrt(): %9.3f\r\n",x);
    dymetap();
    }
return(sqrt(x));
}
/*======================================================================*/
char *sgetint(char *text, short width, short *num)  {
char *cp,*dp,s[40];  /* 40 is maximum length of field. Be careful here! */
short i;
for(cp=text,dp=s,i=0;i<width;*dp++=*cp++,i++);
*dp='\0';
*num = atoi(s);
return(cp);
}
/*======================================================================*/
char *sgetint(char *text, short width, short &num)  {
char *cp,*dp,s[40];  /* 40 is maximum length of field. Be careful here! */
short i;
for(cp=text,dp=s,i=0;i<width;*dp++=*cp++,i++);
*dp='\0';
num = atoi(s);
return(cp);
}
/*======================================================================*/
char *sgetfl(char *text, short width, float *x)  {
char *cp,*dp,s[40];  /* 40 is maximum length of field. Be careful here! */
short i;
for(cp=text,dp=s,i=0;i<width;*dp++=*cp++,i++);
*dp='\0';
*x = atof(s);
return(cp);
}
/*======================================================================*/
char *sgetfl(char *text, short width, float &x)  {
char *cp,*dp,s[40];  /* 40 is maximum length of field. Be careful here! */
short i;
for(cp=text,dp=s,i=0;i<width;*dp++=*cp++,i++);
*dp='\0';
x = atof(s);
return(cp);
}
// ====================================================================
char *sgets(char *text, short width, char *s)  {
char *cp,*dp; 
short i;
for(cp=text,dp=s,i=0;i<width;*dp++=*cp++,i++);
*dp='\0';
return(cp);
}
// ====================================================================
/*  TackExt()
    tack an extension onto the end of a file name if none is already there.
*/  
void TackExt(char *filename, char *ext)
{
    char *ptr;
    ptr = strchr(filename,'.');
    if(ptr)
        return;
    else{
        strcat(filename,".");
        strcat(filename,ext);
        }
    }
// ====================================================================
// TAP  -- Tap a key to continue
void tap()
{
    cerr << "\nTap a key to continue.  ESC to exit.";
    if(getch() == 27) 
        exit(1);
    }
/*=========================================================================*/
// ti -- Get the time and date back in a string.  Use standard C, not
//   special borland functions!
short ti(char *timestring)  {
time_t timer;
struct tm *tmnow;
char f1[40], f2[40], f3[40];

timer = time(NULL);
tmnow = localtime(&timer);

sprintf(timestring,"%d:%02d:%02d on %2d/%02d/%04d.\n",
    tmnow->tm_hour, tmnow->tm_min, tmnow->tm_sec, 
    tmnow->tm_mon+1, tmnow->tm_mday, tmnow->tm_year+1900);
return 0;
}
// ====================================================================
/*** Utilities from Numerical Recipes **********/
char msguti1[] = "Not enough space.\n";

float *vector(short nl, short nh)
{
    float *v;
    v = new float[nh-nl+1];
    if (!v) {
        printf(msguti1);
        return 0;
        }
    v = v-nl;
    return v;
}

// ====================================================================
float **matrix(short nrl, short nrh, short ncl, short nch)
{
    short i,k;
    float **m;

    m = new float*[nrh-nrl+1];
    if (!m) {
        printf(msguti1);
        return 0;
        }
    m -= nrl;
    for(i = nrl; i <= nrh; i++) 
        m[i] = 0;

    for(i=nrl;i<=nrh;i++) {
        m[i]=new float[nch-ncl+1];
        if (!m[i]) {
            printf(msguti1);
            for(k = nrl; k < i; k++) 
                delete (m[k]+ncl);
            delete (m+nrl);
            return(0);
            }
        m[i] -= ncl;
        }
    return m;
    }

// ====================================================================
void free_vector(float *v, short nl)
{
    v = v+nl;
    if(v)
        delete (v);
    }

// ====================================================================
void free_matrix(float **m, short nrl, short nrh, short ncl)
{
    short i;

    for(i=nrh;i>=nrl;i--) {
        if(m[i]+ncl) 
            delete[] (m[i]+ncl);
        }
    delete[] (m+nrl);
    }

// ====================================================================
short *ivector(short nl, short nh)
{
    short *v;

    v = new short[nh-nl+1];
    if (!v) {
        printf(msguti1);
        return(0);
        }
    v = v-nl;
    return v;
    }
/* =================================================================== */
void free_ivector(short *v, short nl)
{
    v = v+nl;
    if(v)
        delete[] (v);
    }
/* =================================================================== */
long *lvector(short nl, short nh)
{
    long *v;

    v = new long[nh-nl+1];
    if (!v) {
        printf(msguti1);
        return(0);
        }
    v = v-nl;
    return v;
    }
/* =================================================================== */
void free_lvector(long *v, short nl)
{
    v = v+nl;
    if(v)
        delete (v);
    }
/* =================================================================== */
short **imatrix(short nrl, short nrh, short ncl, short nch) {

short i,k;
short **m;

m=new short* [nrh-nrl+1];
if (!m) {
    printf(msgutil2,"imatrix",nrl,nrh,ncl,nch);
    return 0;
    }
m -= nrl;
for(i = nrl; i <= nrh; i++) 
    m[i] = 0;

for(i=nrl;i<=nrh;i++) {
    m[i]=new short [nch-ncl+1];
    if (!m[i]) {
        printf(msgutil3,i,"imatrix");
        for(k = nrl; k < i; k++) 
            delete[] (m[k]+ncl);
        delete[] (m+nrl);
        return(0);
        }
    m[i] -= ncl;
    }
return m;
}
/* =================================================================== */
void free_imatrix(short **m, short nrl, short nrh, short ncl) {
short i;
short **n;
short *o;
n=m+nrl;
if(!n) return;
for(i=nrh;i>=nrl;i--) {
    o = m[i] + ncl;
    if(o) {
        delete[] o;
        }
    }
//printf("Finishing free_imatrix...\n");
delete (m+nrl);  
}
