/* DYMSTART -- Various functions that are called when Dyme starts.
	These include InitMat, ReadTriang, ReadOutFix, and some Groups
	and FixBank functions.
*/

#include "dymesys.h"   // Interdyme system files.
#include "user.h"      // User-specific files (triang, ptriang, etc.)

#define OK 0

//======================================================================
/* INITMAT -- New version, that takes a reference to VamFile, which 
	will be set up as the default file.  VamFile must already have been
	opened in the calling program.
*/

short InitMat(VamFile& vf)
{
	/* This routine has been much simplified, putting more of the burden
		on the Matrix and Vector constructors.  Now it just sets up the 
		mats array, and dates and nvecs from the default VamFile. */

	short nvecs;
	if(!vf.IsOpen()) {
		cprintf("VamFile not opened in call to InitMat.\r\n");
		return(ERR);
		}
	// mats = new matid[MAXMAT+1];  // (doesn't work under DOS Extender!)
	startyear = vf.Getstartdate();
	endyear = vf.Getenddate();
	matcount = 0;
	pmatcount = 0;
	return 0;
	}
#ifndef INCL_IVECTOR
/* ReadTriang() *************************************************************
Read the triangularization vector from the file triang.dat

short ReadTriang()
{
	FILE *fpt;
	char s[240];
	short i,j;

	if((fpt = fopen("triang.dat","r")) == NULL){
		printf("Cannot open triang.dat file.\n");
		exit(1);
		}
	if(fgets(s,240,fpt) == NULL){
		return(ERR); // Hit end of file.
		}
	j = 0;
	getint(ntriang,fpt,s,j);
	if((triang = new short[ntriang]) == 0){
		cerr << "Not enough memory for Triang.\n";
		exit(1);
		}
	for(i = 0; i < ntriang; i++){
		if(getint(triang[i], fpt, s, j) == ERR){
			cerr << "Not enough elements in Triang.\n";
			exit(1);
			}
		}
	fclose(fpt);
	return(OK);
	}
*/
/* ReadPTriang() *************************************************************
Read the triangularization vector from the file triang.dat

short ReadPTriang()
{
	FILE *fpt;
	char s[240];
	short i,j;

	if((fpt = fopen("ptriang.dat","r")) == NULL){
		printf("Cannot open ptriang.dat file.\n");
		exit(1);
		}
	if(fgets(s,240,fpt) == NULL){
		return(ERR); // Hit end of file.
		}
	j = 0;
	getint(ntriang,fpt,s,j);
	if((ptriang = new short[ntriang]) == 0){
		cerr << "Not enough memory for Triang.\n";
		exit(1);
		}
	for(i = 0; i < ntriang; i++){
		if(getint(ptriang[i], fpt, s, j) == ERR){
			cerr << "Not enough elements in PTriang.\n";
			exit(1);
			}
		}
	fclose(fpt);
	return(OK);
	}
*/
#endif
/* ReadOutFix() **********************************************************
Read the fix types for output.  The fix type determines what Seidel
does to determine the output.

	e	use the equation to determine the output, having included
		the dis (discrepancy) variable into the final demand. Presumably,
		this is the normal option.  It says, "I want the model to tell
		me what the answer is.  I will limit my interference to adjusting,
		at most, the final demands and possibly the dis variable.
	d	take the output from the Vam file and dump the difference
		between it and what the equation gives into the dis variable.
		This option says, "I know what the answer is, and any apparent
		difference is demand that doesn't really exist."
	i	take the output from the Vam file and add the difference
		between it and what the equation gives to the imports.  This
		option says, "I know what domestic capacity is, and any difference
		between it and demand must be made up by imports."

The input file is a list of the sector titles with the outfix option
in column 18.  This format allows the same file to be used for the
fixes as for the .ttl file in VAM and keeps the outfixes close to the
sector names.
*/
short ReadOutFix(void)
{
	FILE *fpt;
	char s[240],options[501],opt;
	short i,n, err = 0;

	if((fpt = fopen("sectors.ttl","r")) == NULL){
		printf("Cannot open sectors.ttl file.\n");
		exit(1);
		}
	for(i = 1; i<=500; i++){
		if(fgets(s,240,fpt) == NULL) break; // Hit end of file.
		opt=s[17];
		if(opt == 'e' || opt == 'd' || opt == 'i') {
			options[i] = opt;
			}
		else{
			printf("Illegal outfix code %c in sector %d replaced by e.\n", opt, i);
			err++;
			}
		}
	fclose(fpt);
	if(err > 0) tap();
	if(i == 500) n = 500;
	else n = i--;
	outfix = new char [n+1];
	for(i = 1; i <= n; i++)
		outfix[i] = options[i];
	return(OK);
	}
#ifndef INCL_IVECTOR
/* ReadTriang() *************************************************************
Read the triangularization vector from the file triang.dat
*/
short ReadTriang()
{
	FILE *fpt;
	char s[240];
	short i,j;

	if((fpt = fopen("triang.dat","r")) == NULL){
		printf("Cannot open triang.dat file.\n");
		exit(1);
		}
	if(fgets(s,240,fpt) == NULL){
		return(ERR); // Hit end of file.
		}
	j = 0;
	getint(ntriang,fpt,s,j);
	if((triang = new short[ntriang]) == 0){
		cerr << "Not enough memory for Triang.\n";
		exit(1);
		}
	for(i = 0; i < ntriang; i++){
		if(getint(triang[i], fpt, s, j) == ERR){
			cerr << "Not enough elements in Triang.\n";
			exit(1);
			}
		}
	fclose(fpt);
	return(OK);
	}
#endif

