// GBANK.CPP -- Methods and utilities for simple G Bank.

#include <stdio.h>  // Stick with C streams for now!
#include <conio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#if !defined(OS2COMPILE)
#if !defined(DJGPP)
#include <alloc.h>
#endif
#endif
#include <string.h>

#include "message.h"
#include "gbank.h"
#include "constant.h"

#if !defined(ERR)
#define ERR (-1)
#endif
short matchwild(char*, char*);


/*==========================================================================*/
char * GBank::Setbnktitle(const char *s) {
bnkHeader Hdr;

bnktitle = new char[strlen(s)+1];
strcpy(bnktitle,s);
Hdr.jahra=jahra;
Hdr.nobs=nobs;
Hdr.stper=stper;
strcpy(Hdr.bnktitle,bnktitle);
// update bank with title, too.
if((fseek(pbank,0L,0)) != 0) {   /* posit at begin of file */
   cprintf(MSGBK(15));
   return(NULL);
   }
if((fwrite( (void*)&Hdr,sizeof(Hdr),1,pbank))==  0){
		cprintf(MSGBK(1),bnkname);
      return(NULL);
      }
return bnktitle;
}
/*==========================================================================*/
/* ListNames - list series names of a regular GBank */
short GBank::ListNames(FILE* out, char* mask, short wide) {
short i,c,matchcount,toscreen,ObsPerLine;
char *cp,s[40];

if(wide)
	ObsPerLine=5;
else ObsPerLine=1;
if(out==stdout)
	toscreen=TRUE;
i = matchcount = 0;
if(nseries > 100 && toscreen) {
	#if !defined(OS2COMPILE)
	clrscr();
	gotoxy(1,1);
	#endif
	}
cp = names;
while (i<nseries){
	i++;
	if(matchwild(cp,mask)) {
		if(matchcount && matchcount % ObsPerLine == 0) {
			if(toscreen) cprintf("\r\n");
			else fprintf(out,"\n");
			}
		if( (toscreen) && matchcount && matchcount % (ObsPerLine*24) == 0){
			cprintf("\r");
			cprintf(MSGLSC(11));
			c = getch();
			if(c == ESC) break;
			#if !defined(OS2COMPILE)
			clrscr();
			gotoxy(1,1);
			#endif
		}
		if(toscreen) cprintf("%-15s",cp);
		else fprintf(out,"%-15s",cp);
		matchcount++;
		}
	while(*(cp++) != '\0'); 
	}
return(OK);
}
/*==========================================================================*/
// Copy constructor.  This allows you to do a statement like:
//              gb = GBank(other);    where "other" is another GBank object.

GBank::GBank(GBank& gb) {
short i;
classname = new char[strlen(gb.classname)+1];
strcpy(classname,gb.classname);
bnkname = new char[strlen(gb.bnkname)+1];
strcpy(bnkname,gb.bnkname);
indname = new char[strlen(gb.indname)+1];
strcpy(indname,gb.indname);
bnktitle = new char[strlen(gb.bnktitle)+1];
strcpy(bnktitle,gb.bnktitle);
opened=gb.opened;
pbank=gb.pbank;
ipbank=gb.ipbank;
jahra=gb.jahra;
nseries=gb.nseries;
nopy=gb.nopy;
stper=gb.stper;
nobs=gb.nobs;
status=gb.status;
MsgObject=gb.MsgObject;
indx=new unsigned short[nseries];
for(i=0;i<nseries;i++)
	indx[i] = gb.indx[i];
ncin=gb.ncin;
names = new char[ncin];
for(i=0;i<ncin;i++)
	names[i] = gb.names[i];
}
/*==========================================================================*/
// default constructor, for simple declarations.  Use with assignbank
//    method below:
GBank::GBank(void) {
bnktitle=0;
classname=0;
bnkname=0;
indname=0;
nseries=0;
pbank=0;
ipbank=0;
indx=0;
names=0;
status=0;
MsgObject=new GMessage;
}
/*==========================================================================*/
GBank::GBank(const char *rootname, const char *Status,
       short BaseYr, short nobserv, short st) {
// Opens a new bank

short err,i,j;
long posit;
char fname[120];
bnkHeader Hdr;

//printf("Constructing object for GBank %s.\n",rootname);

if(strcmp(Status,"New")!=0) {
   cprintf("GBank constructor only works with New at present.\n\r");
   return;
   }
bnktitle=0;
classname=0;
bnkname=0;
indname=0;
nseries=0;
pbank=0;
ipbank=0;
indx=0;
names=0;
status=0;
MsgObject=new GMessage;

strcpy(fname,"GBank");
classname=new char[strlen(fname)+1];
strcpy(classname,fname);
/* Open .bnk file and initialize. */
strcpy(fname,rootname);
strcat(fname,".bnk");
bnkname= new char[strlen(fname)+1];
strcpy(bnkname,fname);
jahra = Hdr.jahra = BaseYr;
nobs = Hdr.nobs = nobserv;
stper = Hdr.stper = st;
if((pbank = fopen(bnkname,"wb+")) == 0 ||
	(fwrite( (void*)&Hdr,sizeof(Hdr),1,pbank))==  0){
		cprintf(MSGBK(1),bnkname);
		banktype = 'x';
      goto failure;
      }

/* Open the .ind file and initialize. */
strcpy(fname,rootname);
strcat(fname,".ind");
indname= new char[strlen(fname)+1];
strcpy(indname,fname);
status='w';   // write status

if((ipbank = fopen(indname,"wb+")) == 0) {
	cprintf(MSGBK(2),indname);
	goto failure;
   }
posit = 0;
if((fseek(ipbank,posit,0)) != 0) {   /* posit at begin of file */
   cprintf(MSGBK(15));
   goto failure;
   }

nseries=0;
if((fwrite(&nseries,sizeof(short),1,ipbank)) == 0) {    /* read the number of series */
   cprintf("Unable to write to index.\n\r");
   goto failure;
   }
indx = new unsigned short[NSMAX+1];
for(i=0;i<NSMAX;i++)
   indx[i]=0;

ncin=0;
names = new char[NCMAX];
strcpy(names,"");

banktype = 'w';
goto success;
failure:
   opened = -1;
   if(classname) {
		delete[] classname;
		classname=0;
		}
   if(bnkname) {
		delete[] bnkname;
		bnkname=0;
		}
   if(indname) {
		delete[] indname;
		indname=0;
		}
	if(indx) {
		delete[] indx;
		indx=0;
		}
	if(names) {
		delete[] names;
		names=0;
		}
   if(MsgObject) {
      delete MsgObject;
      MsgObject=0;
      }

   return;
success:
   opened = 1;
   return;
}
/*==========================================================================*/
// AssignBank -- Used with simple constructor to assign bank to globally 
//    defined G Bank.
short GBank::AssignBank(const char *rootname) {
short err;
long posit;
char fname[120];
bnkHeader Hdr;

//printf("Constructing object for GBank %s.\n",rootname);

bnktitle=0;
classname=0;
bnkname=0;
indname=0;
nseries=0;
pbank=0;
ipbank=0;
indx=0;
names=0;
status=0;

strcpy(fname,"GBank");
classname=new char[strlen(fname)+1];
strcpy(classname,fname);

/* Open .bnk file and initialize. */
strcpy(fname,rootname);
strcat(fname,".bnk");
bnkname= new char[strlen(fname)+1];
strcpy(bnkname,fname);
if((pbank = fopen(bnkname,"rb+")) == 0 ||
	(fread( (void*)&Hdr,sizeof(Hdr),1,pbank))==  0){
		cprintf(MSGBK(1),bnkname);
		banktype = 'x';
	  goto failure;
	  }
jahra = Hdr.jahra;
jahra = (jahra<1900)?jahra+1900:jahra;
stper = Hdr.stper;
nobs = Hdr.nobs;
bnktitle = new char[strlen(Hdr.bnktitle)+1];
strcpy(bnktitle,Hdr.bnktitle);

/* Open the .ind file and initialize. */
strcpy(fname,rootname);
strcat(fname,".ind");
indname= new char[strlen(fname)+1];
strcpy(indname,fname);
status='r';

if((ipbank = fopen(indname,"rb+")) == 0) {
	cprintf(MSGBK(2),indname);
	goto failure;
   }
posit = 0;
if((fseek(ipbank,posit,0)) != 0) {   /* posit at begin of file */
   cprintf(MSGBK(15));
   goto failure;
   }

if((fread(&nseries,sizeof(short),1,ipbank)) == 0) {    /* read the number of series */
   cprintf(MSGBK(16));
   goto failure;
   }
indx = new unsigned short[nseries+1];

if((fread(indx,sizeof(short),nseries+1,ipbank)) == 0) {  /* read the pointer */
	  cprintf(MSGBK(17));
	  goto failure;
	  }
ncin = indx[nseries]+1;
names = new char[ncin];

if((fread(names,1,ncin,ipbank)) == 0) {  /* read the names  */
   cprintf(MSGBK(18));
   goto failure;
   }

banktype = 'w';
goto success;
failure:
   opened = -1;
   if(classname) {
		delete[] classname;
		classname=0;
		}
   if(bnkname) {
		delete[] bnkname;
		bnkname=0;
		}
   if(indname) {
		delete[] indname;
		indname=0;
		}
	if(indx) {
		delete[] indx;
		indx=0;
		}
	if(names) {
		delete[] names;
		names=0;
		}
   if(MsgObject) {
      delete MsgObject;
      MsgObject=0;
      }
   return -1;
success:
   opened = 1;
   return 0;
}
/*==========================================================================*/
GBank::GBank(const char *rootname)  {
// Opens an existing bank, like G bankcmd, but uses rootname.

short err;
long posit;
char fname[120];
bnkHeader Hdr;

//printf("Constructing object for GBank %s.\n",rootname);

bnktitle=0;
classname=0;
bnkname=0;
indname=0;
nseries=0;
pbank=0;
ipbank=0;
indx=0;
names=0;
status=0;
MsgObject=new GMessage;

strcpy(fname,"GBank");
classname=new char[strlen(fname)+1];
strcpy(classname,fname);

/* Open .bnk file and initialize. */
strcpy(fname,rootname);
strcat(fname,".bnk");
bnkname= new char[strlen(fname)+1];
strcpy(bnkname,fname);
if((pbank = fopen(bnkname,"rb")) == 0 ||
	(fread( (void*)&Hdr,sizeof(Hdr),1,pbank))==  0){
		cprintf(MSGBK(1),bnkname);
		banktype = 'x';
      goto failure;
      }
jahra = Hdr.jahra;
jahra = (jahra<1900)?jahra+1900:jahra;
stper = Hdr.stper;
nobs = Hdr.nobs;
bnktitle = new char[strlen(Hdr.bnktitle)+1];
strcpy(bnktitle,Hdr.bnktitle);

/* Open the .ind file and initialize. */
strcpy(fname,rootname);
strcat(fname,".ind");
indname= new char[strlen(fname)+1];
strcpy(indname,fname);
status = 'r';  // read status

if((ipbank = fopen(indname,"rb")) == 0) {
	cprintf(MSGBK(2),indname);
	goto failure;
   }
posit = 0;
if((fseek(ipbank,posit,0)) != 0) {   /* posit at begin of file */
   cprintf(MSGBK(15));
   goto failure;
   }

if((fread(&nseries,sizeof(short),1,ipbank)) == 0) {    /* read the number of series */
   cprintf(MSGBK(16));
   goto failure;
   }
indx = new unsigned short[nseries+1];

if((fread(indx,sizeof(short),nseries+1,ipbank)) == 0) {  /* read the pointer */
      cprintf(MSGBK(17));
      goto failure;
      }
ncin = indx[nseries]+1;
names = new char[ncin];

if((fread(names,1,ncin,ipbank)) == 0) {  /* read the names  */
   cprintf(MSGBK(18));
   goto failure;
   }

banktype = 'w';
goto success;
failure:
   opened = -1;
   if(classname) {
		delete[] classname;
		classname=0;
		}
   if(bnkname) {
		delete[] bnkname;
		bnkname=0;
		}
   if(indname) {
		delete[] indname;
		indname=0;
		}
	if(indx) {
		delete[] indx;
		indx=0;
		}
	if(names) {
		delete[] names;
		names=0;
		}
   if(MsgObject) {
      delete MsgObject;
      MsgObject=0;
      }
   goto end;
success:
   opened = 1;
end:
   opened=opened;
}
/*==========================================================================*/
// free up memory allocated in constructor.  close data files.
GBank::~GBank() {
//printf(" GBank Destructor called.  Freeing up memory.\n");
if(status=='w')
   windex();  // make sure to write index, if 'n' option used in addser.
if(classname) {
	delete[] classname;
	classname=0;
	}
if(bnkname) {
	delete[] bnkname;
	bnkname=0;
	}
if(indname) {
	delete[] indname;
	indname=0;
	}
if(bnktitle) {
   delete[] bnktitle;
   bnktitle=0;
   }
if(indx) {
	delete[] indx;
	indx=0;
	}
if(names) {
	delete[] names;
	names=0;
	}
if(MsgObject)
	delete MsgObject;
fclose(pbank);
fclose(ipbank);
}
/*==========================================================================*/
/* findname -- finds position of given name in GBank names array. */
short GBank::findname(char *gname) {
short i;
for(i=0;i<nseries;i++)
   if(strcmp(gname,&names[indx[i]])==0)
      return(i);
return(-1);
}
/*==========================================================================*/
/* GetName -- Reverse of findname, gets name at certain position */
char* GBank::GetName(short loc,char *gname) {
if(loc<0 || loc> nseries)
   return(0);
strcpy(gname,&names[indx[loc]]);
return(gname);
}
/*==========================================================================*/
short GBank::windex(void)  { // Write index to index file.
long posit;
short err;
unsigned long core;

posit = 0;
if(( fseek(ipbank,posit,0)) != 0) {   /* position at begining of file */
	cprintf(MSGBK(11));
	return(ERR);
	}

/* nseries is number of series in bank */
if((fwrite(&nseries,sizeof(short),1,ipbank)) == 0) {   /* write number of series */
	cprintf(MSGBK(12));
   return(ERR);
   }

if((fwrite(indx,2,nseries+1,ipbank)) == 0) {  /* write the pointer */
   cprintf(MSGBK(13));
   return(ERR);
   }
/* find ncin, the number of characters actually in the names */
ncin = indx[nseries]+1;
if((fwrite(names,1,ncin,ipbank)) == 0) {  /* write the names  */
   cprintf(MSGBK(14));
   return(ERR);
   }

//core = coreleft();
return(0); 
}
/*==========================================================================*/
short GBank::rindex(void)  { // Read index from index file.
long posit,err;

posit = 0;
if((fseek(ipbank,posit,0)) != 0) {   /* posit at begin of file */
   cprintf(MSGBK(15));
   return(ERR);
   }

if((fread(&nseries,2,1,ipbank)) == 0) {    /* read the number of series */
   cprintf(MSGBK(16));
   return(ERR);
   }
if((fread(indx,2,nseries+1,ipbank)) == 0) {  /* read the pointer */
   cprintf(MSGBK(17));
   return(ERR);
   }
ncin = indx[nseries] + 1;
if((fread(names,1,ncin,ipbank)) == 0) {  /* read the names  */
   cprintf(MSGBK(18));
   return(ERR);
   }

cprintf(MSGBK(19));
return(0);
}

/*==========================================================================*/
//  Virtual version just has two arguments.  It, in turn, calls the 
//    "non-virtual version with the default argument.
short GBank::getser(char *nam, float *f) {
return ( getser(nam,f,-1) );
}
/*==========================================================================*/
// This version of GETSER is associated with the GBank class, and so
//   knows information relevant to finding series from data in the
//   GBank object. 

short GBank::getser(char *nam, float *ss, short SeriesLength) {

short err,i,j,gap,max_obs;
long temp,posit;
float val;
char *pnam;

if((err = findname(nam)) == ERR) goto error;
temp = err;     /* if in workspace, position and read */
posit = temp * (long)nobs * sizeof(float) + temp + sizeof(bnkHeader);
if((err = fseek(pbank,posit,0)) == ERR) {
   cprintf("\r\n cannot fseek \n\r");
   return(ERR);
   }
if((err=fread(&cnpy,1,1,pbank))==0 )  {
   cprintf(" \n cannot read %s from %d\n ",nam,bnkname);
   return(ERR);
   }
nopy=cnpy; 
if(SeriesLength==-1) SeriesLength=nobs;  // check if this arg was given.
for( i=0;i<SeriesLength;i++)
	ss[i] = 0;
if((err=fread((void*)ss,sizeof(float),SeriesLength,pbank))==0)  {
   cprintf("\r\n cannot read %s from %d\n\r",nam,bnkname);
   return(ERR);
   }
return 0;
error:  return(ERR);
} 
/*==========================================================================*/
// ADDSER for GBank class.  Puts a series in a bank.  Returns series number.
short GBank::addser(char *nam, float *f,char wrindex_flag,short SeriesLength) {
short i,err,series_number;
unsigned short j;
long posit,temp;
char cnpy;

cnpy = nopy;
/* does name exist? */
if((series_number = findname(nam)) == -1) {
   series_number = nseries;
   i=0;
   j=*(indx+nseries++);           /* if not add to bank */
   while(nam[i] != '\0') {
      *(names+j++) = nam[i];
      i++;
      }
   *(names+j++) = '\0';
   *(indx+nseries)=j;
   temp = nseries -1;
   }
else {
   temp = series_number;
   }

posit = temp * (long)(nobs*sizeof(float)+1)+sizeof(bnkHeader);   /* posit and write */
if((fseek(pbank,posit,0)) != 0) {
   cprintf(MSGBK(7));
   nseries--;
   return (-1);
   }
// (each series written with its own frequency)
if(SeriesLength==-1) SeriesLength=nobs;
if((fwrite(&cnpy,1,1,pbank))==0 ||
	   (fwrite(f,sizeof(float),SeriesLength,pbank))==0){
	cprintf(MSGBK(8),nam,pbank);
	nseries--;
	return(ERR);
   }

if(wrindex_flag=='y') {  // default is 'y'. remember to write index!
   if((windex()) == ERR) {
      nseries--;
	   return(ERR); /* write index file  */
      }
   }
if(nseries == NSMAX || *(indx+nseries) > NCMAX -20){
	cprintf(MSGBK(9));
	cprintf(MSGBK(10),nseries,*(indx+nseries));
	return(ERR);
	} 

return(series_number);
}
//==========================================================================
short GBank::flush(void) {
fflush(pbank);
fflush(ipbank);
return 0;
}
/*==========================================================================*/
// removed check_lag (DSM - 8/29/93)
/*==========================================================================*/
/* search is to locate the variable name in a series of names */
short search(unsigned short *ind, char *nams, char *nam, short num) {
short i,ten;
char *pn;

for(i=0;i<num;i++) {
   pn = &nams[ind[i]];
	if((strcmp(pn,nam)) == 0) return(i);
   }
return(ERR);
}

