// GROUPS.H -- Class definition of group class.  This allows for an easy
//   and flexible syntax of group specifications.
//#define OS2COMPILE

#if (!defined(GROUPS_H))
#define GROUPS_H  1

#include "matrix.h"

struct GroupDefn  {   // One group definition record, with tag and size.
   size_t GDsize;     // These are variable length.  GDsize is the size.
   short Nin;           // Number of group members, for check.
   short *GroupElems;   // This is the list of group elements.
   char *GroupTag;    // This is the group name
   GroupDefn() {GroupTag=NULL,GroupElems=NULL;}
   ~GroupDefn();
};


class GroupHome  {
 public:
	enum {MAXGROUP=300,MAXTAGLEN=20,MAXSEC=750};
 protected:
   short NGroups;       // Number of groups defined.
   long start, psn[MAXGROUP+1];  // positions of each group defn in file.
 public:
   GroupDefn *gd; // Array of group definitions.  (can be allocated
                       //   NGroups or MAXGROUP )
   short OldGroups;     // Number of groups already in file.
   short GroupsRead;     // Number of groups most recently read, either from 
                       //   ASCII or binary file.  Use for error chks.
 protected:
   FILE *fpbin,*fptext;
   short *mlist;    // A convenient list scratch space for methods.
      // (allocated by constructor)
 public:
   short GroupsReplaced; // Number of groups that had a new group definition
                       //   of <= the old Nin (number of group elements)

   GroupHome(const char *grpfile);   // read in group defns from file  
   GroupHome(void);
   void Load(void);          // read in from GROUPS.BIN.
   short GroupsIn(void) {return GroupsRead;}  // test for success of constructor
   ~GroupHome();

   // fill supplies the magic behind the groups definition sytax:
   short fill(FILE *ifp, char *gString, short msize, short members[]);
   // Sfill borrows from fill, reads from the string gString
   short Sfill(char *gString, short msize, short members[]);

   short GetNGroups(void)  {return NGroups;}
   // GetMemLst gets a member vector consisting of 1's and 0's.
   short GetMemLst(char *tag, short lSize, short *members);
   // lSize allows members array to be different lengths.
	short NinGroup(char *tag);  // Returns number of members.
   short GetGroupNo(char *tag); // Returns index of group in file.
	short GroupList(char *tag, short *list); // returns list of group members
   short PrintGroup(FILE *chk, char *tag); // prints group definition to file

   short AddGroup(char *tag, char *gs);   // Used by constructor.
   short ReadGroup(FILE *iin, char *tag, char *gs);  // ASCII input method

   short wrtHeader(void);
   short rdHeader(void);
	friend short Vector::fix(short t, short SectorFlag=0);
//	friend short FixBank::PutInGroupNums(void);//  

};


#endif

