// MISCTOOLS -- Some miscellaneous tools used by Inforum programs.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>      // for time and date functions
#include <ctype.h>
#include <alloc.h>
#include <conio.h>
#include <mem.h>

#include "message.h"

/* Prototypes... */
float *vector(short nl, short nh);
double *dvector(short nl, short nh);
float **matrix(short nrl, short nrh, short ncl, short nch);
double **dmatrix(short nrl, short nrh, short ncl, short nch);
void free_vector(float *v, short nl);
void free_dvector(double *v, short nl);
void free_matrix(float **m, short nrl, short nrh, short ncl);
void free_dmatrix(double **m, short nrl, short nrh, short ncl);
short *ivector(short nl, short nh);
void free_ivector(short *v, short nl);
long *lvector(short nl, short nh);
void free_lvector(long *v, short nl);
void outofmem(char *s);
void memout(char *s);
short breakname(char *nam, char *vnam);
short breakname(char *nam, char *vnam, short &prow);
short ti(char *timestring);
void tap();
short getopt(FILE *fp, char* line, short maxchar);
void depad(char *s);
short lint(float *f,short nobs_ws);
short linsub(float *x, short m, short n);
short stepint(float *f,short nobs_ws);
short stepsub(float *x, short m, short n);
short matchwild(char *name, char *wildcard);

short IsMissing(float x);
/* End prototypes. */

const short ESC = 27;
const short TRUE = 1;
const short FALSE = 0;

GMessage UtilMsgObject;  // Global object, if misctool is linked.
#define MSGUTIL1(n) UtilMsgObject.GetGMessage("msgutil",n)

/* =================================================================== */
short IsMissing(float x) {
if(x <= -.0000009 && x >= -.0000011)
	return 1;
return 0;
}
/* =================================================================== */
float *vector(short nl, short nh) {

float *v;
unsigned size;

v= new float[nh-nl+1];
if (!v) {
	printf(MSGUTIL1(1),"vector",nl,nh);
	return 0;
	}
return v-nl;
}
/* =================================================================== */
double *dvector(short nl, short nh) {

double *v;

v=new double[nh-nl+1];
if (!v) {
	printf(MSGUTIL1(1),"dvector",nl,nh);
	return 0;
	}
return v-nl;
}
/* =================================================================== */
float **matrix(short nrl, short nrh, short ncl, short nch) {

short i,k;
float **m;

m=new float* [nrh-nrl+1];
if (!m) {
	printf(MSGUTIL1(2),"matrix",nrl,nrh,ncl,nch);
	return 0;
	}
m -= nrl;
for(i = nrl; i <= nrh; i++) 
   m[i] = 0;

for(i=nrl;i<=nrh;i++) {
	m[i]=new float [nch-ncl+1];
	if (!m[i]) {
		printf(MSGUTIL1(3),i,"matrix");
		for(k = nrl; k < i; k++) 
			delete[] (m[k]+ncl);
		delete[] (m+nrl);
		return(0);
		}
	m[i] -= ncl;
	}
return m;
}
/* =================================================================== */
double **dmatrix(short nrl, short nrh, short ncl, short nch) {

short i,k;
double **m;

m= new double* [nrh-nrl+1];
if (!m) {
	printf(MSGUTIL1(2),"dmatrix",nrl,nrh,ncl,nch);
	return(0);
	}
m -= nrl;
for(i = nrl; i <= nrh; i++) 
	m[i] = 0;

for(i=nrl;i<=nrh;i++) {
	m[i]= new double [nch-ncl+1];
	if (!m[i]) {
		printf(MSGUTIL1(3),i,"dmatrix");
		for(k = nrl; k < i; k++) 
         delete[] (m[k]+ncl);
		delete[] (m+nrl);
		return(0);
		}
	m[i] -= ncl;
	}
return m;
}
/* =================================================================== */
void free_vector(float *v, short nl) {
if((v+nl)) 
	delete[] (v+nl);
}
/* =================================================================== */
void free_dvector(double *v, short nl) {
if((v+nl)) 
	delete[] (v+nl);
}
/* =================================================================== */
void free_matrix(float **m, short nrl, short nrh, short ncl) {
short i;
float **n;
float *o;
n=m+nrl;
if(!n) return;
for(i=nrh;i>=nrl;i--) {
	o = m[i] + ncl;
	if(o) {
		delete[] o;
		}
	}
delete (m+nrl);  
}
/* =================================================================== */
void free_dmatrix(double **m, short nrl, short nrh, short ncl) {
short i;
double **n;
double *o;
n=m+nrl;
if(!n) return;
for(i=nrh;i>=nrl;i--) {
   o=m[i]+ncl;
	if(o) 
		delete[] (m[i]+ncl);
	}
delete[] (m+nrl);
}
/* =================================================================== */
short *ivector(short nl, short nh) {
short *v;
v=new short [nh-nl+1];
if (!v) {
	printf(MSGUTIL1(1),"ivector",nl,nh);
	return(0);
	}
return v-nl;
}
/* =================================================================== */
void free_ivector(short *v, short nl) {
if((v+nl) ) 
	delete[] (v+nl);
}
/* =================================================================== */
long *lvector(short nl, short nh) {
long *v;

v=new long [nh-nl+1];
if (!v) {
	printf(MSGUTIL1(1),"lvector",nl,nh);
	return(0);
	}
return v-nl;
}
/* =================================================================== */
void free_lvector(long *v, short nl) {
if((v+nl)) 
	delete[] (v+nl);
}
/* =================================================================== */
void outofmem(char *s) {
printf("Out of memory");
if(strlen(s) != 0) printf(" at %s.\n",s);
else printf(".\n");
exit(1);
}
/* =================================================================== */
// BREAKNAME is overloaded: a simple version and a more complicated.
/*** breakname divides a vector-sector name into a vector name and a sector.
     (The sector is the return value.  */

short breakname(char *nam, char *vnam) {
short len,i,j, tag, sec;
char s[10];

len = strlen(nam);
i = len-1;
while(isdigit(nam[i])) i--;
tag = i;
i++;
j = 0;
while((s[j++]=nam[i++]) != '\0');
if((len = strlen(s)) == 0) sec = 1;
else sec = atoi(s);
i = 0;
while(i <= tag){
	vnam[i] = nam[i];
	i++;
	}
vnam[i] = '\0';
return(sec);
}
/*=========================================================================*/
/* breakname() devides a vector-sector name into a vector name and a sector 
	For a name of the form pce27, "pce" will be put into vnam and the return
	value will be 27.  For a name of the form am3.7, (am row 3, col 7),
	vname will be returned as "am", *prow as 3, and the return value is 7.
*/ 
short breakname(char *nam, char *vnam, short &prow) {

short len,i,j, tag, sec;
char s[10];

len = strlen(nam);
i = len-1;
while(isdigit(nam[i])) i--;
tag = i;
i++;
j = 0;
while((s[j++]=nam[i++]) != '\0');
if((len = strlen(s)) == 0) sec = 0; // Was sec = 1, changed 23 Nov 92 for indexcmd
else sec = atoi(s);

if(nam[tag] == '.'){
	i = tag-1;
	while(isdigit(nam[i])) i--;
	tag = i;
	i++;
	j = 0;
	while((s[j++] = nam[i++]) != '.');
	s[j-1] = '\0';
	if((len = strlen(s)) == 0) prow = -1;
	else prow = atoi(s);
	}

i = 0;
while(i <= tag){
	vnam[i] = nam[i];
	i++;
	}
vnam[i] = '\0';
return(sec);
}
/*=========================================================================*/
// ti -- Get the time and date back, in a friendly format, in a string.
short ti(char *timestring)  {
struct time dtime; // dos.h must be included.
struct date date;
char f1[40], f2[40], f3[40];
gettime(&dtime);
getdate(&date);
if(dtime.ti_hour<12) strcpy(f1," in the morning, on ");
else if(dtime.ti_hour<16) {
   if(dtime.ti_hour<13)
      strcpy(f1," Lunch Time!, on ");
   else strcpy(f1," in the afternoon, on ");
   }
else if(dtime.ti_hour<22) {
   if(dtime.ti_hour>18 && dtime.ti_hour<20)
      strcpy(f1," Dinner Time!, on ");
   else strcpy(f1," in the evening, on ");
   }
else strcpy(f1," at night, on ");
if(dtime.ti_hour>=13) dtime.ti_hour -=12;
if(dtime.ti_hour<1) dtime.ti_hour +=12;

sprintf(timestring,"%d:%02d:%02d%s%d/%02d/%02d.",
	dtime.ti_hour, dtime.ti_min, dtime.ti_sec, f1,
   date.da_mon, date.da_day, date.da_year-1900);
return 0;
}
/*=========================================================================*/
/* tap() = Tap a key to continue. */
void tap(void) {
printf("\nTap a key to continue.  ESC to exit.");
if(getch() == ESC) 
	exit(1);
}
/*======================================================================*/
/* depad() -- remove trailing blanks from a string */
void depad(char *s) {
short i,len;

len = strlen(s);
i = len-1;  
while(s[i] == ' ' && i > 0) 
   i--;
if(i > 0) 
   s[i+1] = '\0';
else s[0] = '\0';
}
/*======================================================================*/
short lint(float *f,short nobs_ws)  {
int i,err;
i = 0;
while(IsMissing(f[i])) i++;
while ((i = linsub(f,i,nobs_ws)) < nobs_ws);
return(0);
}
/*=======================================================================*/
short linsub(float *x, short m, short n)  {

short i,j,istart;
float dif;

i = m;
while(!(IsMissing(x[i])) && i < n) i++;
if (i == n) return(i);
istart = i-1;
while(IsMissing(x[i]) && i < n) i++;
if (i < n ){
	dif = (x[i] - x[istart])/(i - istart);
	for (j = istart+1; j < i; j++)
		x[j] = x[j-1] + dif;
	}
return(i);
} 
/*======================================================================*/
short stepint(float *f,short nobs_ws)  {
short i,err;
i = 0;
while(IsMissing(f[i])) i++;
while ((i = stepsub(f,i,nobs_ws)) < nobs_ws);
return(0);
}
/*=======================================================================*/
short stepsub(float *x, short m, short n)  {

short i,j,istart;
float dif;

i = m;
while(!(IsMissing(x[i])) && i < n) i++;
if (i == n) return(i);
istart = i-1;
while(IsMissing(x[i]) && i < n) i++;
if (i < n ){
	for (j = istart+1; j < i; j++)
		x[j] = x[istart];
	}
return(i);
} 
/*=======================================================================*/
void memout(char *s) {
#ifdef INCL_CORELEFT
printf("Out of memory %s. Core left %ld.\n",s,coreleft());
exit(1);
#endif
}
/*=======================================================================*/
short max(short a, short b){
return a > b ? a : b;
}
short min(short a, short b){
return a > b ? b: a;
}

/* =============================================================== */
/*  This wildcard matching function matches any number of characters
    with a '*', and any single character with a '^'.  Note that '? '
    couldn't be used, because chop takes this to mean: "look at the
    buffer" */
   
short matchwild(char *name, char *wildcard) {
char *cp=wildcard, *dp=name, nextchar, prevchar;
short matches=TRUE;

if(strlen(name)==0) return(FALSE); /* Null name has no match. */
if(strlen(wildcard)==0) return(TRUE);  /* Null wildcard matches everything */

while(*cp) {
	if(*cp=='*'){ /* This pattern matches any number of characters */
		/* If it matches so far, then it will now match to the end: */
		if((nextchar= *(cp+1))=='\0') 
			return(matches);
		else if (nextchar=='^'||nextchar=='*') 
			return(-1); /* Illegal format */
		else {  /* Non-wildcard character follows. */
         /* Match all characters in name up to this character.  If we
   	      don't find a match, return false. */
			while(*dp){
				if(*dp != nextchar) {
					matches=FALSE;
					dp++;
					}
				else {
					matches=TRUE;
					cp++; cp++; dp++;  /* Move both characters to next position */
					break;
					}
				}  /* end while */
			if(!matches) return(FALSE);
         }
      } /* end if '*'   From here go back to main loop. */
   else if(*cp=='^') { /* Any single character matches this, but it most be there. */
		if(*dp=='\0') return(FALSE);
		else {
			matches=TRUE;
			dp++;  cp++;
			}
		} /* end if '^'   */
	else {  /* All other characters */
		if(*cp!=*dp) return(FALSE);
		else {
			cp++; dp++;
			}
		} /* end all other */
	}
/* If wildcard has been "used up", are there still characters in name?
   	Then there is no match. */
if(!*cp && *dp) matches=FALSE;
return(matches);   /* If we get down to here. */
}
//=========================================================================
