/* Packed matrix (Pmatrix) operator and function definitions
	Some functions, such as adding and subtracting of Pmatrices have been
	removed for this copy for use in Dyme.
	*/

#include "dymesys.h" // Interdyme system files.

extern short vcount;
//////////////////////  THE PMATRIX MEMBER FUNCTIONS /////////////////////////
Pmatrix::Pmatrix(short anr, short anc, char atemp, short afr, short afc)
{
	short i;
	char wheretxt[60];

	which = vcount++;
	if(atemp!='p' && atemp!='r') {
		temp = 'r';
		cout << "invalid type, initial temp = " << atemp <<".\n";
		cout << "temp = 'r' assumed.\n";
		}
	else temp = atemp;

	nelem = 0;
	nr = anr;
	nc = anc;
	fr = afr;
	fc = afc;
	name=0;
	lr = fr + nr - 1;
	lc = fc + nc - 1;
	sprintf(wheretxt,"Packed matrix rows");
	if((row=new short*[nr])==0||(rnc=new short[nr])==0||
		(p= new float*[nr])==0)	outofmem(wheretxt);
	row -= fr;
	rnc -= fr;
	p -= fr;
	for(i=fr; i<=lr; i++) {
		rnc[i]=0;
		p[i]=0;
		row[i]=0;
		}
}

Pmatrix::Pmatrix(Pmatrix& a)
{
	short i,j,k;
	char wheretxt[60];

	which = vcount++;
	nr = a.nr;
	nc = a.nc;
	temp = a.temp;
	fr = a.fr;
	fc = a.fc;
	lr = a.lr;
	lc = a.lc;
	name=0;
	nelem = a.nelem;
	sprintf(wheretxt,"Pmatrix Copy");
	if((rnc=new short[nr])==0||(row=new short*[nr])==0||
		(p=new float*[nr])==0) outofmem(wheretxt);
	rnc -=fr;
	row -=fr;
	p -= fr;
	for(i=fr; i<=nr; i++)
		if((rnc[i] = a.rnc[i]) >0 ){
			if((row[i]=new short[a.rnc[i]])==0||(p[i]=new float[a.rnc[i]])==0)
				outofmem(wheretxt);
			for(j=0; j<a.rnc[i]; j++){
				row[i][j] = a.row[i][j];
				p[i][j] = a.p[i][j];
				}
			}
		else {
			row[i]=0;
			p[i]=0;
			}
	}
Pmatrix::Pmatrix(const Matrix& a)
{
	short i,j,k;
	short *rnz;
	float *trow;
	char wheretxt[60];

	sprintf(wheretxt,"Pmatrix <- Matrix conversion");
	if((trow = new float[lc-fc+1]) == 0 ||
		(rnz = new short[lc-fc+1]) == 0) outofmem(wheretxt);
	trow -= a.fc;
	Pmatrix Temp(a.nr, a.nc, 'p', a.fr, a.fc);
	*this = Temp;
	for(i = a.fr; i <= a.lr; i++){
		for(j = a.fc, rnc[i]=0; j <= a.lc; j++)
			if(a.m[i][j]!=0.0)
				rnz[rnc[i]++] = j;
		if((row[i] = new short[rnc[i]]) == 0 ||
			(p[i] = new float[rnc[i]]) == 0) outofmem(wheretxt);
		for(j=0, nelem+=rnc[i]; j<rnc[i]; j++){
			row[i][j] = rnz[j];
			p[i][j] = trow[rnz[j]];
			}
		}
	temp = 'r';
	trow += a.fc;
	delete trow;
	delete rnz;
	}

Pmatrix::Pmatrix() {
nr=nc=fr=lr=fc=lc=0;
nelem=0;
rnc=0;
row=0;
p=0;
}
Pmatrix::~Pmatrix()
{
	#ifdef DBG_DESTRUCT
	printf("Calling Pmatrix destructor.  which = %d seq = %d\n",which,seq);
	#endif
	if(nelem == 0) 
		return; //dummy matrix
	freeh();
	}
void Pmatrix::freeh()
{
	short i;

	if(p != 0){
		for(i = fr; i <= lr; i++){
			if(row[i]!=0) {
				delete row[i];
				row[i] = 0;
				}
			if(p[i]!=0)	{
				delete p[i];
				p[i] = 0;
				}
			}
		delete (p+fr);
		delete (row+fr);
		delete (rnc+fr);
		nelem=0;
		temp=(temp<='q')?'p':'r';
		p = 0;
		}
	}
/* SetAij() 
	SetAij will set row i column j element of packed matrix, if it originally
	is not zero, to value aij.
	It returns OK if succeeded or ERR if failed.
	*/
short Pmatrix::SetAij(short i, short j, float aij)
{
	short k;
	 
	if(i<fr || i > lr || j<fc || j>lc) return (ERR);
	for(k=0; k<rnc[i]; k++)
		if(j == row[i][k]) break;
	if(k==rnc[i]) return(ERR);
	p[i][k] = aij;
	return(OK);
	}
float Pmatrix::operator () (short i, short j)
{
	short rbeg, rend, r;

	if(p==0 || temp=='p' || temp=='r'){
		cerr << "Error: Reference to a deleted or unloaded Pmatrix.\n";
		return 0.0;
		}
	if(i<fr || i>lr || j<fc || j>lc){
		cerr << "illegal matrix index: " << i << "," << j << "." <<"\n";
		cerr << "returning zero. \n";
		return 0.0;
		 }
	 for(r=0; r<rnc[i]; r++) if(j == row[i][r]) return (p[i][r]);
	return 0.0;
}


/* Init() **********************************************************
	Initialize the Pmatrix. 
*/

void Pmatrix::Init()
{
	short i;
	char wheretxt[60];
	sprintf(wheretxt,"Pmatrix initialization");
	if(p==0){
		if((rnc=new short[nr])==0||(row=new short*[nr])==0||(p=new float*[nr])==0)
			 outofmem(wheretxt);
		rnc-=fr;
		row-=fr;
		p-=fr;
		for(i=fr;i<=lr;i++){
			rnc[i]=0;
			row[i]=0;
			p[i]=0;
			}
		}
	else {
		for(i=fr; i<=lr; i++)
			if(rnc[i]!=0){
			rnc[i]=0;
			delete row[i];
			delete p[i];
			row[i]=0;
			p[i]=0;
			}
		}
	nelem =0;
	temp = (temp<='q') ? 'p' : 'r';
	return;
	}

/* ReadP(char *finame, short yr) *********************************
	ReadP pmatrix "pm" for year "yr" from file "fin".
	return OK when succeeds or ERR if fails.

	The only diference between LoadP and ReadP is the latter always 
	calls Init() to initialize the packed matrix while LoadP is assuming 
	all memory needed has been properly allocated.

*/
short Pmatrix::ReadP(char *finame, short yr)
{
	short i;
	pmfid pmh;
	long tos, pmhsize;
	FILE *fin;
	if((fin=fopen(finame, "rb"))==0){
		cerr<<"Error: Can not open input file "<<finame<<".\n";
		return(ERR);
		}
	pmh.ReadIn(fin);

	if(yr< pmh.startyear||yr>pmh.endyear){
		cerr<<"Error: The year " << yr <<  " is out of range.\n";
		return(ERR);
		}
	if(nr!=pmh.nr||nc!=pmh.nc){
		cerr<<"Error: The rows or columns are not match.\n";
		return(ERR);
		}
	// Init allocates memory for rnc, row, and p.
	Init();

	if(fread(&rnc[fr], sizeof(short), nr, fin)!=nr)
		return(ERR);
	for(i=fr; i<=lr; i++) if(rnc[i]!=0){
		if((row[i]=new short[rnc[i]])==0||(p[i]=new float[rnc[i]])==0){
			cerr<<"Error: Out of memory.\n";
			return (ERR);
			}
		if(fread(row[i], sizeof(float), rnc[i], fin)!=rnc[i]) 
			return(ERR);
		}

	nelem = pmh.nelem;
	pmhsize = 80*sizeof(char)+9*sizeof(short)+3*sizeof(long)+sizeof(short*)
		+sizeof(short**)+sizeof(char); // now 119

	tos = pmhsize + sizeof(short)*(nr + nelem) + nelem *
		sizeof(float)*(yr - pmh.startyear);
	fseek(fin, tos, SEEK_SET);

	for(i=fr; i<=lr; i++)
		if(rnc[i]!=0){
			if(fread(p[i], sizeof(float), rnc[i], fin)!=rnc[i]) 
				return(ERR);
			}
	temp=(temp<='q')?'q':'s';
	fclose(fin);  //dsm
	return(OK);
	}

/* LoadP(char *finame, short yr) *********************************
	Load pmatrix "pm" for year "yr" from file "fin".
	return OK when succeeds or ERR if fails
*/
short Pmatrix::LoadP(char *finame, short yr)
{
	short i;
	pmfid pmh;
	long tos;
	FILE *fin;
	long pmhsize=0;

	if((fin=fopen(finame, "rb"))==0){
		cerr<<"Error: Can not open input file "<<finame<<".\n";
		return(ERR);
		}
	pmh.ReadIn(fin);

	if(yr< pmh.startyear||yr>pmh.endyear){
		cerr<<"Error: The year " << yr <<  " is out of range.\n";
		return(ERR);
		}
	if(nr!=pmh.nr||nc!=pmh.nc){
		cerr<<"Error: The rows or columns do not match.\n";
		return(ERR);
		}
	pmhsize = 80*sizeof(char)+9*sizeof(short)+3*sizeof(long)+sizeof(short*)
		+sizeof(short**)+sizeof(char); // now 119
	tos = pmhsize + sizeof(short)*((long)nr + (long)nelem) +
		(long)nelem * sizeof(float)*((long)yr - (long)pmh.startyear);
	fseek(fin, tos, SEEK_SET);

	for(i=fr; i<=lr; i++)
		if(rnc[i]!=0){
			if(fread(p[i], sizeof(float), rnc[i], fin)!=rnc[i])
				return(ERR);
			}
	temp=(temp<='q')?'q':'s';
	fclose(fin);  //dsm
	return(OK);
	}

/* Store(char *finame, short yr) *********************************
	Write pmatrix "pm" for year "yr" to file "fin".
	return OK when succeeds or ERR if fails
*/
short Pmatrix::store(char* finame, short yr)
{
	pmfid pmh1;
	short *frnc;
	FILE *fin;
	short i;
	long tos,pmhsize;
	char wheretxt[60];

	sprintf(wheretxt,"Pmatrix store");
	if((fin=fopen(finame,"rb+"))==0) {
		printf(" Error: Can not open file %s\n", finame);
		return(ERR);
		}
	pmh1.ReadIn(fin);
	if(yr< pmh1.startyear||yr>pmh1.endyear){
		printf("Error: The year %4d is out of range.\n", yr);
		return(ERR);
		}

	if(nr!=pmh1.nr||nc!=pmh1.nc||nelem!=pmh1.nelem) goto not_match;
	if((frnc = new short[nr])==0)
		outofmem(wheretxt);
	fread(frnc, sizeof(short), nr, fin);
	for(i=fr; i<=lr; i++)
		if(rnc[i]!=frnc[i-fr]) goto not_match;
	delete frnc;

	pmhsize = 80*sizeof(char)+9*sizeof(short)+3*sizeof(long)+sizeof(short*)
		+sizeof(short**)+sizeof(char); // now 119
	tos = pmhsize + sizeof(short)*(nr + nelem) + sizeof(float)*
		nelem * (yr - pmh1.startyear);
	fseek(fin, tos, SEEK_SET);
	for(i=fr; i<=lr; i++)
		if(rnc[i]>0)
			fwrite(p[i], 4, rnc[i], fin);
	fclose(fin);
	return(OK);
	not_match:
	printf("Error: Pmatrices do not match. Write failed.\n");
	fclose(fin);
	return(ERR);
	}



void Pmatrix::Display(char *title,short width, short decimalPlaces)
{
	short i, j, nperline;
	char fmt[20];

	if(strlen(title) > 0) printf("\n%s\n", title);

	nperline = 79/width;
	sprintf(fmt,"%s%d.%df", "%",width,decimalPlaces);
	for(i = fr; i <= lr; i++){
		for(j = fc; j <= lc; j++){	
			printf(fmt,this->operator()(i, j));
			if((j - fc % nperline == nperline -1) && j < lc) printf("\n");
			}
		printf("\n");
		}
	}

Pmatrix& Pmatrix::operator = (const Pmatrix& a)
{
	short i,j,kr,kc,kc1,nrmin;
	char wheretxt[60];
	sprintf(wheretxt,"Pmatrix assignment");
	if(nr!=0) {
		if(nr != a.nr || nc != a.nc) 
			cout << "Equating matrices of unequal size.#" << which
			<< "=#" << a.which <<"\n";
		}
	else {
		nr=a.nr;
		nc=a.nc;
		fr=a.fr;
		fc=a.fc;
		lr=a.lr;
		lc=a.lc;
		Init();
		nelem=a.nelem;
		}
	nrmin = nr < a.nr ? nr : a.nr;

	if(nelem!=0) 
		for(i=fr; i<=lr; i++)
			if(rnc[i]!=0){
				rnc[i]=0;
				delete row[i];
				row[i]=0;
				delete p[i];
				p[i]=0;
				}
	nelem = 0;
	if(a.nelem!=0) 					// a is a loaded matrix
		for(i=0; i< nrmin; i++){
			kc = a.rnc[i+a.fr]>nc?nc:a.rnc[i+a.fr];
			if(kc>0)
				if((row[fr+i] = new short[kc]) == 0||
					(p[fr+i] = new float[kc]) == 0) outofmem(wheretxt);
			kr = 0;
			for(j=0;j<kc;j++){
				kc1 = a.row[a.fr+i][j]-a.fc; // why "this" object is not
								  // accessible after this line
				if(kc1 < nc){
					p[fr+i][kr] = a.p[a.fr+i][j];
					row[fr+i][kr++] = fc+kc1;
					}
				}
			rnc[fr+i] = kr;
			nelem += rnc[fr+i];
			}
	temp = (temp <= 'q')?((a.temp=='q'||a.temp=='s')?'q':'p'):
				  ((a.temp=='q'||a.temp=='s')?'s':'r');

	a.freet();
	return(*this);
	}

Pmatrix& Pmatrix::operator = (const Matrix& a)
{ 
	short i,j,kr,kc,kc1,nz,nrmin, ncmin;
	char wheretxt[60];
	sprintf(wheretxt,"Pmatrix assignment from matrix");
	float *trow;
	short *rnz;
	nrmin = nr < a.nr ? nr : a.nr;
	ncmin = nc < a.nc ? nc : a.nc;
	if(nr != a.nr || nc != a.nc) 
//		cout << "Equating matrices of unequal size.#" << which
//		<< "=#" << a.which <<"\n";
	if((trow = new float[ncmin]) == 0 ||
		(rnz = new short[ncmin]) == 0) outofmem(wheretxt);
	trow-=fc;
	if(nelem!=0)
		for(i=fr; i<=lr; i++)
			if(rnc[i]!=0){
				rnc[i]=0;
				delete row[i];
				row[i]=0;
				delete p[i];
				p[i]=0;
				}
	nelem = 0;
	for(i = 0; i <nrmin; i++){
		for(nz = 0, j = 0; j < ncmin; j++)
			if((trow[fc+j] = a.m[a.fr+i][a.fc+j]) != 0.) rnz[nz++] = fc+j;
		if((rnc[fr+i] = nz)>0){
			if((row[fr+i] = new short[nz]) == 0 ||
				(p[fr+i] = new float[nz]) == 0)	outofmem(wheretxt);
			for(j=0, nelem += nz; j<nz; j++) {
				row[fr+i][j] = rnz[j];
				p[fr+i][j] = trow[rnz[j]];
				}
			}
		}
	temp = (temp <= 'q')?((a.temp=='y')?'q':'p'):
				  ((a.temp=='n')?'s':'r');
	a.freet();
	delete rnz;
	delete (trow+fc);
	return(*this);
	}

Pmatrix Pmatrix::operator ~() //Transposition
{
	short i,j,k;
	float flt;
	char wheretxt[60];
	sprintf(wheretxt,"Pmatrix transposition");
	 if(temp != 'q' && temp != 's') {
		cerr << "Pmatrix is not loaded. Transposition failed.\n";
		exit(1);
		}
	Pmatrix Temp(nc,nr,'p',fc,fr);
	for(i=fr; i<=lr; i++)
		for(j=0; j<rnc[i]; j++) Temp.rnc[row[i][j]]++;
	for(i=Temp.fr; i<=Temp.lr; i++)
		if(Temp.rnc[i]>0)
			if((Temp.row[i] = new short[Temp.rnc[i]]) == 0 ||
				(Temp.p[i] = new float[Temp.rnc[i]]) == 0)	outofmem(wheretxt);
	for(i=fc;i<=lc;i++) Temp.rnc[i] = 0; // reset the counter;
	for(i = fr; i<= lr; i++)
		for(j=0; j<rnc[i]; j++){
			Temp.row[row[i][j]][Temp.rnc[row[i][j]]] = i;
			Temp.p[row[i][j]][Temp.rnc[row[i][j]]++] = p[i][j];
			}
	Temp.nelem = nelem;
	Temp.temp = 'q';
	return(Temp);
}

///////////////FRIENDS OF PACKED MATRIX AND MATRIX AND VECTOR  //////////////
Pmatrix operator * (const float a, const Pmatrix& b)
{
	short i,j;
	Pmatrix Temp;
	Temp = b;
	if(b.temp == 'q'){
		for(i=b.fr; i<= b.lr; i++)
			for(j=0; j<b.rnc[i];j++) b.p[i][j] *= a;
		return(Temp);
		}
	for(i=b.fr; i<=b.lr; i++)
		for(j=0; j<b.rnc[i]; j++) Temp.p[i][j] *=a;
	Temp.temp = 'q';
	b.freet();
	return(Temp);
	}
Pmatrix operator / (const Pmatrix& a, const float b)
{
	short i,j;
	float recip;

	recip = 0.;
	if(b != 0.) recip = 1./b;
	else {
		printf("Dividing a matrix by 0. Result is 0.\n");
		exit(1);
		}
	return (recip*a);
	}
/* ==================================================================== */

Vector operator * (const Vector& a, const Pmatrix& b)
{
	short i,j,vcol;

	Vector Temp(a.nr,b.nc,'y',a.fr,b.fc); 
	for(i = 0; i < Temp.nelem; i++){
		Temp.v[i] = 0.;
		}
	if(a.nc != b.nr){
		cout << "vector*matrix dimensions do not match.\n";
		cout << "returning a zero vector.\n";
		}
	else {
		for(i=0; i<a.nelem; i++) // fixed (11/4/94) by dsm, used to be Temp.nelem!
			for(j=0; j<b.rnc[i+b.fr]; j++){
				vcol = b.row[i+b.fr][j] - b.fc;
				Temp.v[vcol] += a.v[i] * b.p[i+b.fr][j];
				}
		}
	a.freet();
	b.freet();
	return(Temp);
	}
/* ==================================================================== */
// % operator is diagonalized pre-multiply, so that each element
//    of the vector is multiplied by the corresponding row 
//    of the matrix.  Vector a is treated as though
//    it were the diagonal elements of a conformable square matrix.
//    This can be useful in converting constant price to current,
//    for example.

Pmatrix operator % (const Vector& a, const Pmatrix& b)
{
	short i,j;
	double sum;

	Pmatrix Temp;
	Temp = b;
	if(a.numelm() != b.rows() ){
		cout << "vector*matrix dimensions do not match.\n";
		cout << "returning a zero Pmatrix.\n";
		return Temp;
		}
	// The following doesn't work under Builder:
	//if(b.temp == 'q'){
	//	for(i=b.fr; i<= b.lr; i++)
	//		for(j=0; j<b.rnc[i];j++) {
	//         b.p[i][j] *= a[i];
	//         }
	//	return(b);
	//	}
	for(i=b.fr; i<=b.lr; i++)
		for(j=0; j<b.rnc[i]; j++) {
			Temp.p[i][j] = Temp.p[i][j] * a.v[i];
			}
	Temp.temp = 'q';
	b.freet();
	return(Temp);

	}
// =====================================================================
/* There had been a Pmatrix*Pmatrix operator declared in MATRIX.H, 
		but it wasn't defined.  This was declared to return Pmatrix.
		Since you don't know which cells will be nonzero in advance,
		returning a Matrix seems more sensible.
*/
Matrix operator * (const Pmatrix& a, const Pmatrix& b) {
short i,j,k,fr,lr,fc,lc,colno;
double sum;

Matrix Temp(a.nr,b.nc,'y',a.fr,b.fc); 
for(i = a.fr; i <= a.lr; i++){
	for(j = b.fc; j <= b.lc; j++)
		Temp.m[i][j] = 0.;
	}
if(a.nc != b.nr){
	cout << "matrix dimensions do not match in * operator.\n";
	cout << "returning a zero matrix.\n";
	}
else {
	for(i = a.fr; i <= a.lr; i++){
		for(j = b.fc; j <= b.lc; j++){
			sum = 0;
			for(k = 0; k < a.rnc[i]; k++) {
				colno = a.row[i][k];
				sum += a.p[i][k]*b(colno,j);
				}
			Temp.m[i][j] = sum;
			}
		}
	}
a.freet();
b.freet();
return(Temp);

}
/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
// diagonalized post mult -> return P-matrix of same configuration
//  This operator follows the logic of the matrix % operators.  Namely,
//   post-multiplication multiplies down the column by the vector, as
//   in converting coefficients to flows.

Pmatrix operator % (const Pmatrix& a, const Vector& b) {
	short i,j,colno;
	// check for conformability : b.nr == a.nc
	Pmatrix Temp;
	Temp = a;
	if(b.rows() != a.nc) {
		printf("Nonconformable Vector in Pmatrix %% operator!\n");
		return Temp;
		}
	//  The following doesn't work under Builder:
	//if(a.temp == 'q'){
	//	for(i=a.fr; i<= a.lr; i++)
	//		for(j=0; j<a.rnc[i];j++) {
	//         colno=a.row[i][j];
	//         a.p[i][j] *= b[colno];
	//         }
	//	return(a);
	//	}
	for(i=a.fr; i<=a.lr; i++)
		for(j=0; j<a.rnc[i]; j++) {
			colno=a.row[i][j];
			Temp.p[i][j] *= b.v[colno];
			}
	Temp.temp = 'q';
	a.freet();
	return(Temp);

}
/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
/* ebemul with packed matrix row.  The arguments are a Pmatrix, the number of a
	row	in the Pmatrix, and a Vector.  The return value is the column Vector 
	by element-by-element multiplication of the Pmatrix row and the Vector.
	The returned vector has the same dimensions as the argument Vector.
	*/
Vector ebemul(const Pmatrix& a, short i, const Vector& b)
{
	short j, j1;
	short k;  // k is minimum of b.nelem and number of nonzeros in row i;

	Vector Temp(b.nr,b.nc,'y');
	k = b.nelem > a.rnc[i] ? a.rnc[i] : b.nelem;
	for(j = 0; j< k; j++)
		if(a.row[i][j]<=b.nelem){ //the col'n number of the nonzero < b.nelem?
			j1 = a.row[i][j] - a.fc; 
			Temp.v[j1] = a.p[i][j]*b.v[j1];
			}
	return (Temp);
	}
Vector operator * (const Pmatrix& a, const Vector& b)
{
	short i,j;
	double sum,x,y,z;
	short n,n0,n1,n2;
	short rt[100];

	Vector Temp(a.nr,b.nc,'y',a.fr,b.fc); 
	for(i = 0; i < Temp.nelem; i++){
		Temp.v[i] = 0.;
		}
	if(a.nc != b.nr){
		cout << "matrix*vector dimensions do not match.\n";
		cout << "returning a zero vector.\n";
		}
	else {
		for(i = 0; i < Temp.nelem; i++){
			sum = 0.;
			n = a.rnc[a.fr+i];
			for(j=0;j<n;j++)
				rt[j] = a.row[a.fr+i][j];
			for(j = 0; j < n; j++) {
				if(i==95 && j==3) {
					n1=a.fr+i;
					x = a.p[a.fr+i][j];
					n2 = a.row[a.fr+i][j];
					y = b.v[n2-a.fc];
					}
				sum += a.p[a.fr+i][j]*b.v[a.row[a.fr+i][j]-a.fc];
				}
			Temp.v[i] = sum;
			}
		}
	a.freet();
	b.freet();
	return(Temp);
	}

Vector operator / (const Pmatrix& a, const Vector& b)
{
	short i,j,k,n;
	double sum;

	Vector Temp(a.nc,b.nc,'y',a.fc,b.fc);

	if(a.nr != b.nr){
		cerr << "Matrices do not have same number of rows in / operator.\n";
		cerr <<  "Using the smaller number.\n";
		}
	n = min(a.nr,b.nr);
	for (i = a.fc; i <= a.lc; i++){
		sum = 0;
		for(k = 0; k < n; k++)
			sum += a(a.fr+k,i)*b.v[k];
		Temp.v[i-a.fc] = sum;
		}
	a.freet();
	b.freet();
	return(Temp);
	}
Vector operator / (const Vector& a, const Pmatrix& b)
{
	short i,j,k,n;
	double sum;

	Vector Temp(a.nc,a.nr,'y',a.fc,b.fr);
	for(i=0; i<a.nelem;i++)
		Temp.v[i] = a.v[i];
	a.freet();
	return(Temp*b);
	}

Vector colsum(const Pmatrix& A){
	short i,j, fr,lr,fc,lc,nc;
	double sum;

	fr = A.firstrow();
	lr = A.lastrow();
	fc = A.firstcolumn();
	lc = A.lastcolumn();
	nc  = A.columns();
	Vector Temp(nc,1,'y',fc,1);
	for(i = fc; i < lc; i++)
		Temp[i] = 0.;
	for(j = fc; j <= lc; j++){
		sum = 0.;
		for(i = fr; i <= lr; i++)
			sum += A(i,j);
		Temp[j] = sum;
		}
	
	return(Temp);
	}
Vector rowsum(const Pmatrix& A){
	short i,j, fr,lr,fc,lc,nr;
	double sum;

	fr = A.firstrow();
	lr = A.lastrow();
	fc = A.firstcolumn();
	lc = A.lastcolumn();
	nr = A.rows();

	Vector Temp(nr,1,'y',fr,1);
	for(j = fr; j < lr; j++)
		Temp[j] = 0.;
	for(i = fr; i <= lr; i++){
		sum = 0.;
		for(j = fc; j <= lc; j++)
			sum += A(i,j);
		Temp[i] = sum;
		}
	
	return(Temp);
	}

/* ====================================================================*/
// Read in the packed matrix header to file fpo, which must already
//   be open.
short pmfid::ReadIn(FILE*fpo)
{
	short err=0;
	size_t count=0;
	fseek(fpo, 0l, SEEK_SET);
	count+=fread(name,sizeof(char),80,fpo);
	count+=fread(&startyear,sizeof(short),1,fpo);
	count+=fread(&endyear,sizeof(short),1,fpo);
	count+=fread(&np,sizeof(short),1,fpo);
	count+=fread(&nr,sizeof(short),1,fpo);
	count+=fread(&nc,sizeof(short),1,fpo);
	count+=fread(&fr,sizeof(short),1,fpo);
	count+=fread(&fc,sizeof(short),1,fpo);
	count+=fread(&lr,sizeof(short),1,fpo);
	count+=fread(&lc,sizeof(short),1,fpo);
	count+=fread(&nelem,sizeof(long),1,fpo);
	count+=fread(&BytesPerYear, sizeof(long),1,fpo);
	count+=fread(&loaded,sizeof(char),1,fpo);
	// The next couple we don't really need, but we read to make the
	//   positioning consistent with DOS
	count+=fread(&rnc,sizeof(short*),1,fpo);
	rnc=0;
	count+=fread(&row,sizeof(short**),1,fpo);
	row=0;
	count+=fread(&FirstMat,sizeof(long),1,fpo);
	if(count!=95)
		return (-1);
	else return 0;
}
/* ====================================================================*/
// Write out the packed matrix header to file fpo, which must already
//   be open.
short pmfid::WriteOut(FILE*fpo)
{
	short err=0;
	size_t count=0;
	fseek(fpo, 0l, SEEK_SET);
	count+=fwrite(name,sizeof(char),80,fpo);
	count+=fwrite(&startyear,sizeof(short),1,fpo);
	count+=fwrite(&endyear,sizeof(short),1,fpo);
	count+=fwrite(&np,sizeof(short),1,fpo);
	count+=fwrite(&nr,sizeof(short),1,fpo);
	count+=fwrite(&nc,sizeof(short),1,fpo);
	count+=fwrite(&fr,sizeof(short),1,fpo);
	count+=fwrite(&fc,sizeof(short),1,fpo);
	count+=fwrite(&lr,sizeof(short),1,fpo);
	count+=fwrite(&lc,sizeof(short),1,fpo);
	count+=fwrite(&nelem,sizeof(long),1,fpo);
	count+=fwrite(&BytesPerYear, sizeof(long),1,fpo);
	count+=fwrite(&loaded,sizeof(char),1,fpo);
	// The next couple we don't really need, but we read to make the
	//   positioning consistent with DOS
	count+=fread(&rnc,sizeof(short*),1,fpo);
	count+=fread(&row,sizeof(short**),1,fpo);
	FirstMat=80*sizeof(char)+9*sizeof(short)+3*sizeof(long)+sizeof(char);
	count+=fwrite(&FirstMat,sizeof(long),1,fpo);;
	
	if(count!=95)
		return (-1);
	else return 0;
}

