// RESECTOR.CPP -- Methods for ReSector Class

#include <stdio.h>  
#include <conio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <new.h>
#include <malloc.h>
#include <string.h>

#include "resector.h"

short **imatrix(short,short,short,short);
char *sgetint(char*, short, short*);
free_imatrix(short **ConCord,short nrl,short nrh,short ncl);

/* The following test program exercises some of the functions of the
	ReSector class.  To test this function, make sure that the following
	line "#define TEST" is not commented out, and make -fresector.
	To turn the class back in to one that can be linked with others,
	put the comment back on again.  For column numbers, we refer to
	the number of sectors in the aggregation scheme.  This causes no
	problems as long as two schemes do not have the same number of 
	sectors.
*/

//#define TEST

#if defined(TEST)

#include "message.h"
#include "messinit.inc"

#define NILIAD 320
#define NDOM   480

void main(void)  {

short err,i,n,l[40];
short idom,iibm,count;

ReSector rs("defn82.key");  // declare ReSector object rs

err = rs.FormAgg(NDOM,NILIAD); // DOM to Iliad
if(err<0) {
	printf("Error in obtaining aggregation from DOM to Iliad.\n");
	exit(1);
	}
err = rs.FormAgg(NILIAD,NDOM);  // Iliad to DOM.
if(err<0) {
	printf("Error in obtaining aggregation from Iliad to DOM.\n");
	exit(1);
	}

printf("\n\nDOM->ILiad concordance:\n");

printf( "Number of splits: %d\n",rs.GetNSplit(NDOM,NILIAD) );
printf("Number of Iliad sectors for DOM 453: %d\n",
	n=rs.GetNcon(NDOM,NILIAD,453) );
rs.GetList(NDOM,NILIAD,453,l);
printf("List of Iliad sectors in DOM 453:\n");
for(i=0;i<n;i++) 
   printf("%4d",l[i]);
printf("\n");

printf("Number of DOM sectors with 1 correspondence  : %4d\n",
	rs.GetNSingles(NDOM,NILIAD) );
printf("Number of DOM sectors with any correspondence: %4d\n",
	rs.GetNAny(NDOM,NILIAD) );
printf("\n");
//=============================
printf("\n\nIliad->DOM concordance:\n");
printf( "Number of splits: %d\n",rs.GetNSplit(NILIAD,NDOM) );
printf("Number of DOM sectors for Iliad 200: %d\n",
	n=rs.GetNcon(NILIAD,NDOM,200) );
rs.GetList(NILIAD,NDOM,200,l);
printf("List of DOM sectors in Iliad 200:\n");
for(i=0;i<n;i++) 
   printf("%4d",l[i]);
printf("\n");

printf("Number of Iliad sectors with 1 correspondence  : %4d\n",
	rs.GetNSingles(NILIAD,NDOM) );
printf("Number of Iliad sectors with any correspondence: %4d\n",
    rs.GetNAny(NILIAD,NDOM) );
printf("\nOne to One's:\n");  // one to one means "only to only"
count=0;
printf(" Iliad DOM\n");
for(i=1;i<=NILIAD;i++) {
   if(rs.GetNcon(NILIAD,NDOM,i)==1) {   
      idom=rs.GetFirst(NILIAD,NDOM,i);
      if(rs.GetNcon(NILIAD,NDOM,idom)==1) {
         count++;
         printf("%4d::%4d  %c",i,idom,(count%6)?' ':'\n');
         }
      }
   }
printf("\n There are %d one to one correspondences between DOM & Iliad.\n",count);
}
#endif

/*======================================================================== */
/* Constructor, read sector definitions from file.  This constructor can
	read and organize the standard DEFN82.SIC file, but it is actually 
	much more general.  The only change from the usual DEFN82.SIC file 
	is that each column must contain an integer specifying how many
*/

#define MAXKEYLEN 350

ReSector::ReSector(const char *s, short NumColumns, short MaxSecs){
FILE *key;
char *cp,line[MAXKEYLEN];
short linecount=0,i,j,idom,iibm,qli[20];

NKey=0;
NColsUsed = NumColumns;
MaxSecsUsed = MaxSecs;
// Initialize the Sector Aggregation Concordance lists, and the split lists.
for(i=0;i<MAXCOL;i++) {
	Vecs[i]=0;
	for(j=0;j<MAXCOL;j++) {
		Secs[i][j]=0;
		SplitList[i][j]=0;
		MemAll[i][j]=0;
		NSplit[i][j]=0;
		}
	}

if((key=fopen(s,"r"))==NULL)  {
   printf("Unable to open %s\n",s);
	Initialized = 0;
   return;
   }

ConCord = imatrix(1,MaxSecsUsed,1,NColsUsed);
			
// Read through definition file, putting it into memory (ConCord);

cp=fgets(line,MAXKEYLEN-1,key);
// Read NSecs for each column on first line:
for(i=0;i<NumColumns;i++) 
	cp=sgetint(&line[i*5],5,&NSecs[i]);

while((cp=fgets(line,MAXKEYLEN-1,key))!=NULL) {
	linecount++;
	short inum;
	if(linecount>MaxSecs) break;
	for(i=1;i<=NumColumns;i++) {
		cp=sgetint(&line[(i-1)*5],5,&inum);
		ConCord[linecount][i]=inum;  // column starts at 1
		}
   } 
if(linecount>MaxSecs) {
	printf("Too many lines in aggregation file: %d.  MaxSecs = %d.\n",
			linecount,MaxSecs);
	return;
	}
NKey = linecount;
	
fclose(key);
Initialized = 1;
}
/*======================================================================== */
/*  Destructor, deletes Vecs, Secs, SplitList and ConCord.
*/
ReSector::~ReSector() {
short i,j;
//printf("Starting ReSector destructor.\n");
for(i=0;i<MAXCOL;i++) {
	// free Vecs
	if(Vecs[i]) {
		//printf("Destructing Vecs[%d]\n",i);
		delete[] Vecs[i];
		}
	for(j=0;j<MAXCOL;j++) {
		if(!MemAll[i][j])
			continue;
		//printf("Destructing i: %d  j: %d\n",i,j);
		// free Secs
		if(Secs[i][j]!=NULL) {
			//printf("Destructing Secs i: %d  j: %d\n",i,j);
			delete[] Secs[i][j];
			}
		// free SplitList
		if(SplitList[i][j]!=NULL) {
			//printf("Destructing SplitList i: %d  j: %d\n",i,j);
			delete[] SplitList[i][j];
			}
		//printf("Finished Destructing i: %d j: %d\n",i,j);
		}
	}
free_imatrix(ConCord,1,MaxSecsUsed,1);
//printf("Finishing ReSector destructor.\n");
}
/*======================================================================== */
/* GetColNo -- This is the trick for referring to aggregation schemes
	by the number of sectors at the top of the file.  This function returns
	(-1) when there is no aggregation scheme with that number of sectors.
*/
short ReSector::GetColNo(short NSec){
for(short i=0;i<NColsUsed;i++)
	if(NSecs[i]==NSec)
		return(i);
return(-1);
}
/*======================================================================== */
/* FORMAGG -- Once the "data" of the aggregation schemes have been read
	in from the constructor, this function performs the real initialization
	work.  This function performs work that used to be done by the old 
	constructor, which was more specific.  It sets up the concordance lists
	and split lists in whichever of the NumColumn * NumColumn directions
	is specified.
   
	For example, to get the Iliad to LIFT initialized, where Iliad was 
	column 4 and LIFT column 6, one would call:
		err=FormAgg(4,6);
	Alternatively, maybe it would be better to handle this by sector
	numberings (more intuitive), such as:
		err=FormAgg(320,85);
*/

short ReSector::FormAgg(short Nfrom, short Nto) {
short from,to;
short Jfrom, Jto;
short i;

// Get column positions in ConCord array:

if((from = GetColNo(Nfrom))==-1) {
	printf("No such aggregation scheme: %d\n",Nfrom);
	return (-1);
	}
if((to = GetColNo(Nto)) == -1) {
	printf("No such aggregation scheme: %d\n",Nto);
	return (-1);
	}

// Set up space for vectors that may be used.
if(!Vecs[from])
	Vecs[from] = new float[Nfrom];
if(!Vecs[to])
	Vecs[to] = new float[Nto];

// Initialize the appropriate elements of Secs and SplitList.
Secs[from][to] = new SecCon[Nfrom];

// Zero it out...
for(i=0;i<Nfrom;i++)  {
	Secs[from][to][i].ncon=0;
	Secs[from][to][i].conlist=NULL;
	}

SplitList[from][to] = new short[Nto];  // maybe this is overkill, but be safe!
MemAll[from][to] = 1;

// Perform the "AddNode" function, for each row of the matrix.
for(i=1;i<=NKey;i++) {
	Jfrom = ConCord[i][from+1];
	Jto = ConCord[i][to+1];
	if(Jfrom<=0 || Jfrom > Nfrom || Jto <= 0 || Jto > Nto)
		continue;
	AddNode(from,to,Jfrom,Jto);
	}

// Determine the number of splits for this combination.

NSplit[from][to]=0;
for(i=0;i<Nfrom;i++)
	if(Secs[from][to][i].ncon>1)
		SplitList[from][to][NSplit[from][to]++]=i+1;  // sectors, starting at 1, not 0
return 0;  // OK.
}
/*======================================================================== */
/*	AddNode -- This adds a node in the Secs vector, which provides a list
	of sectors from the "to" aggregation corresponding to the sector from
	the "from" aggregation.  This is a more general implementation than
	the old version of ReSector.
	This function is called with from and to starting at 0.

*/
short ReSector::AddNode(short from, short to, short Jfrom, short Jto)  {
short i,n;
nod *np,*nplast;

if((n=Secs[from][to][Jfrom-1].ncon)==0) {
	Secs[from][to][Jfrom-1].conlist=new nod;
	Secs[from][to][Jfrom-1].conlist->conc=Jto;
	Secs[from][to][Jfrom-1].conlist->next=NULL;   // end of list
	Secs[from][to][Jfrom-1].ncon++;
	return 0;  // 0 return code means inserted.
	}
else {
	np=Secs[from][to][Jfrom-1].conlist;
	for(i=0;i<n;i++) {
		// if already there, return cardinal index (start at 1)
		if(Jto==np->conc)
			return(i+1);
		nplast=np;
		np=np->next;
		}
	np=new nod;
	nplast->next=np;
	np->conc=Jto;
	np->next=NULL;
	Secs[from][to][Jfrom-1].ncon++;
	return 0;
	}
}
/*======================================================================== */
/* IsNode -- Check if a node is there.  0 based.
*/
short ReSector::IsNode(short from, short to, short Jfrom, short Jto)  { // check on presence of node
nod *np;
if(Secs[from][to][Jfrom-1].ncon==0) return 0;
np=Secs[from][to][Jfrom-1].conlist;
while(np)  {
	if(np->conc==Jto) return 1;
	np=np->next;
	}
return 0;  // not found
}
/*======================================================================== */
/* GetList -- Get a list of sectors corresponding to sector sec -- From
	and to in this routine start at 1.
*/
short ReSector::GetList(short Nfrom, short Nto, short sec, short *list)  {
short i,n;
nod *np;

short from,to;
if((from = GetColNo(Nfrom))==-1) {
	printf("No such aggregation scheme: %d\n",Nfrom);
	return (-1);
	}
if((to = GetColNo(Nto)) == -1) {
	printf("No such aggregation scheme: %d\n",Nto);
	return (-1);
	}
if(!(Secs[from][to]))
	return(-2);
np=Secs[from][to][sec-1].conlist;
if((n=Secs[from][to][sec-1].ncon)==0)
	return(-1);
for(i=0;i<n && np;i++) {
	list[i]=np->conc;
	np=np->next;
	}
return(i);
}
/*======================================================================== */
short ReSector::GetNSplit(short Nfrom, short Nto) {
short from,to;
if((from = GetColNo(Nfrom))==-1) {
	printf("No such aggregation scheme: %d\n",Nfrom);
	return (-1);
	}
if((to = GetColNo(Nto)) == -1) {
	printf("No such aggregation scheme: %d\n",Nto);
	return (-1);
	}
if(!(Secs[from][to]))
	return(-2);
return NSplit[from][to];
}
/*======================================================================== */
/* Gets the number of single sector matches - 1 based.
*/
short ReSector::GetNSingles(short Nfrom, short Nto) {
short count=0;
short from,to;

if((from = GetColNo(Nfrom))==-1) {
	printf("No such aggregation scheme: %d\n",Nfrom);
	return (-1);
	}
if((to = GetColNo(Nto)) == -1) {
	printf("No such aggregation scheme: %d\n",Nto);
	return (-1);
	}
if(!(Secs[from][to]))
	return(-2);
for(short i=1;i<=Nfrom;i++)
	if(GetNcon(Nfrom,Nto,i)==1)
		count++;
return(count);
}
/*======================================================================== */
/*	 Get number of concordance matches.  -> 1 based
*/
short ReSector::GetNcon(short Nfrom, short Nto, short ind) {
short from,to;
if((from = GetColNo(Nfrom))==-1) {
	printf("No such aggregation scheme: %d\n",Nfrom);
	return (-1);
	}
if((to = GetColNo(Nto)) == -1) {
	printf("No such aggregation scheme: %d\n",Nto);
	return (-1);
	}
if(!(Secs[from][to]))
	return(-2);
return Secs[from][to][ind-1].ncon;
}
/*======================================================================== */
/* GetNAny -- Return the number of any concordance matches -> 1 based */
short ReSector::GetNAny(short Nfrom, short Nto) {
short count=0;
short from,to;
if((from = GetColNo(Nfrom))==-1) {
	printf("No such aggregation scheme: %d\n",Nfrom);
	return (-1);
	}
if((to = GetColNo(Nto)) == -1) {
	printf("No such aggregation scheme: %d\n",Nto);
	return (-1);
	}
if(!(Secs[from][to]))
	return(-2);
for(short i=1;i<=Nfrom;i++)
	if(GetNcon(Nfrom,Nto,i)>=1)
		count++;
return(count);
}
/*======================================================================== */
short ReSector::GetFirst(short Nfrom, short Nto, short ind)  {
short from,to;
if((from = GetColNo(Nfrom))==-1) {
	printf("No such aggregation scheme: %d\n",Nfrom);
	return (-1);
	}
if((to = GetColNo(Nto)) == -1) {
	printf("No such aggregation scheme: %d\n",Nto);
	return (-1);
	}
if(!(Secs[from][to]))
	return(-2);
if(ind<=0 || ind>Nfrom)
	return(-3);
if(GetNcon(Nfrom,Nto,ind)==0)
	return(-4);
return Secs[from][to][ind-1].conlist->conc;
}
/*======================================================================== */
SecCon::~SecCon()  {  // thanks, Tim
nod *np;
np=conlist;
while(np) {
   np=conlist->next;
   delete conlist;
   conlist=np;
   }
}
      



