/*  RESECTOR -- Class that embodies DEFN82.IBM, and other types of 
	aggregation keys of this general type.  To be a legal key file, 
	the file must have numbers in columns of width 5, and each column
	must be headed by a number showing the highest legal sector number
	in that aggregation scheme.  The program reads in NUMCOL columns
	(now 6) by default, but can be told to read in more or less by
	supplying this argument in the constructor. 

	Here are the first couple of lines from the file DEFN82.KEY, which
	was the first aggregation file tested with this class:
------------------------------------------------------------------------
  575  571  480  320  432   85  BEA82   
    1    1    1    1    1    1  10100   Dairy farm products                                            
    2    2    2    2    2    1  10200   poultry and eggs                                               
    3    3    3    3    3    1  10301   Meat animals                                                   
    4    4    3    3    3    1  10302   Miscellaneous livestock-horses,bees,honey,etc                  
    5    5    4    4    4    1  20100   Cotton                                                         
    6    6    5    5    5    1  20201   Food grains: wheat,rye,rice,buckwheat                          
    7    7    5    5    6    1  20202   Feed grains: corn,oats,barley,hay,sorghum                      
    8    8    5    5    6    1  20203   Grass seeds                                                    
    9    9    6    6    7    1  20300   Tobacco                                                        
   10   10    7    7    8    1  20401   Fruits                                                         
   11   11    7    7    8    1  20402   Tree nuts                                                      
   12   12    7    7    9    1  20501   Vegetables                                                     
------------------------------------------------------------------------

*/

#if(!defined(RESECTOR_H))
#define RESECTOR_H  1

// This simple structure represents a node in a linked list.
struct nod {
   short conc;
   nod *next;
   };

// This is a Sector Concordance string, implemented as a linked list.
struct SecCon {
   short ncon;
   nod *conlist;
   ~SecCon(); 
   };

class Vector;
class Matrix;
class Pmatrix;

class ReSector  {
 protected:
	/* Constants:
		MAXCOL - Maximum number of aggregations to read in.
		NUMCOL - Default number of aggregations.
		MAXSEC - Maximum number of lines to read in the file.
	*/
	enum constants {MAXCOL=10,NUMCOL=6,MAXSEC=700};
	short NSecs[MAXCOL];     // Number of sectors for each scheme.
	short NSplit[MAXCOL][MAXCOL];      // Current number of splits.
	short MemAll[MAXCOL][MAXCOL];      // Track memory allocation.
	short *SplitList[MAXCOL][MAXCOL];      //  List of splits.
	struct SecCon *Secs[MAXCOL][MAXCOL];  // Sector concordance linked lists.
	float *Vecs[MAXCOL];			 // Vectors for aggregation/disaggregation
	short **ConCord;      // Matrix of concordance data, up to MAXSEC by MAXCOL.
	short Initialized, NKey, NColsUsed, MaxSecsUsed;
 public:
	ReSector(const char *s, short NumColumns=NUMCOL, short MaxSecs=MAXSEC);
	~ReSector();

	short IsInitialized(void) {return Initialized;}
	// Routines with Nfrom, Nto are based on no. of sectors.
	short FormAgg(short Nfrom, short Nto);
	short GetList(short Nfrom, short Nto, short sec, short *list);
	short GetNSplit(short Nfrom, short Nto);
	short GetNSingles(short Nfrom, short Nto);
	short GetNcon(short Nfrom, short Nto, short ind);
	short GetNAny(short Nfrom, short Nto);
	short GetFirst(short Nfrom, short Nto, short ind);	

	short GetColNo(short NSec);
	// routines with from, to are 0 based.
	short AddNode(short from, short to, short Jfrom, short Jto);
	short IsNode(short from, short to, short Jfrom, short Jto);

	// Higher level routines:
	short AggVector(short Nfrom, short Nto, Vector& in, Vector& out, 
		Vector& SplitVector);
	short AggMatRows(short Nfrom, short Nto, Matrix& in, Matrix& out,
		Vector& SplitVector);
	short AggMatRows(short Nfrom, short Nto, Pmatrix& in, Matrix& out,
		Vector& SplitVector);
   short CrossAggMatRows(short Nfrom, short Nto, short Nbetween, 
      Matrix& in, Matrix& out, Vector& SplitVector);
	short CrossAggVector(short Nfrom, short Nto, short Nbetween, 
      Vector& in, Vector& out, Vector& SplitVector);
		
	short CrossAggVectorD(short Nfrom, short Nto, short Nbetween, 
      Vector& in, Vector& out, Vector& SplitVector,
      Vector& SplitDefault); // chk for miss val, use default.
   short doCrossAggVec(short Nfrom, short Nto, short Nbetween, 
      Vector& in, Vector& out, Vector& SplitVector, Vector& SplitDefault,
      short WithDefault=0);

};
#endif  // End class definitions for ReSector class.
