/* TSERIES.CPP -- Functions for Tseries, MacFixBank, and MacFixHome. */

// Use macro variable capabilities of Interdyme:
#include "dymesys.h"

short IsMissing(float x);
#if defined(INCL_MACRO)

//======================================================================
// Determine if a fix will be applied to this variable in the current 
//   year.
short Tseries::IsFixed(void) {
short i;
return mfh.IsFixed(seq);
}
//======================================================================

short MacroFixHome::IsFixed(short seq) {
short i;

for(i=0;i<nfix;i++) {
	if(MFixVec[i]->MacroNumber==seq) {
		switch (MFixVec[i]->FixType) {
			case 'o': // all these cases are treated as actuals. ("oigs")
			case 'i':
			case 'g':
			case 's':
			// modifier fixes
			case 'm':
			case 'c':
			// dynamic growth and index fixes
			case 'n':
			case 'w':
			case 't':
			#ifdef INCL_EQFIX
			case 'e':  // equation fix 
			case 'f':  // follow fix
			case 'h':  // share fix
			#endif
			if(!IsMissing(MFixVec[i]->FixValue))
				return 1;
			else
				continue;
			default:
				continue;
			}  // end cases
		} // end if
	} // end loop i
return 0;

}
//======================================================================
// This function assumes that GBank gb has been defined globally (in dyme.inc
//    for example) and assigned properly.  It also assumes that the main
//    simulation VamFile has been opened as vf[0], also global.
//    It reads the Tseries in from the G forecast bank, dyme.bnk.
void Tseries::in(char *name)
{
	short i, spread,err;
	tsname=0;
	short endd, baseyr;
	endd = (short)vf[0].Getenddate();
	baseyr = gb.GetBaseYr();
	spread = endd - baseyr + 1 ;
	if(( ts = new float[spread]) == 0){
		cerr << "Not enough memory for " << name << ".\n";
		exit(1);
		}
	if((i = gb.findname(name)) == -1){
		printf("Cannot find %s in the time series bank.\n",name);
		seq = -1;
		for(i = 0; i < spread; i++) ts[i] = -0.000001;
		} 
	else{
		tsname = new char[strlen(name)+1];
		strcpy(tsname,name);
		seq = i;
		Tsptr[i] = this;
		if(gb.getser(tsname,ts,spread)==-1) {
			printf("Cannot read %s\n",name);
			}
		}
	}
/*  out() ********************************************************/
/*  Write the macro Tseries out to the G forecast bank, dyme.bnk */
void Tseries::out(GBank& gb)
{
	long posit;
	short spread,err;
	char cnpy;

	spread = (short)vf[0].Getenddate() - (short)gb.GetBaseYr() + 1 ;
	gb.Setnopy(1); // set default frequency to annual (^_^)
	if(gb.addser(tsname,ts,'y',spread)==-1)  {
		printf("Cannot write to G bank series %d.\n",seq);
		}
}
/* operator [] for Tseries *****************************************/
/* subscript operator - Determine the position of t, t-1, etc, from
	the base year of the G forecast bank, dyme.bnk
*/
float& Tseries::operator [](short t)
{
	short sd;
	sd = (short)(gb.GetBaseYr() +.00001); //startdate is a float 
	sd = t - sd;
	if(sd >= 0)     return(ts[sd]);
	else return(ts[0]);
	}
// ==========================================================================
// up - As a default, update the Tseries so that it has the same value as
//   last year.
void Tseries::up()
{
	short i;
	i = t - gb.GetBaseYr();
	if(fabs(ts[i] + .000001) < .0000001) ts[i] = ts[i-1];
	}
// ==========================================================================
//  A destructor is necessary for Tseries because of ts and tsname.
Tseries::~Tseries()  {
	if(ts) 
		delete[] ts;
	if(tsname) 
		delete[] tsname;
}
// ==========================================================================
/* cum(output, input, spill_rate) -- the cum function for run *********/
double cum(Tseries& y, float x, float s)
{
	y[t] = y[t-1]*(1. - s) + x;
	return(y[t]);
	}
// ==========================================================================
double peak(Tseries& y, float x, float s)
{
	float floor;

	floor = y[t-1]*(1.-s);
	y[t] = x > floor? x : floor;
	return(y[t]);
	} 
// ==========================================================================
double pct(Tseries& y)
{
	double d;
	if(IsMissing(y[t-1])||IsMissing(y[t]))
		d=0;
	else 
		d=(fabs(y[t-1])<1.e-6)?0:((y[t]-y[t-1])/y[t-1]*100.);
	return(d);
	} 
// ==========================================================================
double pos(float x)
{
	if(x>0) return(x);
	return(0.);
	}
double sq(float x) {
return x*x;
}
/* storets() -- store all of the macro time series */
void storets(void){
	short i,ns_ws;
	ns_ws = gb.Getnseries();
	for(i = 0; i < ns_ws; i++)
		if(Tsptr[i] != 0) Tsptr[i]->out(gb);
	}
/* upets() -- make this year's value equal to last year's */
void upets(void){
	short i,ns_ws;
	ns_ws = gb.Getnseries();
	for(i = 0; i < ns_ws; i++)
		if(Tsptr[i] != 0) Tsptr[i]->up();
	}
/////////////////////////////////////////////////////////////////////////////
// ===========================================================================
//   MacroFix functions follow:
// ===========================================================================

/*==========================================================================*/
short bnkHeader::WriteOut(FILE* fp) {
size_t count = 0;
count += fwrite(&jahra,sizeof(short),1,fp);
count += fwrite(&stper,sizeof(short),1,fp);
count += fwrite(&nobs,sizeof(short),1,fp);
count += fwrite(bnktitle,sizeof(char),80,fp);
if(count != 83)
	return -1;
else
	return 0;
}
/*==========================================================================*/
short bnkHeader::ReadIn(FILE* fp) {
size_t count = 0;
count += fread(&jahra,sizeof(short),1,fp);
count += fread(&stper,sizeof(short),1,fp);
count += fread(&nobs,sizeof(short),1,fp);
count += fread(bnktitle,sizeof(char),80,fp);
if(count != 83)
	return -1;
else
	return 0;
}
/*==========================================================================*/
short MacFix::ReadIn(FILE *fp) {
size_t count=0;
count += fread(&command,sizeof(char),1,fp);
count += fread(Mname,sizeof(char),15,fp);
count += fread(&startfix,sizeof(short),1,fp);
count += fread(&endfix,sizeof(short),1,fp);
count += fread(&funlen,sizeof(short),1,fp);
count += fread(&funpos,sizeof(long),1,fp);
count += fread(&function,sizeof(char*),1,fp);
function=0;
if(count != 21)
	return(-1);
else return 0;
}
/*==========================================================================*/
short MacFix::WriteOut(FILE *fp) {
size_t count=0;
count += fwrite(&command,sizeof(char),1,fp);
count += fwrite(Mname,sizeof(char),15,fp);
count += fwrite(&startfix,sizeof(short),1,fp);
count += fwrite(&endfix,sizeof(short),1,fp);
count += fwrite(&funlen,sizeof(short),1,fp);
count += fwrite(&funpos,sizeof(long),1,fp);
count += fwrite(&function,sizeof(char*),1,fp);
if(count != 21)
	return(-1);
else return 0;
}
/*==========================================================================*/



MacroFix::MacroFix(char *MName) {

	MacroName=new char[strlen(MName)+1];
	strcpy(MacroName,MName);
	FixType=0;
	RhoStart=0;
	Rho=0;
	FixValue=0;
	Error=0;
	funlen=0;
	funpos=0;
	function=0;
	}
// ===========================================================================
// Copy constructor for MacroFix:
MacroFix::MacroFix(MacroFix& mf) {
	MacroName=new char[strlen(mf.MacroName)+1];
	strcpy(MacroName,mf.MacroName);
	FixNumber=mf.FixNumber;
	MacroNumber=mf.MacroNumber;
	FixType=mf.FixType;
	RhoStart=mf.RhoStart;
	Rho=mf.Rho;
	Error=mf.Error;
	FixValue=mf.FixValue;
	#if defined(INCL_EQFIX)
	funlen=mf.funlen;
	funpos=mf.funpos;
	if(mf.funlen>0) {
		function = new char[funlen];
		strcpy(function,mf.function);
		}
	#endif
	}

// ===========================================================================
MacroFix::~MacroFix(void) {
	if(MacroName) 
		delete[] MacroName;
	#if defined(INCL_EQFIX)
	if(function)
		delete function;
	#endif
}
/*==========================================================================*/
MacFix::MacFix(MacFix& m)  { // copy constructor

command=m.command;
strcpy(Mname,m.Mname);
startfix=m.startfix;
endfix=m.endfix;
	#if defined(INCL_EQFIX)
funlen=m.funlen;
funpos=m.funpos;
if(funlen>0) {
	function=new char[funlen];
	strcpy(function,m.function);
	}
	#endif
}
/*==========================================================================*/
MacFix& MacFix::operator=(const MacFix& m) {
command=m.command;
strcpy(Mname,m.Mname);
startfix=m.startfix;
endfix=m.endfix;
#if defined(INCL_EQFIX)
funlen=m.funlen;
funpos=m.funpos;
if(funlen>0) {
	function=new char[funlen];
	strcpy(function,m.function);
	}
#endif
return *this;
}
// ===========================================================================
// NameToFix separates "fixname" like trffrn:c into macroname and fixtype.
short NameToFix(char *fixname, char *mname, char& fixtype) {
	char *cp;
	if((cp=strchr(fixname,':'))==NULL) return(-1);
	fixtype=*(cp+1);
	strcpy(mname,fixname);
	cp=strchr(mname,':');
	*cp=0;
	return 0;
}
/////////////////////////////////////////////////////////////////////////////
// ===========================================================================
/* The MacroFixHome Load function allocates a vector of MacroFix pointers of
	length MAXMACFIX.  For each fix in the MacroFixBank that is valid, i.e., that
	fixes a macrovariable in the currently loaded G Macro databank, a 
	MacroFix object is allocated on to the array, and initialization data
	from the "index" file is loaded.  If the fix is a rho-adjustment, the
	rho start date and value of Rho is also loaded.  Other fix values are 
	read by FixUpdate for the given year, t.
*/
short MacroFixHome::Load(MacFixBank& mfb, GBank& gb)  {
	short ifx,n,iret,ns_ws;
	char fixname[30], mname[30], fixtype;
	float ss[MacFixBank::NOMAX];

	nfix=0;
	ns_ws=gb.Getnseries();
	nmacro=ns_ws;

	n=mfb.Getnseries();
	for(ifx=0;ifx<n;ifx++) {
		// Walk through the fixnames in the MacFixBank:
		if((iret=mfb.GetFixName(ifx,fixname))<0) 
			continue;
		// Separate into macro name and fix type:
		if((iret=NameToFix(fixname,mname,fixtype))<0) continue;
		// See if the macro name is in the G model bank: (relies on global
		//    ns_ws and nam)
		if((iret=gb.findname(mname))<0) 
			continue;
		MFixVec[nfix] = new MacroFix(mname);
		MFixVec[nfix]->MacroNumber=iret;
		MFixVec[nfix]->FixNumber=ifx;
		MFixVec[nfix]->FixType=fixtype;
		
		if(fixtype=='r')   // if rho-adjustment, get Rho value
			mfb.GetRhoFix(ifx,MFixVec[nfix]->Rho,MFixVec[nfix]->RhoStart);
		nfix++;
		}
	return 0;
}
// ===========================================================================
//  Default void constructor, needed when declaring without arguments.
MacroFixHome::MacroFixHome() {
	nfix = nmacro = 0;
	for(short i=0;i<MacFixBank::MAXMACFIX;i++)
		MFixVec[i] = 0;
}
// ===========================================================================
// Copy Constructor:    (also uses MacroFix copy constructor)
MacroFixHome::MacroFixHome(MacroFixHome& mfh)  {
	nfix=mfh.nfix;
	nmacro=mfh.nmacro;
	for(short i=0;i<nfix;i++)
		MFixVec[i] = new MacroFix(*(mfh.MFixVec[i])); 
	// check that this uses MacroFix copy constructor correctly
}
// ===========================================================================
MacroFixHome::~MacroFixHome(void) {
	short i;
	if(nfix==0) return;
	for(i=0;i<nfix;i++)
		if(MFixVec[i]) delete MFixVec[i];
}
// ===========================================================================
/* FixUpdate updates the values of all non-rho and non-skip fixes for the
	current year t.
*/

short MacroFixHome::FixUpdate(short t, MacFixBank &mfb) {
	short i,ifx;
	float val;

	for(i=0;i<nfix;i++) {
		if(MFixVec[i]->FixType=='r' || MFixVec[i]->FixType=='k')
			continue;
		ifx=MFixVec[i]->FixNumber;
		mfb.GetFixObs(ifx,t,val);
		MFixVec[i]->FixValue=val;
		}
	#if defined(INCL_EQFIX)
	Interval = GDate(t,1,1) - BaseDate + 1;
	#endif
	return 0;
}
// ===========================================================================
/* GetRho gets back rho-adjustment data, given the "seq" or sequence number
	of the model macrovariable.
*/
short MacroFixHome::GetRho(short seq, short& fixnum,float& rho, 
		short& rhostart, float& error) {
short i;
for(i=0;i<nfix;i++) {
	if(MFixVec[i]->MacroNumber==seq && MFixVec[i]->FixType=='r') {
		rho=MFixVec[i]->Rho;
		rhostart=MFixVec[i]->RhoStart;
		error=MFixVec[i]->Error;
		fixnum=i;
		return 0;
		}
	}
return(-1);  // No fixes on macrovariable.
}
//===================================================================
/* PutError saves the error back to the MacroFixHome element.*/
short MacroFixHome::PutError(short fixnum, float error) {
	MFixVec[fixnum]->Error=error;
	return 0;
}
//===================================================================
short IsMissing(float x) {
	if( x < -.0000009 && x > -.0000011)
		return 1;
	return 0;
}
//===================================================================
short MacroFixHome::ApplyFixes(short seq, float& depend) {
short i,err;
short funlen,fixno,sd;
char *cp;
char fnstring[500],name[500], *pname;
Chopper ch;
#ifdef INCL_EQFIX
float *f = new float[Interval+1];
#endif
float FixVal, LagVal, rhs, FixBaseVal, SeriesBaseVal, ratio;
short fixbaseyr;  // used in dynamic index and growth fixes
pname=name;

sd = (short)(gb.GetBaseYr() +.00001); //startdate is a float 

for(i=0;i<nfix;i++) {
	if(MFixVec[i]->MacroNumber==seq) {
		switch (MFixVec[i]->FixType) {
			case 'k':
				depend=Tsptr[seq]->ts[t-sd]; // or do we want to read directly from the file?
				break;
			case 'o': // all these cases are treated as actuals. ("oigs")
			case 'i':
			case 'g':
			case 's':
				if(!IsMissing(MFixVec[i]->FixValue))
					depend =  MFixVec[i]->FixValue;
				break;
			case 'm':
				if(!IsMissing(MFixVec[i]->FixValue))
					depend *= MFixVec[i]->FixValue;
				break;
			case 'c':
				if(!IsMissing(MFixVec[i]->FixValue))
					depend += MFixVec[i]->FixValue;
				break; 
			// Here to implement the "dynamic" fixes.  They are a little bit
			//   more complicated than the rest.
			case 'n':   // dynamic index fix
				// Base year of fix is startfix
				if( (fixbaseyr=mfbank.Getstartfix(i)) >= t )
					continue;
				err = mfbank.GetFixObs(i+1,fixbaseyr,FixBaseVal);
				if(IsMissing(FixBaseVal) || fabs(FixBaseVal) < 1.0e-7) {
					cprintf("Error applying dynamic index fix on %s in %d: Base value of fix is zero!\r\n",
						Tsptr[seq]->tsname,t);
					return -1;
					}
				SeriesBaseVal = Tsptr[seq]->ts[fixbaseyr-sd];
				if(!IsMissing(MFixVec[i]->FixValue)) {
					depend = MFixVec[i]->FixValue * SeriesBaseVal / FixBaseVal;
					}
				break;
			case 'w':  // dynamic growth rate fixes
			case 't':
				// For growth rates, base is 1 year before first fix.
				if( (fixbaseyr=mfbank.Getstartfix(i)-1) >= t )
					continue;
				SeriesBaseVal = Tsptr[seq]->ts[t-1-sd];
				FixVal = MFixVec[i]->FixValue;
				depend = SeriesBaseVal * exp(FixVal/100.); // 2 -> .02

				break;
			#ifdef INCL_EQFIX
			case 'e':  // equation fix 
				FixVal=MFixVec[i]->FixValue;
				if(!IsMissing(FixVal)) {   // apply fix only in years specified
					VecFixNumber = -1;
					// Here is where the bulk of the work is done:
					MacFixNumber = fixno = MFixVec[i]->FixNumber;
					cp=mfbank.GetFunction(fixno);
					if(cp!=NULL)
					strcpy(fnstring,cp);  
					strcat(fnstring,"\n");  // help chopper
					ch.SetSource(fnstring);
					err = ch.rhs(f,pname);
					FixVal = MFixVec[i]->FixValue;  // save fix value
					depend = f[Interval-1];
					}
				break;
			case 'f':  // follow fix
				FixVal=MFixVec[i]->FixValue;
				LagVal=Tsptr[seq]->ts[t-sd-1];
				if(IsMissing(LagVal) || fabs(LagVal)<1.0e-7) {
					cprintf("Cannot apply follow fix to %s in %d.\r\n",
						Tsptr[seq]->tsname,t);
					cprintf("  Lagged value of variable is zero or missing.\r\n");
					return(-1);
					}
				if(!IsMissing(FixVal)) {   // apply fix only in years specified
					VecFixNumber = -1;
					// Here is where the bulk of the work is done:
					MacFixNumber = fixno = MFixVec[i]->FixNumber;
					cp=mfbank.GetFunction(fixno);
					if(cp!=NULL)
					strcpy(fnstring,cp);  
					strcat(fnstring,"\n");  // help chopper
					ch.SetSource(fnstring);
					err = ch.rhs(f,pname);
					// f is the value of the expression
					// FixVal is the extra growth rate to apply
					// LagVal is the lagged value of the variable
					//   (must not be zero or missing)
					FixVal = MFixVec[i]->FixValue;  // save fix value
					ratio = f[Interval-1]/f[Interval-2];
					if(ratio <= 1.0e-7) {
						cprintf("Cannot apply follow fix to %s in %d.\r\n",
							Tsptr[seq]->tsname,t);
						cprintf(" Follow expression %s is zero or negative.\r\n",
							fnstring);
						}
					depend = LagVal*exp(FixVal/100+log(f[Interval-1]/f[Interval-2]));
					}
				break;
			case 'h':  // share fix
				FixVal=MFixVec[i]->FixValue;
				if(!IsMissing(FixVal)) {   // apply fix only in years specified
					VecFixNumber = -1;
					// Here is where the bulk of the work is done:
					MacFixNumber = fixno = MFixVec[i]->FixNumber;
					cp=mfbank.GetFunction(fixno);
					if(cp!=NULL)
					strcpy(fnstring,cp);  
					strcat(fnstring,"\n");  // help chopper
					ch.SetSource(fnstring);
					err = ch.rhs(f,pname);
					FixVal = MFixVec[i]->FixValue;  // save fix value
					rhs = f[Interval-1];
					if(IsMissing(rhs))
						depend = MISSING;
					else
						depend = FixVal * rhs;
					}
				break;
			#endif
			case 'r':
				continue;
			default:
				continue;
			}  // end cases
		} // end if
	} // end loop i

#ifdef INCL_EQFIX
if(f)
	delete f;
#endif
return 0;
}
//===================================================================
/* Modify -- Modify handles 9 types of fixes:
	(skip='k',ovr='o,mul='m',cta='c',ind='i',gro='g',stp='s',eqn='e',rho='r')
	When the setrho flag is set to 'y', the rho-adjustment is applied.
	Normally, all equations are called at the end of each year with
	setrho set to 'y'.  Otherwise the rho-adjustment is not applied.
	For all other fixes, more than one type of fix is legal, but two
	fixes of the same type (i.e. 'i') on the same variable will not
	both take.  In this case, the fix given later in the fixes file
	overwrites the fix given earlier in the file.
*/

/* The following function assumes we have opened up MacFixBank, and
	initialized a MacroFixHome object called "mfh".  If MacFixesRead
	is false (global dyme.inc variable), this routine returns.
*/

// modify is a method of Tseries, (seq is member of Tseries)
void Tseries::modify(float depend, short eqType) {
	::modify(seq,depend,eqType);
	return;
}

void modify(short seq, float depend, short eqType) {
/* Note that we cannot locate fixes with seq number under current setup.
	This is because there may be more than one fix on a given macrovariable,
	and they are stored in different locations.
*/
short err;
float rho,error,actual,lagged,SaveDepend;
short rhostart,fixnum,jahrba;

jahrba = gb.GetBaseYr();
if(MacFixesRead) {
	if(setrho == 'y') {
		if((err=mfh.GetRho(seq,fixnum,rho,rhostart,error))==-1) return; // no rho-adj.
		if(rho == 0) return; // no rho-adjustment
		if(rhostart>t) return;  // not ready yet.
		if(rhostart<t && t>godate) {
			if(error < .000001 && error > -.000001)
				error = 0.; // Prevent arithmetic underflow.
			error *= rho; //set rho-adjust error for next year
			mfh.PutError(fixnum,error);
			return;
			}
		// initialize rhoerror -- How we do this depends on eqType:

		actual = Tsptr[seq]->ts[t-jahrba];
		lagged = Tsptr[seq]->ts[t-jahrba-1];
		if(eqType == Tseries::REG) {
			error = (actual - depend);
			}
		else if(eqType == Tseries::LOG) {
			error = log(actual) - depend;
			}
		else if(eqType == Tseries::DIFF) {
			error = (actual - lagged) - depend;
			}
		else if(eqType == Tseries::PERC) {
			error = 100 * ((actual - lagged) / lagged) - depend;
			}
		else if(eqType == Tseries::DIFF_LOG) {
			error = (log(actual) - log(lagged)) - depend;
			}
//  NOTE: you can uncomment these if you have a macrovariable named 
//     "pop" for population:
//              else if(eqType == Tseries::LOG_PC) {
//                      error = (log(actual) / pop[t]) - depend;
//                 }
//              else if(eqType == Tseries::DIFF_PC) {
//                      error = ((actual/pop[t]) - (lagged/pop[t-1])) - depend;
//                 }
//              else if(eqType == Tseries::DIFF_LOG_PC) {
//                      error = (log(actual / pop[t]) - log(lagged / pop[t-1]))
//                                               - depend;
//                 }
		else {
			printf("Illegal type of equation");
			tap();
			}
		error *= rho;

		mfh.PutError(fixnum,error);
		return;
		} // end "setrho=='y' "


	// Here when not setting rho-adjust error
	// A rhostart >= t works like a skip up to that date.
	if((err=mfh.GetRho(seq,fixnum,rho,rhostart,error))==-1) { // no rho-adj.
		rhostart=0;
		error=0.;
		rho=0.;
		}
// experiment, to get PERC modify to work:
//   if(rhostart < t && rhostart != 0)  {
	if(rhostart < t)  {
		// Apply rho-adjustment, if any.
		depend += error;
		/* Modify "depend", depending on the type of equation, to 
			return the desired value.  For example, with DIFF_LOG:
			say the the equation for the variable V is in first 
			difference log form, i.e., what you get from the equation
			is:
				log(V) - log(V[1]) = depend
			but what you want is V, so the conversion is:
				V = exp(depend + log(V[-1]))
		*/
		lagged = Tsptr[seq]->ts[t-jahrba-1];
		if(eqType == Tseries::REG) {
			// nothing done
			}
		else if(eqType == Tseries::LOG) {
			depend = exp(depend);
			}
		else if (eqType == Tseries::DIFF) {
			depend = lagged + depend;
			}
		else if (eqType == Tseries::PERC) {
			depend = (depend * lagged/100) + lagged;
			}
		else if (eqType == Tseries::DIFF_LOG) {
			depend = exp(log(lagged) + depend);
			}
//      else if (eqType == Tseries::LOG_PC) {
//            depend = exp(depend) * pop[t];
//            }
//      else if (eqType == Tseries::DIFF_PC) {
//            depend = (depend + (lagged/pop[t-1])) * pop[t];
//            }
//      else if (eqType == Tseries::DIFF_LOG_PC) {
//            depend = (exp(depend + log(lagged/pop[t-1])) * pop[t];
//            }
		else {
			cprintf("Illegal type of equation.\r\n");
			tap();
			}
		} // end if(rhostart..
	// Now apply all other types of fixes.  A "skip" is treated just like
	//    an override fix, but it applies the current value of the value in
	//    the model bank.  Other fixes may be applied after it!

	// Apply fixes, and copy equation-calculated values only after year
	//    of rhostart.  Do not apply fixes when setrho == 'y'
	if(rhostart < t && setrho != 'y') {
		SaveDepend=depend;  // If no valid fix, save previous value.
		err=mfh.ApplyFixes(seq,depend); // Like a modify inside of modify.
		// Copy value of depend over into tseries object in Tsptr array:
		if(!IsMissing(depend))
			Tsptr[seq]->ts[t-jahrba]=depend;
		else
			Tsptr[seq]->ts[t-jahrba]=SaveDepend;
		}
	}
else  // (macfixes not read, still use depend.)
	if(t > MacEqStartDate) Tsptr[seq]->ts[t-jahrba]=depend;

err=err;
return;
}

//////////////////////////////////////////////////////////////////////////////
/*==========================================================================*/
short MacFixBank::Load(const char *rootname)  {
// Opens an existing MacFixBank, like G AssignBank.

short i,err;
long posit,funcspos=0,cumpos=0;
char fname[120];
bnkHeader Hdr;
size_t MacFixSize=0;

//printf("Constructing object for MacFixBank %s.\n",rootname);

names=NULL;
indx=NULL;
indname=NULL;
bnktitle=NULL;
bnkname=NULL;
opened=-1;

strcpy(fname,"MacFixBank");
classname=new char[strlen(fname)+1];
strcpy(classname,fname);

/* Open already existing macro fixes .bnk file and initialize. */
strcpy(fname,rootname);
strcat(fname,".bnk");
bnkname= new char[strlen(fname)+1];
strcpy(bnkname,fname);
if((pbank = fopen(bnkname,"rb")) == 0 ||
			Hdr.ReadIn(pbank) == -1){
		printf(msgbk1,bnkname);
		banktype = 'x';
	goto failure;
	}
jahra = Hdr.jahra;
jahra = (jahra<1900)?jahra+1900:jahra;
stper = Hdr.stper;
nobs = Hdr.nobs;
bnktitle = new char[strlen(Hdr.bnktitle)+1];
strcpy(bnktitle,Hdr.bnktitle);

/* Open macro fixes ".ind" file and initialize. */
strcpy(fname,rootname);
strcat(fname,".ind");
indname= new char[strlen(fname)+1];
strcpy(indname,fname);

if((ipbank = fopen(indname,"rb")) == 0) {
	printf(msgbk2,indname);
	goto failure;
	}
posit = 0;
if((fseek(ipbank,posit,0)) != 0) {   /* posit at begin of file */
	printf(msgbk15);
	goto failure;
	}

if((fread(&nseries,sizeof(short),1,ipbank)) == 0) {
	printf(msgbk16);
	goto failure;
	}
indx = new unsigned short[nseries+1];

if((fread(indx,sizeof(unsigned short),nseries+1,ipbank)) == 0) {  /* read the pointer */
	  printf(msgbk17);
	  goto failure;
	  }
ncin = indx[nseries]+1;
names = new char[ncin];

if((fread(names,1,ncin,ipbank)) == 0) {  /* read the names  */
	printf(msgbk18);
	goto failure;
	}

// allocate fixes array.  This is read after names information.
mfx = new MacFix[nseries];

for(i=0;i<nseries;i++) {
	err = mfx[i].ReadIn(ipbank);
	mfx[i].function=0;
	}
//  The following doesn't work without "packed" structures:
//if((fread(mfx,sizeof(MacFix),nseries,ipbank)) == 0) {  /* read the fix array */
//	  printf(msgbk17);
//	  goto failure;
//	  }
// For Equation Fixes:
nfunc=0;   cumpos=0;  
	#if defined(INCL_EQFIX)
// chk if we want sizeof(char*) at end of this:, yes for consistency with
//   DOS sizeof(MacFix)!
MacFixSize = 16*sizeof(char)+3*sizeof(short)+sizeof(long)+sizeof(char*);
funcspos = sizeof(short) + sizeof(unsigned short)*(nseries+1) +
	sizeof(char)*ncin + MacFixSize*nseries;
for(i=0;i<nseries;i++) {
	if(mfx[i].funlen>0) {
		nfunc++;
		mfx[i].funpos = cumpos;
		cumpos += mfx[i].funlen;  // = strlen(func) + 1
		posit = funcspos + mfx[i].funpos;
		if((fseek(ipbank,posit,0)) != 0) {
			printf(msgbk15);
			goto failure;
			}
		mfx[i].function = new char[mfx[i].funlen];
		fread(mfx[i].function,mfx[i].funlen,1,ipbank);
		}
	}
	#endif
err=err;
banktype = 'm';
goto success;
failure:
	opened = -1;
	if(mfx) delete[] mfx;
	if(names) delete[] names;
	if(indx) delete[] indx;
	return -1;
success:
	opened = 1;
	return 0;
}
/*==========================================================================*/
// Copy constructor.  This allows you to do a statement like:
//              mfb = MacroFixBank("macfixes",gb);
MacFixBank::MacFixBank(MacFixBank& mf) {
	short i;
	classname = new char[strlen(mf.classname)+1];
	strcpy(classname,mf.classname);
	bnkname = new char[strlen(mf.bnkname)+1];
	strcpy(bnkname,mf.bnkname);
	indname = new char[strlen(mf.indname)+1];
	strcpy(indname,mf.indname);
	bnktitle = new char[strlen(mf.bnktitle)+1];
	strcpy(bnktitle,mf.bnktitle);
	opened=mf.opened;
	pbank=mf.pbank;
	ipbank=mf.ipbank;
	jahra=mf.jahra;
	nseries=mf.nseries;
	stper=mf.stper;
	nobs=mf.nobs;
	// Now, macfxbnk elements:
	mfx = new MacFix[nseries];
	for(i=0;i<nseries;i++)
		mfx[i] = mf.mfx[i];
	indx=new unsigned short[nseries];
	for(i=0;i<nseries;i++)
		indx[i] = mf.indx[i];
	ncin=mf.ncin;
	names = new char[ncin];
	for(i=0;i<ncin;i++)
		names[i] = mf.names[i];
} 
/*==========================================================================*/
// free up memory allocated in constructor.  close data files.
MacFixBank::~MacFixBank() {
	short i;
	//printf("Destructor for MacFixBank called.  Freeing up memory.\n");
	if(classname) 
		delete[] classname;
	if(bnkname) 
		delete[] bnkname;
	if(bnktitle) 
		delete[] bnktitle;
	if(indname) 
		delete[] indname;
	if(mfx) 
		delete[] mfx;
	if(indx) 
		delete[] indx;
	if(names) 
		delete[] names;
	if(MsgObject)
		delete MsgObject;
	fclose(pbank);
	fclose(ipbank);
}
/*==========================================================================*/
short MacFixBank::rindex(void)  { // Read fix array from index file.
long posit,cumpos,funcspos;
short i,err;
size_t MacFixSize = 0;

posit = 0.0;
if((fseek(ipbank,posit,0)) != 0) {   /* posit at begin of file */
	printf(msgbk15);
	return(ERR);
	}

if((fread(&nseries,sizeof(short),1,ipbank)) == 0) {    /* read the number of series */
	printf(msgbk16);
	return(ERR);
	}
if((fread(indx,sizeof(unsigned short),nseries+1,ipbank)) == 0) {  /* read the pointer */
	printf(msgbk17);
	return(ERR);
	}
ncin = indx[nseries] + 1;
if((fread(names,1,ncin,ipbank)) == 0) {  /* read the names  */
	printf(msgbk18);
	return(ERR);
	}

for(i=0;i<nseries;i++) {
	mfx[i].ReadIn(ipbank);
	mfx[i].function=0;
	}
//if((fread(mfx,sizeof(MacFix),nseries,ipbank)) == 0) {
//   printf(msgbk17);
//   return(ERR);
//   }
/* Read the functions, if they exist. */
	#if defined(INCL_EQFIX)
MacFixSize = 16*sizeof(char)+3*sizeof(short)+sizeof(long)+sizeof(char*);
funcspos = sizeof(short) + sizeof(unsigned short)*(nseries+1) +
	sizeof(char)*ncin + MacFixSize*nseries;
cumpos = 0;
for(i=0;i<nseries;i++) {
	if(mfx[i].funlen>0) {
		mfx[i].funpos = cumpos;
		cumpos += mfx[i].funlen;
		posit = funcspos + mfx[i].funpos;
		if((fseek(ipbank,posit,0)) != 0) {
			printf(msgbk15);
			goto failure;
			}
		mfx[i].function = new char[mfx[i].funlen];
		fread(mfx[i].function,mfx[i].funlen,1,ipbank);
		}
	}
	#endif
cprintf(msgbk19);
return(OK);
failure:
return(-1);
}
/*==========================================================================*/
// GetFix retrieves the fix series from .bnk file using the
//   which variable, which is the index of the fix in the fix list, and
//   therefore also the index in the .bnk file.  This is like the virtual
//   getser function without the series name.

short MacFixBank::GetFix(short which,float *s) {
	short err,i,j,gap,max_obs;
	long temp,posit,bnkheadsize;
	float val;
	char *pnam,cnpy;

	for( i=0;i<nobs;i++)
		s[i] = 0;
	bnkheadsize=sizeof(short)*3+sizeof(char)*80;
	posit = which * (long)nobs * sizeof(float) + which + bnkheadsize;
	if((err = fseek(pbank,posit,0)) == ERR) {
		printf(" \n cannot fseek \n");
		return(ERR);
		}
	if((err=fread(&cnpy,1,1,pbank))==0 )  {
		printf(" \n cannot read fix %d from %s\n ",which,bnkname);
		return(ERR);
		}
	if((err=fread((void*)s,sizeof(float),nobs,pbank))==0)  {
		printf(" \n cannot read fix %d from %s\n ",which,bnkname);
		return(ERR);
		}  err=err;
	return 0;
	error:       return(ERR);
}
//===================================================================
/* GetFixName takes an index into the fixnames array, and returns the
	fixname at that index.
*/
short MacFixBank::GetFixName(short i, char* name){
	unsigned short pt;
	pt=indx[i];
	if(pt>ncin) return(-1); // illegal index
	strcpy(name,&names[pt]);
	return 0;
}
//===================================================================
short MacFixBank::GetRhoFix(short which, float& rho, short& rhostart){
	long posit,temp,bnkheadsize;
	short err;
	char cnpy;
	temp=which;
	float rs;

	bnkheadsize=sizeof(short)*3+sizeof(char)*80;
	posit = temp * (long)(nobs*sizeof(float)+1)+bnkheadsize;
	if((fseek(pbank,posit,0)) != 0) {
		printf(msgbk7);
		return (-1);
		}
	if((err=fread(&cnpy,1,1,pbank))==0 )  {
		printf(" \n cannot read fix %d from %s\n ",which,bnkname);
		return(ERR);
		}
	if((err=fread((void*)&rho,sizeof(float),1,pbank))==0)  {
		printf(" \n cannot read rho from %s\n ",bnkname);
		return(ERR);
		}
	if((err=fread((void*)&rs,sizeof(float),1,pbank))==0)  {
		printf(" \n cannot read rs from %s\n ",bnkname);
		return(ERR);
		}   err=err;
	rhostart=(short)rs;
	return 0;
}
//===================================================================
short MacFixBank::GetFixObs(short which, short t, float& value){
	long posit,temp,bnkheadsize;
	char cnpy;
	short base,err;

	temp=which;
	base=(jahra<1900)?jahra+1900:jahra; // t is 4-digit
	bnkheadsize=sizeof(short)*3+sizeof(char)*80;
	posit = temp * (long)(nobs*sizeof(float)+1)
			+ (long)(t-base)*sizeof(float) + 1 + bnkheadsize;
	if((fseek(pbank,posit,0)) != 0) {
		printf(msgbk7);
		return (-1);
		}
	if((err=fread((void*)&value,sizeof(float),1,pbank))==0)  {
		printf(" \n cannot read rho from %s\n ",bnkname);
		return(ERR);
		} err=err;
	return 0;
}
/*==========================================================================*/
char* MacFixBank::GetFunction(short fixno) {
char *cp;
if(fixno < 0 || fixno > nseries)
	return NULL;
if(mfx[fixno].funlen==0)
	return NULL;
return (cp=mfx[fixno].function);
}
/*==========================================================================*/
short MacFixBank::Getstartfix(short fixno) {
if(fixno < 0 || fixno > nseries)
	return -1;

return mfx[fixno].startfix;
}
/*==========================================================================*/
MacFix::MacFix() {
strcpy(Mname,"");
startfix=-1;
endfix=-1;
funlen=0;
funpos=0;
function=0;
}
/*==========================================================================*/
MacFix::~MacFix() {
	#if defined(INCL_EQFIX)
if(function)
	delete function;
	#endif
}
/*==========================================================================*/
// Simple constructor for MacFixBank, does nothing.
MacFixBank::MacFixBank(void) {
	classname=0;
	bnkname=0;
	bnktitle=0;
	indname=0;
	mfx=0;
	indx=0;
	names=0;
}
/*==========================================================================*/
char *MacFixBank::Setbnktitle(const char *s) {
	short err;
	bnkHeader Hdr;

	bnktitle = new char[strlen(s)+1];
	strcpy(bnktitle,s);
	Hdr.jahra=jahra;
	Hdr.nobs=nobs;
	Hdr.stper=stper;
	strcpy(Hdr.bnktitle,bnktitle);
	// update bank with title, too.
	if((fseek(pbank,0L,0)) != 0) {   /* posit at begin of file */
		printf(msgbk15);
		return(NULL);
		}
	if((err=Hdr.WriteOut(pbank)) == -1){
			printf(msgbk1,bnkname);
			return(NULL);
			}
	err=err;
	return bnktitle;
}
/*==========================================================================*/
short MacFixBank::Announce(short bankno) {
// Give announcement that is normally given by bankcmd in G.
cprintf(msgbk19);
cprintf(msgbk3,bnktitle);
cprintf(msgbk4,jahra,stper);
cprintf(msgbk5,nobs);
cprintf(msgbk6,nseries);
return 0;
}
#endif
