/* Config.cpp reads the dyme.cfg file and draws the opening screen. */

#include "dymesys.h"  // dyme system files
#include "user.h"     // user-specific files

/* configure() -- Read the vam.cfg file. **********************/
short configure(void)
{
    short len,err = 0;
    char line[80],szGoDate[6],szStopDate[6],szUseall[4],szDebug[6],szMaxit[4];
    char szDiscDate[6], szOptSpecFileName[60],szMacEqStartDate[6],
        szRandomDraws[7],szAdditiveRandomErrors[5],
        szRandomRegressionCoefficients[5];
    FILE *fpcfg;
    
    len = strlen(CfgFileName);
    if(len == 0) strcpy(CfgFileName,"dyme.cfg");    
    if((fpcfg = fopen(CfgFileName,"r+")) == 0){
        printf("Cannot open %s as the configuration file.\n",CfgFileName);
        return(ERR);
        }
    err += getopt(fpcfg,RunTitle,76);
    err += getopt(fpcfg,szGoDate,4); 
    err += getopt(fpcfg,szStopDate,4);
    err += getopt(fpcfg,szMacEqStartDate,4);
    err += getopt(fpcfg,szDiscDate,4);
    err += getopt(fpcfg,szUseall,3);
    err += getopt(fpcfg,VecFixFileName,60);
    err += getopt(fpcfg,MacFixFileName,60);
    err += getopt(fpcfg,VamFileName,60);
    err += getopt(fpcfg,GbankName,60);
    err += getopt(fpcfg,szDebug,5);
    err += getopt(fpcfg,szMaxit,3);
    err += getopt(fpcfg,szOptSpecFileName,60);// a file name
    err += getopt(fpcfg,szRandomDraws,7); // number
    err += getopt(fpcfg,szAdditiveRandomErrors,5); // yes or no
    err += getopt(fpcfg,szRandomRegressionCoefficients,5); // yes or no
    if(err < 0){
        // Printf added by CA 2005 Aug 23
        printf("There were %d errors in reading the configuration file.\n", -err);
        exit(1);
        }
/***************************/
    #if defined(INCL_SCREEN)
    clrscr();
    videoinit();
    // _setcursortype(_NOCURSOR);

    /* in the winpdef function, the string written must have EXACTLY the same 
       length as the field written.  On the other hand, strings padded with 
       blanks at the end WILL NOT work in fopen.
       */
    _setcursortype(_NORMALCURSOR);

    wopen(0,0,24,79,1,RED|_BLUE,WHITE|_BLUE);
    wtitle("[  SLIMDYME MODEL -- RUN OPTIONS  ]",TCENTER,WHITE|_BLUE);
    wprints(1,2,YELLOW|_BLUE,"Esc accepts initial values. Enter at end of screen accepts new values.");
    wprints(2,2,YELLOW|_BLUE,"Ctrl-End moves to end of screen. Use arrow and Ctrl-arrow to position.");
    wprints(3,35,YELLOW|_BLUE,"Run Title");
    winpdef(4,1,RunTitle,"****************************************************************************",0,LBLUE|_LGREY,1,NULL);
    wprints(6,2,YELLOW|_BLUE,"Start year");
    winpdef(6,15,szGoDate,"9999",0,LBLUE|_LGREY,1,NULL);
    wprints(6,25,YELLOW|_BLUE,"Finish year");
    winpdef(6,40,szStopDate,"9999",0,LBLUE|_LGREY,1,NULL);
    wprints(6,50,YELLOW|_BLUE,"Discrepancy year");
    winpdef(6,70,szDiscDate,"****",0,LBLUE|_LGREY,1,NULL);
    wprints(8,2,YELLOW|_BLUE,"Use all available data?");
    winpdef(8,28,szUseall,"***",0,LBLUE|_LGREY,1,NULL);
    wprints(10,2,YELLOW|_BLUE,"Vector Fix File");
    winpdef(10,17,VecFixFileName,"************************************************************",0,LBLUE|_LGREY,1,NULL);
    wprints(12,2,YELLOW|_BLUE,"Macro Fix File");
    winpdef(12,17,MacFixFileName,"************************************************************",0,LBLUE|_LGREY,1,NULL);
    wprints(14,2,YELLOW|_BLUE,"Vam File Name");
    winpdef(14,17,VamFileName,"************************************************************",0,LBLUE|_LGREY,1,NULL);
    wprints(18,2,YELLOW|_BLUE,"Gbank Name");
    winpdef(18,17,GbankName,"************************************************************",0,LBLUE|_LGREY,1,NULL);
    wprints(22,2,YELLOW|_BLUE,"Debug");
    winpdef(22,17,szDebug,"*****",0,LBLUE|_LGREY,1,NULL);
    wprints(22,42,YELLOW|_BLUE,"Maxit");
    winpdef(22,57,szMaxit,"***",0,LBLUE|_LGREY,1,NULL);
    wtextattr(WHITE|_LGREY);
    winpread();
    wclose();
    #endif

    godate = atoi(szGoDate);
    stopdate = atoi(szStopDate);
    godate = atoi(szGoDate);
    MacEqStartDate = atoi(szMacEqStartDate);
    DiscDate = atoi(szDiscDate);
    useall = szUseall[0];
    debug = atoi(szDebug);
    maxit = atoi(szMaxit);
    depad(VecFixFileName);
    depad(MacFixFileName);
    depad(VamFileName);
    depad(GbankName);
    TackExt(VamFileName,"vam");
    
    depad(szOptSpecFileName);
    if(strcmp(szOptSpecFileName,"none")== 0 || strlen(szOptSpecFileName) == 0)
        MaxFlag = 'n';
    else{
        MaxFlag = 'y';
        strcpy(OptSpecFileName,szOptSpecFileName);
        }
    num_draw = atof(szRandomDraws);
    depad(szAdditiveRandomErrors);
    depad(szRandomRegressionCoefficients);
    if(szAdditiveRandomErrors[0] == 'y') rnderr = 1;
    else rnderr = 0;
    if(szRandomRegressionCoefficients[0]== 'y') rndcoef = 1;
    else rndcoef = 0;
    // Check consistency of desires
    if (num_draw > 0 && rnderr + rndcoef == 0) num_draw = 0;
    if(MaxFlag == 'y' && num_draw >0){
        printf("You cannot have both optimization and stochastic simulation.\n");
        printf("Fix .cfg file.\n");
        exit(1);
        }
    fclose(fpcfg);
    return(OK);
    }       

