/* PADS.CPP -- This is the class that implements the simulation of
   PADS - the "Perhaps Adequate Demand System" of consumption.
*/

#include "dymesys.h"

#if defined(INCL_PADS)
Pads::Pads(){
    neq = 0;
    ng = 0;
    nsg = 0;
    }

Pads::Pads(char *fname){
    r(fname);
    }
void Pads::r(char *fname){
    char card[241],*s,typ;
    short i,j,k,m;
    float x;
    FILE *fp;

    filename = new char[strlen(fname)+1];
    strcpy(filename,fname);
    if((fp = fopen(filename,"r")) == 0){
        printf("Cannot open %s.\n",filename);
        return;
        }
    fgets(card,241,fp);
    i = 0;
    getint(neq,fp,card,i);
    getint(rhostart,fp,card,i);
    getint(ng,fp,card,i);
    getint(nsg,fp,card,i);
    sectors   = ivector(1,neq);
    rhos      = vector(1,neq);
    errors    = vector(1,neq);
    group    = ivector(1,neq);
    subgroup = ivector(1,neq);
    popnr     = ivector(1,neq);
    cstarnr   = ivector(1,neq);
    trendnr   = ivector(1,neq);
    insystem  = ivector(1,neq);
    lambda.resize(neq,1,1,1);
    share.resize(neq,1,1,1);
    mu.resize(ng,1,1,1);
    nu.resize(nsg,1,1,1);
    types     = new char[neq+1];
    coefs     = matrix(1,neq,1,4);
    // zero out
    for(i = 1; i <= neq; i++){
        sectors[i] = 0;
        rhos[i] = 0;
        errors[i] = 0;
        group[i] = 0;
        subgroup[i] = 0;
        popnr[i] = 0;
        cstarnr[i] = 0;
        trendnr[i] = 0;
        insystem[i] = 0;
        lambda[i] = 0;
        share[i] = 0.;
        types[i] = ' ';
        for(j = 1; j <= 4; j++)
        coefs[i][j] = 0;
        }
    //Read the mu and nu
    fgets(card,241,fp);
    j = 0;
    for(i = 1;i<=ng;i++) {
        getfloat(x,fp,card,j);
        mu[i] = x;
        }
    fgets(card,241,fp);
    j = 0;
    for(i = 1;i<=nsg;i++) {
        getfloat(x,fp,card,j);
        nu[i] = x;
        }

    // read individual equations
    i = 0;
    while(fgets(card,241,fp) != NULL){
        if(strlen(card) <= 1) break;
        i++;
        if(i > neq){
            printf("%d is too many equations in %s.\n",i,filename);
            exit(1);
            }
        j = 0;
        getint(sectors[i],fp,card,j);
        getint(group[i],fp,card,j);
        getint(subgroup[i],fp,card,j);
        getint(popnr[i],fp,card,j);
        getint(cstarnr[i],fp,card,j);
        getint(trendnr[i],fp,card,j);
        getint(insystem[i],fp,card,j);
        getfloat(rhos[i],fp,card,j);
        getfloat(share[i],fp,card,j);
        getfloat(lambda[i],fp,card,j);
        for(k = 1; k <= 4; k++)
            getfloat(coefs[i][k],fp,card,j);
        }
        // debug
        /*printf("From constructor: t=%d:\n",t);
        for(i = 1; i<=10;i++){
                printf("\n%2d %6.3f %6.3f ",i,share[i], lambda[i]);
                for(k = 1; k <= 4; k++)
                        printf(" %8.3f",coefs[i][k]);
                }
        tap();
        */
    fclose(fp);
    }


Pads::~Pads(){
        //free_ivector(sectors,1);  // done in ~Equation().
        //free_vector(rhos,1);
        //free_vector(errors,1);
        free_ivector(group,1);
        free_ivector(subgroup,1);
        free_ivector(popnr,1);
        free_ivector(cstarnr,1);
        free_ivector(trendnr,1);
        free_ivector(insystem,1);
        }
//=====================================================================
//  Version of equation with prices, income and total population:

void Pads::padsf(Vector& pce, Vector& p, Vector& tempi,
      Tseries& pop, Tseries& pcetotR){
        short i,j,k;
        float lpbar, lpt, incterm,incterm1,incterm2, consump, cadj, dif,
                pcesum, coefsum, pcei;
        Vector lp(neq), pg(ng), lpg(ng), psg(nsg), lpsg(nsg), //pce prices
        wg(ng), wsg(nsg), //sum of shares (weights) in groups and subgroups.
        pt(neq); //price term of PADS.

        //Debug
        /*printf("\nFrom padsf t=%d:\n",t);
        for(i = 1; i<=10;i++){
                printf("\n%2d %6.3f %6.3f ",i,share[i], lambda[i]);
                for(k = 1; k <= 4; k++)
                        printf(" %8.3f",coefs[i][k]);
                }
        */
        for(i = 1; i <= neq; i++){
                if(p[i] <= 0.) {
                        printf("Negative price for pce category %d replaced by 1.\n",i);
                        p[i] = 1.;
                        }
                lp[i] = log(p[i]);
                }
        lpbar = dot(lp,share);
        lpg.set(0.); lpsg.set(0.), wg.set(0.), wsg.set(0.);
        for(i = 1; i <= neq; i++){
                if(group[i] > 0){
                        lpg[group[i]] += share[i]*lp[i];
                        wg[group[i]] += share[i];
                        }
                if(subgroup[i] > 0){
                        lpsg[subgroup[i]] += share[i]*lp[i];
                        wsg[subgroup[i]] += share[i];
                        }
                }
        for(i = 1; i <= ng; i++){
                lpg[i] = lpg[i]/wg[i];
                }
        for(i = 1; i <= nsg; i++){
                lpsg[i] = lpsg[i]/wsg[i];
                }
        // Compute full price term for each product.
        for(i = 1; i <= neq; i++){
                lpt = 0.;
                if(insystem[i] == 1){
                        lpt = -lambda[i]*(lp[i] - lpbar);
                        if(group[i] > 0){
                                lpt += -mu[group[i]]*(lp[i] - lpg[group[i]]);
                                }
                        if(subgroup[i] > 0){
                                lpt += -nu[subgroup[i]]*(lp[i] - lpsg[subgroup[i]]);
                                }
                        for(k = 1; k <= neq; k++){
                                lpt += -lambda[k]*share[k]*(lp[i] - lp[k]);
                                }
                        }
                lpt = exp(lpt);
                pt[i] = lpt;

                incterm1 =      (coefs[i][1] + coefs[i][4]*tempi[trendnr[i]])*pop[t];

                incterm2 =  coefs[i][2]*pcetotR[t]+
                                                coefs[i][3]*(pcetotR[t] - pcetotR[t-1]);
                //printf("Sector %d %10.1f %10.1f \n",i,incterm1,incterm2);
                consump = pt[i]*(incterm1 +incterm2);
                pcei = pce[i];
            cadj = rhoadj(consump,pce[i],i);
            pce[i] = cadj;
                }
        coefsum = 0.;
        for(i = 1; i <= neq; i++)
                coefsum += coefs[i][2];
        pcesum = pce.sum();
        dif = pcetotR[t] - pcesum;
        dif = dif/coefsum;
        for(i = 1; i <= neq; i++)
                pce[i] += dif*coefs[i][2];
        //pt.Display("price term");
        }
//=====================================================================
// Version of equation with weighted populations (popul), and
//    cstar variables.  Total pce used for spreading only!

void Pads::padsf(Vector& pce, Vector& p, Vector& popul, Matrix& Cstar,
           Vector& tempi, Tseries& pcetotR){

short i,j,k;
float lpbar, lpt, incterm1, incterm2, consump, dif, pcesum, coefsum;

Vector lp(neq), pg(ng), lpg(ng), psg(nsg), lpsg(nsg), //pce prices
   wg(ng), wsg(nsg), //sum of shares (weights) in groups and subgroups.
        pt(neq); //price term of PADS.

/* Debug print coefficient
printf("From padsf t=%d:\n",t);
for(i = 1; i<=10;i++){
        printf("\n%2d %6.3f %6.3f ",i,share[i], lambda[i]);
        for(k = 1; k <= 4; k++)
                printf(" %8.3f",coefs[i][k]);
        }
*/
for(i = 1; i <= neq; i++){
        if(p[i] <= 0.) {
                printf("Negative price for pce category %d replaced by 1.\n",i);
                p[i] = 1.;
                }
        lp[i] = log(p[i]);
        }
lpbar = dot(lp,share);
lpg.set(0.); lpsg.set(0.), wg.set(0.), wsg.set(0.);
for(i = 1; i <= neq; i++){
        if(group[i] > 0){
                lpg[group[i]] += share[i]*lp[i];
                wg[group[i]] += share[i];
                }
        if(subgroup[i] > 0){
                lpsg[subgroup[i]] += share[i]*lp[i];
                wsg[subgroup[i]] += share[i];
                }
        }
for(i = 1; i <= ng; i++){
        lpg[i] = lpg[i]/wg[i];
        }
for(i = 1; i <= nsg; i++){
        lpsg[i] = lpsg[i]/wsg[i];
        }

// Compute full price term for each product:

for(i = 1; i <= neq; i++){
        j = sectors[i];
        lpt = 0.;
        if(insystem[i] == 1){
                lpt = -lambda[i]*(lp[i] - lpbar);
                if(group[i] > 0){
                        lpt += -mu[group[i]]*(lp[i] - lpg[group[i]]);
                        }
                if(subgroup[i] > 0){
                        lpt += -nu[subgroup[i]]*(lp[i] - lpsg[subgroup[i]]);
                        }
                for(k = 1; k <= neq; k++){
                        lpt += -lambda[k]*share[k]*(lp[i] - lp[k]);
                        }
                }
        lpt = exp(lpt);
        pt[j] = lpt;

        incterm1 = (coefs[i][1] + coefs[i][4]*tempi[j])*popul[j];

        incterm2 =  coefs[i][2]*Cstar[0][j] + coefs[i][3]*(Cstar[0][j] - Cstar[1][j]);
        
        //printf("Sector %d %10.1f %10.1f \n",i,incterm1,incterm2);
        consump = pt[j]*(incterm1 +incterm2);

        pce[j] = rhoadj(consump,pce[j],i);
        }

        // Spreading starts here, spread pcetotR weighted by 
        // relative important to income (consumption budget) level.
        coefsum = 0.;
        for(i = 1; i <= neq; i++)
                coefsum += coefs[i][2];

        pcesum = pce.sum();
        dif = pcetotR[t] - pcesum;
        dif = dif/coefsum;
        
        for(i = 1; i <= neq; i++)
                pce[i] += dif*coefs[i][2];
}

//=====================================================================
// Version of equation with weighted populations (popul), and
// cstar variables.  NO total pce used for spreading.

void Pads::padsf(Vector& pce, Vector& p, Vector& popul, Matrix& Cstar,
           Vector& tempi){

short i,j,k;
float lpbar, lpt, incterm1, incterm2, consump, cadj;

Vector lp(neq), pg(ng), lpg(ng), psg(nsg), lpsg(nsg), //pce prices
   wg(ng), wsg(nsg), //sum of shares (weights) in groups and subgroups.
        pt(neq); //price term of PADS.

/* Debug print coefficient
printf("From padsf t=%d:\n",t);
for(i = 1; i<=10;i++){
        printf("\n%2d %6.3f %6.3f ",i,share[i], lambda[i]);
        for(k = 1; k <= 4; k++)
                printf(" %8.3f",coefs[i][k]);
        }
*/

for(i = 1; i <= neq; i++){
        if(p[i] <= 0.) {
                printf("Negative price for pce category %d replaced by 1.\n",i);
                p[i] = 1.;
                }
        lp[i] = log(p[i]);
        }
lpbar = dot(lp,share);
lpg.set(0.); lpsg.set(0.), wg.set(0.), wsg.set(0.);
for(i = 1; i <= neq; i++){
        if(group[i] > 0){
                lpg[group[i]] += share[i]*lp[i];
                wg[group[i]] += share[i];
                }
        if(subgroup[i] > 0){
                lpsg[subgroup[i]] += share[i]*lp[i];
                wsg[subgroup[i]] += share[i];
                }
        }
for(i = 1; i <= ng; i++){
        lpg[i] = lpg[i]/wg[i];
        }
for(i = 1; i <= nsg; i++){
        lpsg[i] = lpsg[i]/wsg[i];
        }

// Compute full price term for each product:

for(i = 1; i <= neq; i++){
        j = sectors[i];
        lpt = 0.;
        if(insystem[i] == 1){
                lpt = -lambda[i]*(lp[i] - lpbar);
                if(group[i] > 0){
                        lpt += -mu[group[i]]*(lp[i] - lpg[group[i]]);
                        }
                if(subgroup[i] > 0){
                        lpt += -nu[subgroup[i]]*(lp[i] - lpsg[subgroup[i]]);
                        }
                for(k = 1; k <= neq; k++){
                        lpt += -lambda[k]*share[k]*(lp[i] - lp[k]);
                        }
                }
        lpt = exp(lpt);
        pt[j] = lpt;

        incterm1 = (coefs[i][1] + coefs[i][4]*tempi[j])*popul[j];

        incterm2 = coefs[i][2]*Cstar[0][j] + coefs[i][3]*(Cstar[0][j] - Cstar[1][j]);

        //printf("Sector %d %10.1f %10.1f \n",i,incterm1,incterm2);
        consump = pt[j]*(incterm1 +incterm2);

        pce[j] = rhoadj(consump,pce[j],i);
        }

}

#endif
