/*
  IDGROUPS.CPP -- Only group functions used in the Interdyme modeling
  framework are included in this file.  GROUPS.H has the full interface
  listed.
*/

#include "dymesys.h"

#if defined(INCL_GROUPS)

struct GBuf {
    size_t GDsize;
    char GroupTag[GroupHome::MAXTAGLEN+1];
    short Nin;
    short GroupElems[GroupHome::MAXSEC];
    } GBF;

/*==========================================================================*/
// Uses Nin field to return how many in group.  -1 signifies no group defined
short GroupHome::NinGroup(char *tag)  {
   short i;
   for(i=0;i<NGroups;i++) {
      if(strcmp(tag,gd[i].GroupTag)==0)
           return(gd[i].Nin);
      }
   return(-1);
}
/*==========================================================================*/
// Make sure to declare list big enough in calling program!
short GroupHome::GroupList(char *tag, short *list)  {
short count=0,i,k,n;

   for(i=0;i<NGroups;i++) {
       if(strcmp(tag,gd[i].GroupTag)==0)  {
           n=gd[i].Nin;
         for(k=0;k<n;k++)
            list[k]=gd[i].GroupElems[k];
         }
      }
   return count ;
   }
/*==========================================================================*/
GroupHome::GroupHome(void)
{ // Simple default constructor for global declaration.
    }
/*==========================================================================*/
void GroupHome::Load(void)  {

    short GroupNo, i,j,n, gds,err,ignore,type, lastgroup, group;
    char *cp,s[100],gString[110];
    short msize, *members;
    long pos;
    NGroups=GroupsRead=0;
    size_t rbyte;
    struct stat buffer;
    short Exists;

   err=stat("groups.bin",&buffer);
    Exists=(!err);  // err == 0 when file exists!
    if(!Exists) goto NewFile;

    // test here for file existence:
    if((fpbin=fopen("groups.bin","rb+"))==NULL)  {
        printf(msgcno,"groups.bin");
        return;
        }

    //printf("Constructor called.  Reading from GROUPS.BIN.\n");
    err=rdHeader();  // reads in NGroups, start, psn
    if(NGroups<=0) goto NewFile;

    mlist=new short[MAXSEC+1];    // combination of previously defined groups.
    msize=MAXSEC+1;   // size of member lists to be searched
    gd=new GroupDefn[MAXGROUP];  // Allocate space for group definition records.

    for(GroupNo=0; GroupNo<NGroups; GroupNo++)  {
        pos=psn[GroupNo];
        if((fseek(fpbin,pos,SEEK_SET)) != 0) {   /* posit at end of header */
            printf(msgbk15);
            return;
            }
        rbyte=psn[GroupNo+1]-psn[GroupNo];
        fread(&GBF.GDsize, sizeof(short), 1, fpbin);
        fread(&GBF.Nin, sizeof(short), 1, fpbin);
        n=GBF.Nin;
        if((err=fread(GBF.GroupElems, sizeof(short),n,fpbin))!=n){
            printf("Error reading group definition.\n");
            return;
            }
        // size of character string obtained by deduction
        rbyte=GBF.GDsize - (n+2)*sizeof(short);
        fread(GBF.GroupTag, sizeof(char), rbyte, fpbin);

        gd[GroupNo].GDsize=GBF.GDsize;
        gd[GroupNo].GroupTag=new char[strlen(GBF.GroupTag)+1];
        strcpy(gd[GroupNo].GroupTag,GBF.GroupTag);
        n=gd[GroupNo].Nin=GBF.Nin;
        gd[GroupNo].GroupElems=new short[n];
        for(i=0;i<n;i++)
            gd[GroupNo].GroupElems[i] = GBF.GroupElems[i];
        GroupsRead++;
        }
    msize=msize;  err=err;
    return;

    NewFile:
    NGroups=GroupsRead=0;
    // save separate GroupsRead variable for reading additional groups to an
    //   existing file.
    for(i=0;i<MAXGROUP;i++) psn[i]=0L;
    psn[0]=start=sizeof(short) + sizeof(long) +
            MAXGROUP * sizeof(long); // header space
    if((fpbin=fopen("groups.bin","wb+"))==NULL)  {
        printf(msgcno,"groups.bin");
        return;
    }
    mlist=new short[MAXSEC+1];    // combination of previously defined groups.
    msize=MAXSEC+1;   // size of member lists to be searched
    gd=new GroupDefn[MAXGROUP];  // Allocate space for group definition records.
    wrtHeader();
    return;
    }
/*========================================================================*/
GroupHome::~GroupHome() {
    short i;
    //printf("Destructor called.  Is Everything OK?\n");
   if(gd)
    delete[] gd;
   if(mlist)
    delete[] mlist;
    fclose(fpbin);
    }
/*========================================================================*/
// Write header information to binary file GROUPS.BIN: NGroups, start,
//    psn
short GroupHome::wrtHeader(void) {
    short err;
    fseek(fpbin,0L,SEEK_SET);
    fwrite(&NGroups,sizeof(short),1,fpbin);
    fwrite(&start,sizeof(long),1,fpbin);
    if((err=fwrite(psn,sizeof(long),MAXGROUP,fpbin))!=MAXGROUP)  {
        printf("Error writing header!  Help!\n");
        return(-1);
        }
    err=err;
    return 0;
    }
/*========================================================================*/
// Read header information from binary file GROUPS.BIN: NGroups, start,
//    psn
short GroupHome::rdHeader(void) {
short err;
fseek(fpbin,0L,SEEK_SET);
fread(&NGroups,sizeof(short),1,fpbin);
fread(&start,sizeof(long),1,fpbin);
    if((err=fread(psn,sizeof(long),MAXGROUP,fpbin))!=MAXGROUP)  {
        printf("Error reading header!  Help!\n");
        return(-1);
        }
    err=err;
    return 0;
    }
/* =========================================================================*/
GroupDefn::~GroupDefn() {
    if(GroupElems)
    delete[] GroupElems;
    if(GroupTag)
      delete[] GroupTag;
    }

#endif  // end groups section
