/* IDRUN.CPP --
	The member functions of the classes are included here, because
	if we included the entire class definition, all member functions
	would be linked in, making DYME.EXE too large.
*/

#include "dymesys.h"
#include "user.h"
short shiftback(short t);
short fixer(FILE *iin);
short putfix(FILE *iin, char kind);
#define TRUE 1
#define FALSE 0
#define OK  0

const float SMALL = 1.0e-7;

// Functions relating to matrices:
#if defined(INCL_MATRIX)

short vorm::unload(void) {
if(seq>=0 && seq<=matcount) {
	mats[seq].loaded='n';
	printf(" Unload called.  seq = %d  matcount=%d\n",seq,matcount);
	}
return 0;
}
//=======================================================================
// For declaring packed matrices by name in Interdyme -> version 
//		without VamFile specified.   Use default VamFile.
Pmatrix::Pmatrix(char *pmname, char readflag, char writeflag, char *pmfname)
{
	SetUpPmatrix(vf[0],pmname,readflag,writeflag,pmfname);
	}
//=======================================================================
// Version with VamFile specified.
Pmatrix::Pmatrix(VamFile& vf,char *pmname, char readflag,
	char writeflag, char *pmfname)
{
	SetUpPmatrix(vf,pmname,readflag,writeflag,pmfname);
	}

Pmatrix::r(char *pmname, char readflag, char writeflag, char *pmfname){
    SetUpPmatrix(vf[0],pmname,readflag,writeflag,pmfname);
	}
//=======================================================================
short Pmatrix::SetUpPmatrix(VamFile& vf,char *pmname, 
	char readflag, char writeflag, char *pmfname)
{
	FILE *fpop, *fpo;
	long lpos, fdsize;
	pmfid pmh;
	size_t pmhsize;
	char binname[80],wheretxt[60];
	jmp_buf this_point;
	char c;
	short lvn,err;

	if(!vf.IsOpen()) {
		cprintf("VamFile not opened in Pmatrix initialization.\r\n");
		return(ERR);
		}
	lvn = vf.findname(pmname);
	if(lvn<0) {
		printf("No such vector as %s.\n", pmname);
		return ERR;
		}
	mats[matcount].vfp=&vf;
	strcpy(mats[matcount].name,pmname);
	mats[matcount].rows = vf.Getnrows(lvn);
	mats[matcount].cols = vf.Getncols(lvn);
	c = vf.Gettype(lvn);
	// (c was type flag)
	if(c=='p' || c == 'P') {
		mats[matcount].lags = vf.GetPmatNo(pmname);
		mats[matcount].type = 'p';
		mats[matcount].pmfname=new char[80];
		mats[matcount].offset = vf.Getoffset(lvn);
		}
	else {
		cprintf("Error: You are trying to declare a Packed matrix: %s.\r\n",pmname);
		cprintf("  In the Vam file it is not a packed matrix!\r\n");
		tap();
		return(ERR);
		}
	mats[matcount].vptr = 0;
	mats[matcount].mptr = 0;
	mats[matcount].pmatptr = 0;

	if(strlen(pmfname) == 0){
		/*	This matrix not used. Probably in vam.cfg for future use.
			Setting read and write flags like this will prevent
			CheckDeclar from complaining and will insure that the
			matrix does no harm.
			*/
		mats[matcount].readflag = 'n';
		mats[matcount].writeflag = 'n';
		nelem = 0;
		return ERR;
		}

	if(strcmp(pmfname,"same")!=0 && strcmp(pmfname,"home")!=0 )
		strcpy(binname, pmfname);
	else { //If pmfname is "same", get the real name from the vam.bin file
		fdsize = 2*sizeof(float) + 2*sizeof(long) + 2*sizeof(short) +
			sizeof(unsigned short) + 80*sizeof(char);
		lpos = fdsize + mats[matcount].lags * 32;
		fpo = vf.Getfp();
		// Weird: Under linux, the next line is blotting out the string:
		err=fseek(fpo, lpos, SEEK_SET);
		if(err==-1)
			cprintf("Error seeking\n\r");
		fread(binname, 32, 1, fpo);
		}
	if((fpop = fopen(binname, "rb+"))==NULL){
		printf("Cannot open %s to define Packed Matrix.\n", binname);
		exit(1);
		}
	pmh.ReadIn(fpop);
	pmhsize = 80*sizeof(char)+9*sizeof(short)+3*sizeof(long)+sizeof(short*)
		+sizeof(short**)+sizeof(char); // now 119

	mats[matcount].readflag = readflag;
	mats[matcount].writeflag = writeflag;
	mats[matcount].NumElements = pmh.nelem;
	mats[matcount].LastDat = 0;
	nelem = pmh.nelem;
	if((nr = mats[matcount].rows)!=pmh.nr||(nc = mats[matcount].cols)!=pmh.nc){
		printf("P-matrix file and vam file do not match\n");
		exit(1);
		}
	fr = 1;
	fc = 1;
	mats[matcount].type = 'p';

	temp = 'r';
	lr = fr + nr - 1;
	lc = fc + nc - 1;
	sprintf(wheretxt,"Pmatrix row");
	if((row=new short*[nr])==0||(rnc=new short[nr])==0||
		(p= new float*[nr])==0)	outofmem(wheretxt);
	row -= fr;
	rnc -= fr;
	p -= fr;
	fread(&rnc[fr], 2, nr, fpop);
	short j;
	for(j=fr; j<=lr; j++){
		if(rnc[j]>0) {
			if((row[j]=new short[rnc[j]])==0||(p[j]=new float[rnc[j]])==0)
				outofmem(wheretxt);
			fread(row[j], 2, rnc[j], fpop);
			}
		else {
			rnc[j]=0;
			p[j]=0;
			row[j]=0;
			}
		}
	mats[matcount].pmatptr = this;
	seq=matcount;
	strcpy(mats[matcount].pmfname, binname);
	mats[matcount].loaded='y';
	matcount++;
	if(matcount>MAXMAT) {
		printf("Maximum number of matrices (%d) exceeded on %s.\n",
			matcount,mats[matcount].name);
		exit(1);
		}
	DymeLoaded='y';
	name = new char[strlen(pmname)+1];
	strcpy(name,pmname);
	which = vcount++;
	LastDat=0;
	fclose(fpop);
	return 0;
	}
//=======================================================================
// For declaring vectors by name in interdyme.  This has been rewritten
//	(dsm-10/18/94) to work the same as the one below, but take the
//	default VamFile, when none is specified.

Vector::Vector(char *name, char rf, char wf)
{	// for now, vf[0] is the global default VamFile.
	SetUpVector(vf[0],name,rf,wf);
	}
//=======================================================================
/* This version of the constructor for declaring vectors by
	name in Interdyme is distinguished by taking a
	Vamfile as its first argument.  By default, this is the read
	and write file for the Vector.  When we do a global load, or an
	explicit load, this VamFile will be used.
*/
Vector::Vector(VamFile& vf,char *name, char rf, char wf)
{
	SetUpVector(vf,name,rf,wf);
	}
void Vector::r(char *name, char rf, char wf){
    SetUpVector(vf[0],name,rf,wf);
    }
/*==========================================================================*/
/* SETUPVECTOR -- This function really does all the work for the
	Matrix, Vector and Pmatrix declarations.
*/
short Vector::SetUpVector(VamFile& vf,char *aname, char rf, char wf)
{
	char c;
	short lvn;
	// For this constructor, we don't use the mats array, but rather
	//	check the VamFile info to tell how to read.  Note: this function
	//	must be declared a friend of VamFile in vamfile.h.
	if(!vf.IsOpen()) {
		cprintf("VamFile not opened in Vector initialization.\r\n");
		return(ERR);
		}
	lvn = vf.findname(aname);
	if(lvn<0) {
		printf("No such vector as %s.\n", aname);
		return ERR;
		}
	nr=vf.Getnrows(lvn);
	nc=vf.Getncols(lvn);
	if(nr > 1 && nc > 1){
		printf("%s must be declared to be a matrix, not a vector.\n");
		return ERR;
		}
	// matcount is global. It keeps track of the highest element used
	//	in mats.
	mats[matcount].readflag = rf;
	mats[matcount].writeflag = wf;
	temp = 'n';
	fr = 1;
	fc = 1;
	lr = fr + nr - 1;
	lc = fc + nc - 1;
	nelem = nr + nc - 1;
	first = 1;
	if(nelem < 0 || nelem > 16000){
		cout << nelem << "is an illegal number of elements for a vector.\n";
		return ERR;
		}
	if(nelem == 0)
		v = 0;
	else if((v = new float[nelem]) == 0){
		printf("Out of memory allocating %s.\n",name);
		return ERR;
		}
	// Set up the information on this vector
	mats[matcount].vfp=&vf;
	strcpy(mats[matcount].name,aname);
	name = new char[strlen(aname)+1];
	strcpy(name,aname);
	mats[matcount].rows = vf.Getnrows(lvn);
	mats[matcount].cols = vf.Getncols(lvn);
	c = vf.Gettype(lvn);
	// (c was type flag)
	#if defined(INCL_FIX)
	if(strcmp(mats[matcount].name,"fix")==0)
		fixesPlace = matcount;
	#endif
	if(c=='p' || c == 'P') {
		cprintf("Error: You are trying to declare a vector: %s.\r\n",aname);
		cprintf("  In the Vam file it is a packed matrix!\r\n");
		tap();
		return(ERR);
		}
	mats[matcount].lags = vf.Getnlags(lvn);
	mats[matcount].NumElements = mats[matcount].rows*mats[matcount].cols;
	mats[matcount].offset = vf.Getoffset(lvn);
	mats[matcount].type = 'v';
	mats[matcount].mptr = 0;
	mats[matcount].vptr = v;
	mats[matcount].vecptr = this;
	mats[matcount].pmatptr = 0;
	seq = matcount;  // This now points to the index in the *mats* array.
	mats[matcount].loaded='y';
	mats[matcount].LastDat = 0;
	matcount++;
	if(matcount>MAXMAT) {
		printf("Maximum number of matrices (%d) exceeded on %s.\n",
			matcount,mats[matcount].name);
		exit(1);
		}
	DymeLoaded='y';
	name = new char[strlen(aname)+1];
	strcpy(name,aname);
	LastDat=0;
	which = vcount++;
	return 0;
	}
/*==========================================================================*/
// For declaring matrices by name in Interdyme -> Version without VamFile
//	specified.  Use default.
Matrix::Matrix(char *name, char rf, char wf)
{
	SetUpMatrix(vf[0],name,rf,wf);
	}
/*==========================================================================*/
// Version with VamFile specified.
Matrix::Matrix(VamFile& vf,char *name, char rf, char wf)
{
	SetUpMatrix(vf,name,rf,wf);
	}
/*==========================================================================*/
void Matrix::r(char *name, char rf, char wf)
{
	SetUpMatrix(vf[0],name,rf,wf);
	}
/*==========================================================================*/


short Matrix::SetUpMatrix(VamFile& vf,char *aname, char rf, char wf)
{
	char c,l;
	short lvn;
	short i;
	if(!vf.IsOpen()) {
		cprintf("VamFile not opened in Vector initialization.\r\n");
		return(ERR);
		}
	lvn = vf.findname(aname);
	if(lvn<0) {
		printf("No such matrix as %s.\n", aname);
		return ERR;
		}

	c = vf.Gettype(lvn);
	// c is the type flag
	l = vf.Getnlags(lvn);
	if(l==0 && c == 'v') {
		cprintf("The vector %s does not have any lags!  Cannot be declared as a matrix.\n\r",
			aname);
		cprintf("Change the value in the lags column and then remake your HIST.VAM file.\n\r");
		tap();
		}
	// matcount is global. It keeps track of the highest element used
	//	in mats.
	mats[matcount].readflag = rf;
	mats[matcount].writeflag = wf;
	temp = 'n';
	fr = 1;
	fc = 1;
	lr = fr + nr - 1;
	lc = fc + nc - 1;

	// Set up the information on this matrix.
	mats[matcount].vfp=&vf;
	strcpy(mats[matcount].name,aname);
	mats[matcount].rows = vf.Getnrows(lvn);
	mats[matcount].cols = vf.Getncols(lvn);
	mats[matcount].lags = vf.Getnlags(lvn);
	if(mats[matcount].lags >0){
		nr = mats[matcount].lags +1;
		nc = max(mats[matcount].rows, mats[matcount].cols);
		fc = 1;
		fr = 0;
		lr = fr + nr - 1;
		lc = fc + nc - 1;
		}
	if(c=='p' || c == 'P') {
		cprintf("Error: You are trying to declare a matrix: %s.\r\n",aname);
		cprintf("  In the Vam file it is a packed matrix!\r\n");
		tap();
		return(ERR);
		}
	mats[matcount].NumElements = mats[matcount].rows*mats[matcount].cols;
	mats[matcount].offset = vf.Getoffset(lvn);
	mats[matcount].type = 'm';
	mats[matcount].vptr = 0;
	mats[matcount].mptr = 0;
	mats[matcount].pmatptr = 0;

	if(mats[matcount].lags > 0){
		 /* Store lagged values of a vector as rows of a matrix.  Current
			 value is row 0, so there are 1 more rows than lagged values. */
		if(mats[matcount].rows > 1 && mats[matcount].cols >1){
			printf("Not prepared for lagged matrices like %s.\n",aname);
			return ERR;
			}
		mats[matcount].type = 'l';
		mats[matcount].loaded = 'y';
		// Set the vptr of this mat to the vptr of the vector as vector.
		// This is used for shifting back the matrix in the store function.
		// The declaration as a vector must preceed the declaration as a matrix.
		for( i = 0;i < matcount;i++){
			if(strcmp(aname, mats[i].name) == 0) break;
			}
		if(i==matcount){
			printf("The declaration of the lagged values of %s as a matrix must come after\n",name);
			printf("   the declaration as a vector.\n");
			exit(1);
			}
		else mats[matcount].vptr = mats[i].vptr;
		}
	else{
		nr = mats[matcount].rows;
		nc = mats[matcount].cols;
		fr = 1;
		fc = 1;
		}
	lr = fr + nr - 1;
	lc = fc + nc - 1;
	if((m = new float* [nr]) == 0)
		goto gripe;
	m -= fr;
	LastDat = 0;
	mats[matcount].LastDat = 0;
	mats[matcount].mptr = m;
	mats[matcount].matptr = this;
	seq = matcount;
	for(i = fr; i <= lr; i++){
		if((m[i] = new float [nc]) == 0) goto gripe;
		m[i] -= fc;
		}
	mats[matcount].loaded='y';
	matcount++;  // global count of matrices and variables assigned.
	DymeLoaded='y';
	name = new char[strlen(aname)+1];
	strcpy(name,aname);
	return 0;
	gripe:
		printf("Out of memory assigning %s.\n",name);
		exit(1);
		return(-1); // to suppress warning
	}
/**************************************************************************/
short Matrix::load(short year){
	short i;
	unsigned long locinvam;
	VamFile *vfp = mats[seq].vfp;
	FILE* fpo = vfp->Getfp();

	// seq is the position of this Matrix in the mats[] array.
	locinvam = (long)vfp->Getstartpsn() +
		(long)vfp->Getnnpy() *
		((long)year -
		(long)vfp->Getstartdate())*4L +
		(long)mats[seq].offset*4L;

	fseek(fpo,locinvam,SEEK_SET);
	for(i = 1; i <= nr; i++){
		if(fread(&m[i][1],4,nc,fpo) != nc){
			printf("Error reading matrix %d.\n",i);
			return(ERR);
			}
		}
	return(OK);
	}
/**************************************************************************/
short Matrix::load(VamFile& vf, char* vname, short year){
	short i;
	unsigned long locinvam,offset;
	VamFile *vfp = &vf;
	FILE* fpo = vfp->Getfp();
	int lvn,vnr,vnc;

	lvn = vf.findname(vname);
	if(lvn<0) {
		printf("No such matrix as %s.\n", vname);
		return ERR;
		}
	vnr = vf.Getnrows(lvn);
	vnc = vf.Getncols(lvn);
	if(vnr != nr || vnc != nc) {
		printf("Matrix %s has wrong dimensions!\n",vname);
		printf("Rows = %d.  Columns = %d.  Declared (%d,%d).\n",vnr,vnc,nr,nc);
		return ERR;
		}

	offset = vf.Getoffset(lvn);
	locinvam = (long)vfp->Getstartpsn() +
		(long)vfp->Getnnpy() *
		((long)year -
		(long)vfp->Getstartdate())*4L +
		(long)offset*4L;

	fseek(fpo,locinvam,SEEK_SET);
	for(i = 1; i <= nr; i++){
		if(fread(&m[i][1],4,nc,fpo) != nc){
			printf("Error reading matrix %d.\n",i);
			return(ERR);
			}
		}
	return(OK);
	}
/**************************************************************************/
short Matrix::store(VamFile& vf, char* vname, short year){
	short i;
	unsigned long locinvam,offset;
	VamFile *vfp = &vf;
	FILE* fpo = vfp->Getfp();
	int lvn,vnr,vnc;

	lvn = vf.findname(vname);
	if(lvn<0) {
		printf("No such matrix as %s.\n", vname);
		return ERR;
		}
	vnr = vf.Getnrows(lvn);
	vnc = vf.Getncols(lvn);
	if(vnr != nr || vnc != nc) {
		printf("Matrix %s has wrong dimensions!\n",vname);
		printf("Rows = %d.  Columns = %d.  Declared (%d,%d).\n",vnr,vnc,nr,nc);
		return ERR;
		}

	offset = vf.Getoffset(lvn);
	locinvam = (long)vfp->Getstartpsn() +
		(long)vfp->Getnnpy() *
		((long)year -
		(long)vfp->Getstartdate())*4L +
		(long)offset*4L;

	fseek(fpo,locinvam,SEEK_SET);
	for(i = 1; i <= nr; i++){
		if(fwrite(&m[i][1],4,nc,fpo) != nc){
			printf("Error writing matrix %d.\n",i);
			return(ERR);
			}
		}
	return(OK);
	}
/**************************************************************************/
short Matrix::store(short year){
	short i;
	unsigned long locinvam,offset;

	VamFile *vfp = mats[seq].vfp;
	FILE* fpo = vfp->Getfp();

	int lvn,vnr,vnc;

	locinvam = (long)vfp->Getstartpsn() +
	(long)vfp->Getnnpy()*( (long)year -
	(long)vfp->Getstartdate())*4L +
		mats[seq].offset*4;

	fseek(fpo,locinvam,SEEK_SET);
	for(i = 1; i <= nr; i++){
		if(fwrite(&m[i][1],4,nc,fpo) != nc){
			printf("Error writing matrix %d.\n",i);
			return(ERR);
			}
		}
	return(OK);
	}

//=======================================================================
short Vector::load(short year){
	short i;
	unsigned long locinvam;
	VamFile *vfp = mats[seq].vfp;
	FILE* fpo = vfp->Getfp();

	locinvam = (long)vfp->Getstartpsn() +
	(long)vfp->Getnnpy()*( (long)year -
	(long)vfp->Getstartdate())*4L +
		mats[seq].offset*4;
	fseek(fpo,locinvam,SEEK_SET);
	if(fread(v,4,nelem,fpo) != nelem){
			printf("Error reading vector.\n");
			return(ERR);
			}
	return(OK);
	}
//=======================================================================
short Vector::load(VamFile& vf, char* vname, short year){
	short i;
	unsigned long locinvam,offset;
	VamFile *vfp = &vf;
	FILE* fpo = vfp->Getfp();
	short lvn,vnr,vnc;

	lvn = vf.findname(vname);
	if(lvn<0) {
		printf("No such vector as %s.\n", vname);
		return ERR;
		}
	vnr = vf.Getnrows(lvn);
	vnc = vf.Getncols(lvn);
	if(vnr != nr || vnc != nc) {
		printf("Vector %s has wrong dimensions!\n",vname);
		printf("Rows = %d.  Columns = %d.  Declared (%d,%d).\n",vnr,vnc,nr,nc);
		return ERR;
		}

	offset = vf.Getoffset(lvn);
	locinvam = (long)vfp->Getstartpsn() + 
	(long)vfp->Getnnpy()*( (long)year - 
	(long)vfp->Getstartdate())*4L + 
		(long)offset*4L;
	fseek(fpo,locinvam,SEEK_SET);
	if(fread(v,4,nelem,fpo) != nelem){
			printf("Error reading vector.\n");
			return(ERR);
			}
	return(OK);
	}
//=======================================================================
short Vector::store(VamFile& vf, char* vname, short year){
	short i;
	unsigned long locinvam,offset;
	VamFile *vfp = &vf;
	FILE* fpo = vfp->Getfp();
	short lvn,vnr,vnc;

	lvn = vf.findname(vname);
	if(lvn<0) {
		printf("No such vector as %s.\n", vname);
		return ERR;
		}
	vnr = vf.Getnrows(lvn);
	vnc = vf.Getncols(lvn);
	if(vnr != nr || vnc != nc) {
		printf("Vector %s has wrong dimensions!\n",vname);
		printf("Rows = %d.  Columns = %d.  Declared (%d,%d).\n",vnr,vnc,nr,nc);
		return ERR;
		}

	offset = vf.Getoffset(lvn);
	locinvam = (long)vfp->Getstartpsn() +
	(long)vfp->Getnnpy()*( (long)year -
	(long)vfp->Getstartdate())*4L +
		(long)offset*4L;
	fseek(fpo,locinvam,SEEK_SET);
	if(fwrite(v,4,nelem,fpo) != nelem){
			printf("Error reading vector.\n");
			return(ERR);
			}
	return(OK);
	}
//=======================================================================

/* load(short yr)

	Exclusive use for dyme
	Load pmatrix "pm" for year "yr" from file "fin".
	Return OK when succeeds or ERR if fails.
*/
short Pmatrix::load(short yr)
{
	return(LoadP(mats[seq].pmfname, yr));
	}
/* Store Pmatrix assuming default file name */
short Pmatrix::store(short yr) {
	return(store(mats[seq].pmfname,yr));
}

//==========================================================================
/* load() function for Interdyme

Load all of the vectors and matrices read from VAM.
Whether or not to read a vector or matrix is controlled by its readflag
in conjunction with the useall flag set by a question at the beginning of
the run.  The useall Question is "Do you want to use all available data?"
Answer n if running a historical simulation.  Answer y if running a forecast
when some, but not all, data is available beyond the start of the run.

If readflag is:
  n  never read the vector
  y  always read the vector (default)
  i  read only in the initial year
  a  read if the data is available and useall is y.

Try this: The idea behind using the 'a' flag is that you may want to keep
the last year's solution value in the vector as a starting value for the
next year.  This is a problem when you want to restart the model in a future
year, after the last data has run out.  In this case, you may want to
explicitly load the lagged data for that vector.

At the beginning of the run, load also the lagged values of vectors requiring
lagged values.
*/

short load(short t)
{
	short i,j,k,loadt;
	unsigned short nelem;
	FILE *fpop; // For packed matrix file
	long locinvam; // Location in vam of what we want to read.
	long ppos; /*	Location of nonzero coefficients for year t
					in packed matrix file. */
	size_t rcount;  // number of bytes read.
	#if defined(INCL_EQFIX)
	Interval = GDate(t,1,1) - BaseDate + 1;
	#endif
    FILE* fpo;
    VamFile *vfp;
	for(k = 0; k < matcount; k++){
		loadt = t;
		if(mats[k].loaded=='n')
			continue;
		vfp = mats[k].vfp;
		fpo = vfp->Getfp();
        if(fpo == 0)
            printf("fpo is 0\n");

		if(mats[k].readflag == 'n') continue;
		if(mats[k].readflag == 'i' && t != godate) continue;
		if(mats[k].readflag == 'a'){
			if(useall == 'n' && t > godate) continue;
			else if(t == godate && t > mats[k].LastDat)
				loadt = t-1;
			else if(t > mats[k].LastDat) continue;
			}
		locinvam = (long)vfp->Getstartpsn() +
			(long)vfp->Getnnpy()*
			(loadt - (long)vfp->Getstartdate())*4L +
			mats[k].offset*4L;
		fseek(fpo,locinvam,SEEK_SET);
		// Read vectors
		if(mats[k].type == 'v'){
			nelem = mats[k].NumElements;
			rcount = fread(mats[k].vptr, 4, nelem, fpo);
			if(rcount != nelem){
				printf("Error reading vector %d.\n",k);
				return(ERR);
				}
			}
		// Read matrices
		else if(mats[k].type == 'm'){
			nelem = mats[k].cols;
			for(i = 1; i <= mats[k].rows; i++) {
				rcount = fread(&mats[k].mptr[i][1],4,nelem,fpo);
				if(rcount != nelem){
					printf("Error reading matrix %d.\n",i);
					return(ERR);
					}
				}
			}
		else if(mats[k].type == 'p'){
			if(mats[k].pmatptr->LoadP(mats[k].pmfname, t)==ERR){
				printf("Error reading packed matrix file %s\n");
				exit(1);
				}
			}

		if(mats[k].type != 'l') continue;
        // Load lagged values of a vector into the matrix that stores them
        if(t > godate) continue ;
		nelem = mats[k].NumElements;
		for(i = 0; i <= mats[k].lags; i++){
			locinvam = (long)vfp->Getstartpsn() +
				(long)vfp->Getnnpy()*
				(long)(loadt - i - vfp->Getstartdate())*4L +
				long(mats[k].offset*4);
			fseek(fpo,locinvam,0);
			rcount = fread(&mats[k].mptr[i][1], 4, nelem, fpo);
			if(rcount != nelem){
				printf("Error reading lagged values of vector %d.\n",i);
				return(ERR);
				}
			}
		}

	return(OK);
	}
/* Store() ***********************************************************
Store all of the vectors and matrices back into VAM.  Then shift all
the vectors having lagged values.
*/
short store(short t)
{
	short i,j,k,lag;
	unsigned short nelem;
	long locinvam; // Location in vam
	for(k = 0; k < matcount; k++){
		VamFile *vfp = mats[k].vfp;
		FILE* fpo = vfp->Getfp();
		// If this is a matrix of lagged values of a vector, shift them back.
		if(mats[k].type == 'l'){
			nelem = max(mats[k].rows,mats[k].cols);
			for(lag = mats[k].lags; lag >= 2; lag--){
				for(i = 1; i <= nelem; i++)
					mats[k].mptr[lag][i] = mats[k].mptr[lag-1][i];
				}
			// Put the vector version into both the lagged once position
			// and the no-lag position of the matrix.
			for(i = 1; i <= nelem; i++){
				mats[k].mptr[1][i] = mats[k].vptr[i-1];
				mats[k].mptr[0][i] = mats[k].vptr[i-1];
				}
			}
		if(mats[k].writeflag == 'n') continue;
		locinvam = (long)vfp->Getstartpsn() +
			(long)vfp->Getnnpy()*
			(t - (long)vfp->Getstartdate())*4L +
			mats[k].offset*4;
		fseek(fpo,locinvam,SEEK_SET);
		if(mats[k].type == 'v'){
			nelem = mats[k].NumElements;
			if(fwrite(mats[k].vptr, 4, nelem, fpo) != nelem){
				printf("Error writing vector %d.\n",i);
				return(ERR);
				}
			}
		if(mats[k].type == 'm'){
			nelem = mats[k].cols;
			for(i = 1; i <= mats[k].rows; i++)
				if(fwrite(&mats[k].mptr[i][1],4,nelem,fpo) != nelem){
					printf("Error writing matrix %d.\n",i);
					return(ERR);
					}
			}

		if(mats[k].type == 'p'){
			if(mats[k].pmatptr->store(mats[k].pmfname, t)==ERR){
				printf("Error writing packed matrix %s.\n",mats[k].pmfname);
				return(ERR);
				}
			}

		}
	return(OK);
	}
/* Shiftback() **************************************
This function shifts down the rows of the matrices whose rows are lagged
values of vectors.  This action is also accomplished in the store() function,
but it is given here as a separate function for use in spin() when optimizing
or doing stochastic simulation, because in those cases the vam file should
not be modified.

*/


short shiftback(short t){
	short i,j,k,lag;
	unsigned short nelem;
	long locinvam; // Location in vam
    //VamFile *vfp;
    //FILE *fpo;
    Vector x(9);
	for(k = 0; k < matcount; k++){
		VamFile *vfp = mats[k].vfp;
		FILE *fpo = vfp->Getfp();
		// If this is a matrix of lagged values of a vector, shift them back.
		if(mats[k].type == 'l'){
			nelem = max(mats[k].rows,mats[k].cols);
            /* debug
            if(strcmp(mats[k].name,"p") == 0){
                lag = mats[k].lags;
                printf("\nBefore shift\n");
                for(i=0; i <= lag; i++){
                    for(j = 1; j <= 9; j++)
                        x[j] = mats[k].mptr[i][j];
                    x.Display(mats[k].name, 7, 0);
                    }
                }
            // end debug */
			for(lag = mats[k].lags; lag >= 2; lag--){

				for(i = 1; i <= nelem; i++){
					mats[k].mptr[lag][i] = mats[k].mptr[lag-1][i];
                    }
 				}
			// Put the vector version into both the lagged once position
			// and the no-lag position of the matrix.
			for(i = 1; i <= nelem; i++){
				mats[k].mptr[1][i] = mats[k].vptr[i-1];
				mats[k].mptr[0][i] = mats[k].vptr[i-1];
				}
            /* debug
            if(strcmp(mats[k].name,"pdm") == 0){
                lag = mats[k].lags;
                printf("\nAfter shift\n");
                for(i = 0; i <= lag; i++){
                    for(j = 1; j <= 9; j++)
                        x[j] = mats[k].mptr[i][j];
                    x.Display(mats[k].name, 7, 0);
                    }
                continue;
                }
             // end debug */
			}
		if(mats[k].writeflag == 'n') continue;
        continue;
		}
	return(OK);
	}
/****************************************************************/

/*=========================================================================*/

/* Put the current values of Vector a into row 0 of Matrix B */
double update(Vector& a, Matrix& B)
{
	short n,i;
	double dif,sum;
	n = a.numelm();
	if(n > B.columns()){
		printf("Vector too big for matrix.\n");
		return(-1.);
		}
	sum = 0;
	for(i = 1; i <= n; i++){
		//dif = a[i] - B[0][i];
		//sum += dif*dif;
		B[0][i] = a[i];
		}
	//sum = sqrt(sum/n);
	return(sum);
	}
/* CheckDeclar() ******************************************************
Check that all objects in vam have been declared in dyme or marked
to be neither read nor written.  If silence option is set, we assume
that you know what you're doing.
*/
short CheckDeclar(short silence)
{
	char c;
	short i, error;

	error = 0;
	for(i = 0; i < matcount; i++){
		if(mats[i].type=='m' && mats[i].mptr == 0){
			if(mats[i].readflag == 'n' && mats[i].writeflag == 'n')
				continue; // nothing is being done with this matrix.
			if(!silence) {
				printf("Matrix %s in Vam has not been declared in Dyme.\n",
					mats[i].name);
				printf("It will be neither read nor written by Dyme.\n");
				}
			mats[i].readflag = 'n';
			mats[i].writeflag = 'n';
			error++;
			}
		if(mats[i].type=='p' && mats[i].pmatptr == 0){
			if(mats[i].readflag == 'n' && mats[i].writeflag == 'n') 
				continue; // nothing is being done with this matrix.
			if(!silence) {
				printf("Packed matrix %s in Vam has not been declared in Dyme.\n",
					mats[i].name);
				printf("It will be neither read nor written by Dyme.\n");
				}
			mats[i].readflag = 'n';
			mats[i].writeflag = 'n';
			error++;
			}
		if(mats[i].type == 'v' && mats[i].vptr == 0){
			if(!silence) {
				printf("Vector %s in Vam has not been declared in Dyme.\n",
					mats[i].name);
				printf("It will be neither read nor written by Dyme.\n");
				}
			mats[i].readflag = 'n';
			mats[i].writeflag = 'n';
			error++;
			}
		if(mats[i].type == 'l' && mats[i].vptr == 0 && mats[i].mptr == 0){
			if(!silence) {
				printf("The Vector with lagged values, %s, in Vam \n",mats[i].name);
				printf("has not been declared as a vector in Dyme.\n");
				printf("It will be neither read nor written by Dyme.\n");
				}
			mats[i].readflag = 'n';
			mats[i].writeflag = 'n';
			error++;
			}
		}
	if(error > 0){
		if(!silence) {
			query: printf("Do you wish to continue? (y or n): ");
			c = getch();
			printf("\n");
			if (c == 'y' || c == 'Y') return(error);
			if (c == 'n' || c == 'N') exit (1);
			goto query;
			}
		}
	return(0);
	}
/**************************************************************************/
/* Multiply each row of matrix m by the corresponding element of vector v.*/
void  rowscale(Vector& v, Matrix& m, char skipdiag)
{
	short i,j,nr,nc;
	float x;

	nr = m.rows();
	nc = m.columns();
	for(i = 1; i <= nr; i++){
		x = v[i];
		for(j = 1; j <= nc; j++){
			if(skipdiag == 'n' || i!=j) m[i][j] = x*m[i][j];
			}
		}
	}

/**************************************************************************/
/* Scale the columns of matrix m to be equal to one. */
void  colnorm(Matrix& m,Vector& scalefactor )
{
	short i,j,nr,nc;
	double  sum;

	nr = m.rows();
	nc = m.columns();
	for(j = 1; j <= nc; j++){
		sum = 0.;
		for(i = 1; i <= nr; i++){
			sum+=m[i][j];
			}
		if(fabs(sum)<1.e-10){
			printf("Warnning: Divided by zero in colnorm\n");
			continue;
			}
		sum = 1./sum;
		scalefactor[j] *= sum;
		for(i = 1; i <= nr; i++)
			m[i][j] *= sum;
		}
	}

/**************************************************************************/
/* Multiply each row of packed matrix pm by the corresponding element of vector v.*/
void  rowscale(Vector& v, Pmatrix& pm, char skipdiag)
{
	short i,j,nr,nc;
	float x;

	nr = pm.rows();
	nc = pm.columns();
	for(i = 1; i <= nr; i++){
		x = v[i];
		for(j = 1; j <= nc; j++){
			if(skipdiag == 'n' || i!=j) pm.SetAij(i, j, x*pm(i,j));
			}
		}
	}

/**************************************************************************/
/* Scale the columns of matrix m to be equal to one. */
void  colnorm(Pmatrix& pm,Vector& scalefactor )
{
	short i,j,nr,nc;
	double  sum;

	nr = pm.rows();
	nc = pm.columns();
	for(j = 1; j <= nc; j++){
		sum = 0.;
		for(i = 1; i <= nr; i++){
			sum+=pm(i,j);
			}
		if(fabs(sum)<1.e-10){
			printf("Warnning: Divided by zero in colnorm\n");
			continue;
			}
		sum = 1./sum;
		scalefactor[j] *= sum;
		for(i = 1; i <= nr; i++)
			pm.SetAij(i, j, sum*pm(i, j));
		}
	}
/**************************************************************************/
short lastdata(void){
	FILE *fpld;
	char found,line[250],name[41];
	short k,date,err;

	if((fpld = fopen("lastdata","r"))==NULL) {
		cprintf("The file lastdata was not opened.  Create this file if you are \r\n");
		cprintf(" going to call the lastdata() function.\r\n");
		tap();
		return(-1);
		}
	while(fgets(line,249,fpld) != NULL){
		sscanf(line,"%s %d",name,&date);
		found = 'n';
		for(k = 0;k < matcount; k++){
			if(strcmp(name,mats[k].name)==0){
				mats[k].LastDat = date;
				if(mats[k].type == 'm') {
					if(! mats[k].matptr) {
						cprintf("\r\nMatrix %s never declared.\r\n  Cannot apply lastdata.  Exiting...\n",
							name);
						exit(1);
						}
					mats[k].matptr->LastDat = date;
					}
				else if(mats[k].type == 'v' || mats[k].type == 'l') {
					if(! (mats[k].vecptr || mats[k].matptr) ) {
						cprintf("\r\nVector %s never declared.\r\n  Cannot apply lastdata.  Exiting...\n",
							name);
						exit(1);
						}
					mats[k].vecptr->LastDat = date;
					}
				else if(mats[k].type =='p')  {
					if(! mats[k].pmatptr) {
						cprintf("\r\nPacked Matrix %s never declared.\r\n  Cannot apply lastdata.  Exiting...\n",
							name);
						exit(1);
						}
					mats[k].pmatptr->LastDat = date;
					}
				else printf("Unrecognized type %c for %s.\n",mats[k].type,name);
				found = 'y';
				break;
				}
			}
		if(found == 'n'){
			printf("Cannot set Last Date for %s.\n",name);
			tap();
			}
		}
	fclose(fpld);
	err = 0;
	for(k=0;k<matcount;k++){
		if(mats[k].readflag == 'a' && mats[k].LastDat == 0){
			printf("%s has a read flag of 'a' but no date of last available data.\n",
				mats[k].name);
			err++;
			}
		}
	if(err > 0) tap();
	return(err);
	}
//////////////////////////////////////////////////////////////////////////////
/* scale() -- scale the vector to the right total by right direction scaling
	This is handled like an "all" group fix.
	*/
double Vector::scale(float control)
{
	short i, allgroup[MAXSEC];
	float *v1;

	for(i=0; i < nelem; i++)
		allgroup[i] = i+1;

	v1 = v - 1;
	return rdscale(v1,nelem,allgroup,control);
	}
#endif  // end matrix-vector section

//===============================================================================
/* rdscale() -- right direction scaling *****************************
	v1 is a pointer to floats such that v1[1] is the first element.
	n is the number of elements in the group
	group is a pointer to n integers, the sector numbers entering
		into the scaling
	control is what they are being scaled to equal.
	This routine does no range checking of the sector numbers.
	*/

double rdscale(float *v1, short n, short *group, float control,
	short ScaleNegatives)
{
	float psumma=0.0; // sum of positive elements
	float nsumma=0.0; // sum of negative elements
	float fac=0.0;		// scaling factor
//	double psumma=0.0; // sum of positive elements
//	double nsumma=0.0; // sum of negative elements
//	double fac=0.0;		// scaling factor
	float x=0.0;
	short i, j, allzeroes=TRUE;



	psumma = nsumma = 0.;
    // n is number of elements in group
	for(i = 0; i < n; i++){
		x = v1[group[i]];
		if(x > 0) psumma += x;
		else nsumma += x;
		if(fabs(x)>=.0000001) allzeroes=FALSE;
		}
	if(allzeroes && fabs(control)>.000001)
		return(0);  // do nothing if we have no data to scale.
    // following avoids scaling in a case where the fix is almost certainly
    // erroneous  CA debug
    if((fabs(control) <= 0) && (psumma > 0) && (fabs(nsumma) <= 0))
        return(0);

	if(ScaleNegatives) {
		// Quadratic formula
		if(psumma > 0.){
			fac = control; // To get into double
			fac = fac*fac - psumma*nsumma*4.; // discriminant
			fac = safesqrt(fac);
			fac = (fac + control)/(psumma*2);
			}
		else if(fabs(control) >= .000001)
			fac = nsumma/control;
		else fac = 0.;
		}
	else {
		control -= nsumma;
		fac = control/psumma;
		}

	for(i = 0; i < n; i++){
		j = group[i];
		if(v1[j ] > 0.) v1[j] *= fac;
		else if(ScaleNegatives) {
			// Next line is necessary when both control and sum are 0.
			if(fabs(fac)<=.000001) v1[j]=0.0; // prevent divide by 0.
			else v1[j] /= fac;
			}
		arith(" in rdscale.",i+1);
		}
	return(fac);
	}
/* rdscale() -- right direction scaling *****************************
	v1 is a pointer to floats such that v1[1] is the first element.
	n is the number of elements in the group
	group is a pointer to n integers, the sector numbers entering
		into the scaling
	control is what they are being scaled to equal.
	This routine does no range checking of the sector numbers.
	*/

/* ============================================================== */
// This alternate version of rdscale takes a Vector& argument, instead
//	 of float*.
double rdscale(Vector& v1, short n, short *group, float control,
	short ScaleNegatives)
{
	double psumma; // sum of positive elements
	double nsumma; // sum of negative elements
	double fac;		// scaling factor
	double z;
	float x;
	short i, j, allzeroes=TRUE;

	arith(" In rdscale location C",0);
	psumma = nsumma = z = 0.;
	for(i = 0; i < n; i++){
		j = group[i];
		x = v1[j];
		z = x;
		arith(" In rdscale location A",i);
		if(x > 0.) psumma = psumma + z;
		else nsumma = nsumma + z;
//		printf("psumma = %lf  z = %lf\n",psumma,z);
		arith(" In rdscale location B",i);
		if(fabs(z)>=.0000001) allzeroes=FALSE;
		}
	arith(" in rdscale, location 0",0);
	// Quadratic formula
	if(allzeroes && fabs(control)>.000001)
		return(0);  // do nothing if we have no data to scale.
	if(psumma > 0.){
		fac = control; // To get into double
		arith(" In rdscale location 1",0);
		fac = fac*fac - psumma*nsumma*4.; // descriminant
		arith(" In rdscale location 2",0);
		fac = safesqrt(fac);
		arith(" In rdscale location 3",0);
		fac = (fac + control)/(psumma*2);
		arith(" In rdscale location 4",0);
		}

	else if(fabs(control) >= .000001)
		fac = nsumma/control;
	else fac = 0.;
	arith(" In rdscale location 5",0);
	for(i = 0; i < n; i++){
		j = group[i];
		if(v1[j ] > 0.) {
			v1[j] = v1[j] * fac;
			arith(" In rdscale location 6",j);
			}
		else {
			if(fabs(fac)<=.000001) v1[j]=0.0; // prevent divide by 0.
			else v1[j] = v1[j] / fac;
			arith(" In rdscale location 7",j);
			}
		}
	ScaleNegatives = ScaleNegatives;
	return(fac);
	}
//==============================================================================
/* pscale() -- proportional scaling
	vec is a pointer to the elements of the vector to be scaled.
	n is the number of elements in the vector.
	group is a vector of 1s and 0s that indicate which elements are
	to be included in the scaling.
	control is the value to which the sum of the elements is to be scaled.
*/


double pscale(Vector &vec, short n, short *group, float control)
{
   double sfac=0;
	float sum=0.0;		// sum of elements
	float absum=0.0;	// sum of alsolute values of  elements
	short i, j;
	int allzeroes=1, nopos=1, noneg=1;

	for(i=1;i<=n;i++) {
		if(group[i-1]) {
			if(vec[i]>SMALL) {
				allzeroes = 0;
				nopos = 0;
				}
			else if(vec[i] < -SMALL) {
				allzeroes = 0;
				noneg = 0;
				}
			sum += vec[i];
			absum += fabs(vec[i]);
			}
		}
	if(fabs(control-sum) < SMALL)
		return(0.);
	else if(allzeroes) {
		if(fabs(control) > SMALL)
			return(0.);
		else {
			printf("The function should not reach this point\n");
			return(0.);
			}
		}
	else if(noneg || nopos) {
		// Calculated ordinary scaling factor
		if(fabs(sum) > SMALL)
			sfac = control / sum;
		else
			sfac = 0.;
		for(i=1;i<=n;i++) {
			if(group[i-1])
				vec[i] = sfac*vec[i];
			}
		return(sfac);
		}
	else {
		// Calculate proportional scaling factor
		if(fabs(absum)>SMALL)
			sfac = (control - sum) / absum;
		else
			sfac = 0.;
		for(i=1;i<=n;i++) {
			if(group[i-1])
				vec[i] += sfac*fabs(vec[i]);
			}
		return(sfac);
		}
}
