// matrix.h -- Various class and struct definitions necessary for vector
//    and matrix classes.


#if(!defined(MATRIX_H))
#define MATRIX_H  1
// This structure is used for packed matrices
typedef struct {
	char name[80];
	short startyear, endyear;
	short np; // number of pmatrix in this file
	short nr; // Number of rows
	short nc; // Number of columns
	short fr; // First row
	short fc; // First column
	short lr; // Last row
	short lc; // Last column
	long nelem; //number of nonzero coefficients in this pmatrix
	long BytesPerYear; // 4*nelem
	char loaded; /* 'y', if rnc and row has been allocated. 'n' for not */
	short *rnc;/* rnc[i] is the number of nonzero coefficients in row i;
				the range of i is: fr <= i <= lr.
				*/
	short **row;  /* row[i][j] is the column number of the jth nonzero
				coefficient in row i. j must be in the interval
				0 <= j < rnc[i].
				*/
	long FirstMat; // length of this file header = psn in file of first coeff
	// Always use these two methods, as they will ensure byte consistency!
	//  (found at bottom of pmatrix.cpp)
	short WriteOut(FILE* fpo);
	short ReadIn(FILE* fpi);
	} pmfid;

// Class declarations for simple matrix and vector package
class vorm // General Class for deriving Vector or Matrix classes
{
	protected:
	short which;	// Identifier for debugging; set = to vcount at creation;
	short nr;		// Number of rows
	short nc;		// Number of columns
	short fr;		// First row
	short fc;		// First column
	short lr;		// Last row
	short lc;		// Last column
	char temp;	// temp = 'y' is vector is temporary, otherwise 'n'
	short seq;	// this matrix corresponds to mats[seq].
	char DymeLoaded; // Indicates if it was read into a mats[] array. (y|n)
	char *name; // Used with dyme, when matrix declared by name.
	/*	freeh() frees the heap memory used for the object.  It must
		be written differently for matrices and vectors, but freet()
		(described below) looks the same for the two classes, except that the
		freeh() it calls is different.  Hence, freeh() is a good candidate to
		be a virtual function.  The user may also call freeh() to free up the
		heap that was need for an object that was needed for several uses but
		is then no longer needed after a certain point in the program. 
		Hence, freeh() is public.  */

	public:
	enum {OK=1, MAXSEC=750}; // Special OK for matrix class.
	short LastDat; // last year for which data is available for this vorm
	virtual void freeh(){return;} 
	/* 	freet() frees the heap memory used by objects created by operators
		when those objects are used by other operators. Thus, in the line
		"C = A + B;" the + first creates a temporary to hold A + B. The =
		then copies this temporary to C and deletes the temporary by a
		call to freet().  In the line "A = B", the = operator will also
		call freet, but since B is not temporary, freet() returns without
		calling freeh() to do the deletion. */

	protected:
	void freet(){if(temp == 'y'||temp == 'p' || temp == 'q') freeh();}
	public:
	short unload(void); 
	short rows(){return nr;}
	short columns(){return nc;}
	short firstrow(){return fr;}
	short firstcolumn(){return fc;}
	short lastrow(){return lr;}
	short lastcolumn(){return lc;}
	void SetLastData(short date);
	short LastData(void);
	char *GetName(void) {return name;};
	};

class IVector;
class Matrix;  // This line allows the "friends" of Vector to refer to Matrix.
class Pmatrix; // Same reason
class Equation; // Similarly for Equation (used in Seidel)
class VamFile;

class Vector : public vorm
{
 private:
	short nelem;	// Number of elements
	short first;  // the first row if a column vector, the first col if a row
	float *v;	// pointer to the vector itself
	short SetUpVector(VamFile& vf,char *name, char rf, char wf);

 public:
	short numelm(){return nelem;}
	short ReadA(char *filename); // Read ASCII from a file.
	void Display(FILE* fp,char *title,short width=6, short decimalPlaces=1,
		short pagewidth=76);
	void Display(char *title = "",short width = 6, short decimalPlaces = 1);
	Vector(short anr, short anc = 1, char atemp = 'n', short afr = 1, short afc = 1);
	Vector(Vector& a);
	Vector();
	// Special for DYME
	Vector(char *name, char rf = 'y', char wf = 'y');
	Vector(VamFile& vf,char *name, char rf = 'y', char wf = 'y');
    void r(char *name, char rf = 'y', char wf = 'y');
	// end Dyme specials.
	 ~Vector();
	Vector& operator = (const Vector& ROp);
	Vector operator - (const Vector &a);
	Vector operator ~ (); // Transposition
	void freeh(); // free heap memory 
	float& operator [] (const short i);	// range checked element
	float sum(short first=0, short last=0);
	float csum(char *Group);
	short First(void);
	double scale(float control);
	short fix(short t, short SectorFlag=0);  // default, normal fix routine
	// If SectorFlag > 0, do single sector, if (-1), do groups only
	short fix(short t, Vector &dd);  // "import fix"
	char GetFixType(short Sector);
	short IsFixOpt(short sector, char flag);
	
	void set(float value);
//	void resize(int nr, int nc = 1, int fr = 0, int fc = 0);// Changed by CA 6 Oct 2005
	void resize(int nr, int nc = 1, int fr = 1, int fc = 1);
	short load(short t); // for use in dyme
	short load(VamFile& vf, char* vname, short year);
	short store(short t); // for use in dyme
	short store(VamFile& vf, char* vname, short year);

	friend void operator << (Matrix& a, const Vector& b);
	friend Vector operator + (const Vector& a, const Vector& b);
	friend Matrix operator * (const Vector& a, const Vector& b);	// ab
	friend Vector operator * (const Matrix& a, const Vector& b);
	friend Vector operator * (const Vector& a, const Matrix& b);
	friend Vector operator * (const float a, const Vector& b);
	friend Vector operator / (const Vector& a, const float b);
	// a/b forms the product a'b without creating a'.
	friend Matrix operator / (const Vector& a, const Vector& b);	// a'b
	friend Vector operator / (const Matrix& A, const Vector& b);	// A'b
	friend Vector operator / (const Vector& a, const Matrix& b);	// a'B
	friend Matrix operator % (const Matrix& a, const Vector& b); // diagonalized *
	friend Matrix operator % (const Vector& b, const Matrix& a); // diagonalized *
	friend Vector ebediv(const Vector& a, const Vector& b); // elem by elem division
	friend Vector ebemul(const Vector& a, const Vector& b); // elem by elem multiplication
	friend Vector ebemul(const Matrix& a, short i, const Vector& b);// ebe with row i of matrix a
	friend Vector ebemul(const Pmatrix& a, short i, const Vector& b);// ebe with row i of Pmatrix a
	friend double dot(const Vector& a, const Vector& b); // dot (inner) product
    friend short Seidel(Matrix& A, Vector& q, Vector& f, IVector& itriang,
	    float toler);
	friend short Seidel(Matrix& A, Vector& q, Vector& f, short *triang,
		float toler);
	friend short Seidel(Pmatrix& A, Vector& q, Vector& f, Vector& dump, 
		short *triang, float toler);
	friend short Seidel(Pmatrix& A, Vector& q, Equation& ImportEq, Vector& fd,
		Vector& im, Vector& ven, Vector& ex, Vector& def,
		Vector& dump, short *triang, float toler);
	friend short Seidel(Pmatrix& A, Vector& q, Vector& imp, Equation& ImportEq, Vector& fd, 
		 Vector& im, Vector& ven, Vector& ex, Vector& def,
		 Vector& dump, short *triang, float toler);
	friend short PSeidel(Matrix& A, Vector& p, Vector& v, short *triang, 
		float toler);
	friend short PSeidel(Matrix& A, Vector& p, Vector& v, IVector& triang, 
		float toler);

	friend short PSeidel(Pmatrix& A, Vector& p, Vector& v, float toler);
	friend Pmatrix operator + (const Matrix& a, const Pmatrix& b);
	friend Pmatrix operator + (const Pmatrix& a, const Matrix& b);
	friend Pmatrix operator - (const Matrix& a, const Pmatrix& b);
	friend Pmatrix operator - (const Pmatrix& a, const Matrix& b);
	friend Vector operator * (const Vector& a, const Pmatrix& b);
	friend Vector operator * (const Pmatrix& a, const Vector& b);
	friend Vector operator / (const Vector& a, const Pmatrix& b); // a'B
	friend Vector operator / (const Pmatrix& a, const Vector& b); // A'b
	friend Pmatrix operator % (const Vector& a, const Pmatrix& b);
	friend Pmatrix operator % (const Pmatrix& a, const Vector& b);
	};


class IVector : public vorm
{
 private:
	short nelem;	// Number of elements
	short first;  // the first row if a column vector, the first col if a row
	short *v;	// pointer to the vector itself

 public:
	short numelm(){return nelem;}
    short First(void) {return(first);}
    short Last(void) {return(first+nelem-1);}
	short ReadA(char *filename); // Read ASCII from a file.
	void Display(char *title = "",short width = 6);
	IVector(short anr, short anc = 1, char atemp = 'n', short afr = 0, short afc = 0);
	IVector(IVector& a);
	IVector();
	~IVector();
	IVector& operator = (const IVector& ROp);
	IVector operator - (const Vector &a);
	void freeh(); // free heap memory
	short& operator [] (short i);	// range checked element
	short sum(short first=0, short last=0);
	void set(short value);
	void resize(int nr, int nc = 1, int fr = 0, int fc = 0);

	friend IVector operator + (const IVector& a, const IVector& b);
	};

class Pmatrix;

class Matrix : public vorm
{
 private:
	float **m;
	short SetUpMatrix(VamFile&vf,char *name,char rf, char wf);
	short DoFix(short t, short lvn, short SectorFlag);

 public:
	short ReadA(char *filename); // Read ASCII from a file.
   	Matrix();      // Default constructor; allows global declaration for later resizing
    void resize(int nr, int nc, int fr=1, int fc=1);
	void Display(char *title = "",short width = 6, short decimalPlaces = 2);
    void Display(FILE *fp,char *title = "",short width = 6, short decimalPlaces = 2);
	short load(short t); // for use in dyme to load individual matrix
	short load(VamFile& vf, char* vname, short year);
	short store(short year); // for use in dyme to store individual matrix
	short store(VamFile& vf, char* vname, short year);
	void set(float value);// set entire matrix to given value.
	/* 	Invert this matrix; return determinant.  Start row reduction with
		the pivot in row startrow and stop when the pivot has been done
		for row endrow.  If startrow is 0, start with first row; if
		endrow is 0, end with last row.  */

	short fix(short t, char *mname); // can specify matrix name
	short fix(short t, short SectorFlag=0);  // default, normal fix routine

	double invert(short startrow = 0, short endrow = 0);
	Matrix(short anr, short anc, char atemp = 'n', short afr = 1, short afc = 1);
	Matrix(Matrix& a);
	// Special for DYME
	Matrix(char *name, char rf = 'y', char wf = 'y');
	Matrix(VamFile& vf,char *name, char rf = 'y', char wf = 'y');
    void r(char *name, char rf = 'y', char wf = 'y');

	// End special DYME stuff...

	~Matrix();
	Matrix& operator = (const Matrix& a);
	Matrix& operator = (const Pmatrix& a);
	Matrix operator ~ ();  // Transposition
	Matrix operator ! ();  // Inversion
	float * operator [] (short i);	// range checked element
	float& operator () (short i, short j); // range-checked element
	void freeh(); // free heap memory 
	void claimh(); // claim heap memory

	friend void operator << (Matrix& a, const Vector& b);
	friend Matrix operator + (const Matrix& a, const Matrix& b);
	friend Matrix operator - (const Matrix& a, const Matrix& b);
	friend Matrix operator * (const Matrix& a, const Matrix& b);
	friend Matrix operator * (const Pmatrix& a, const Pmatrix& b);
	friend Vector operator * (const Vector& a, const Matrix& b);
	friend Vector operator * (const Matrix& a, const Vector& b);
	friend Matrix operator * (const Vector& a, const Vector& b);
	friend Matrix operator * (const float a, const Matrix& b);
	friend Matrix operator / (const Matrix& b, const float a);
	friend Matrix operator / (const Matrix& a, const Matrix& b); // form A'B without ~A
	friend Vector operator / (const Matrix& A, const Vector& b); // A'b
	friend Vector operator / (const Vector& a, const Matrix& b); // a'B
	friend Matrix operator / (const Vector& a, const Vector& b); // a'b
	friend Matrix operator % (const Matrix& a, const Vector& b); // diagonalized *
	friend Matrix operator % (const Vector& b, const Matrix& a); // diagonalized *
	friend Vector ebemul(const Matrix& a, short i, const Vector& b);// ebe with row i of matrix a
    friend short Seidel(Matrix& A, Vector& q, Vector& f, IVector& itriang,
	    float toler);
	friend short Seidel(Matrix& A, Vector& q, Vector& f, short *triang,
		float toler);
	friend short PSeidel(Matrix& A, Vector& p, Vector& v, short *triang,
		float toler);
	friend short PSeidel(Matrix& A, Vector& p, Vector& v, IVector& triang,
		float toler);

	friend Pmatrix;

	friend Pmatrix operator + (const Matrix& a, const Pmatrix& b);
	friend Pmatrix operator + (const Pmatrix& a, const Matrix& b);
	friend Pmatrix operator - (const Matrix& a, const Pmatrix& b);
	friend Pmatrix operator - (const Pmatrix& a, const Matrix& b);
	};

class Pmatrix : public vorm
{
 private:
	short SetUpPmatrix(VamFile& vf,char *pmname, char readflag,
		char writeflag, char *pmfname="");
 public:
	// Made public, to avoid need for friend functions (DSM -2/22/95)
	long nelem;	// number of nonzero coefficients
	short *rnc;	/* rnc[i] is the number of nonzero coefficients in row i;
				the range of i is: fr <= i <= lr.
				*/
	short **row;	/* row[i][j] is the column number of the jth nonzero
				coefficient in row i. j must be in the interval
				0 <= j < rnc[i].
				*/
	float **p;	/* p[i][j] is the jth nonzero coefficient in row i.
				i is in the range fr <= i <= lr;
				j is in the range 0 <= j < rnc[i].
				 */
	void Init();
	short ReadP(char *finame, short yr); // read from a binary file.
	// original load has been changed to LoadP since load(short yr) is
	// defined by Clopper for use in dyme. load() is thus redefined 6/30/93
	short LoadP(char *finame, short yr); // load data from a binary file
	short load(short yr); // this is for dyme
	short store(char* finame, short yr); // Write Pmatrix to a binary file
	short store(short yr);

	short fix(short t, char *mname);  // can specify name
	short fix(short t, short SectorFlag=0);
	short DoFix(short t, short lvn, short SectorFlag=0);

	void Display(char *title = "", short width = 8, short decimalPlaces = 5);
	Pmatrix();  // Default constructor.
	Pmatrix(short anr, short anc, char atemp = 'r', short aft = 1, short afc = 1);
	// Special constructors for Interdyme:
	Pmatrix(char *pmname, char readflag, char writeflag,
		char *pmfname="same");
	Pmatrix(VamFile& vf,char *pmname, char readflag, char writeflag,
		char *pmfname="same");
    r(char *pmname, char readflag, char writeflag,char *pmfname="same");
	// End special Interdyme stuff.
	Pmatrix(Pmatrix& a);
	Pmatrix& operator = (const Pmatrix& a);

	Pmatrix& operator = (const Matrix& a);
	Pmatrix(const Matrix& a);
	friend Matrix& Matrix::operator = (const Pmatrix& a);

	~Pmatrix();

	Pmatrix operator ~(); //Transposition
	short SetAij(short i, short j, float aij);
	float operator () (short i, short j);
	void freeh(); // free heap memory
	friend Pmatrix operator + (const Pmatrix& a, const Pmatrix& b);
	friend Pmatrix operator + (const Matrix& a, const Pmatrix& b);
	friend Pmatrix operator + (const Pmatrix& a, const Matrix& b);
	friend Pmatrix operator - (const Pmatrix& a, const Pmatrix& b);
	friend Pmatrix operator - (const Matrix& a, const Pmatrix& b);
	friend Pmatrix operator - (const Pmatrix& a, const Matrix& b);
	friend Vector operator * (const Vector& a, const Pmatrix& b);
	friend Matrix operator * (const Pmatrix& a, const Pmatrix& b);
	friend Pmatrix operator % (const Vector& a, const Pmatrix& b); // diagonalized pre mult.
	friend Pmatrix operator % (const Pmatrix& a, const Vector& b); // diagonalized post mult
	friend Vector operator * (const Pmatrix& a, const Vector& b);
	friend Vector ebemul(const Pmatrix& a, short i, const Vector& b);// ebe with row i of Pmatrix a
	friend Pmatrix operator * (const float a, const Pmatrix& b);
	friend Pmatrix operator / (const Pmatrix& a, const float b);
	friend Vector operator / (const Pmatrix& a, const Vector& b); //A'b
	friend Vector operator / (const Vector& a, const Pmatrix& b); // a'B

	// This one is used in IdLift:
	friend short Seidel(Pmatrix& A, Vector& q, Vector& imp, Vector& ven,
		Vector& sdc, Equation& impeq, Equation& veneq, Vector& f, 
		Vector& dd, Vector& ex, Vector& inter, Vector& impshr, Matrix& Use, 
		Matrix& Pdmlag, Matrix& Fpilag,
		IVector& triang, float toler);

	friend short Seidel(Pmatrix& A, Vector& q, Equation& ImportEq, Vector& fd,
		Vector& im, Vector& ven, Vector& ex, Vector& def,
		Vector& dump, short *triang, float toler);
	friend short Seidel(Pmatrix& A, Vector& q, Vector& f, Vector& dump,
		short *triang, float toler);
	friend short Seidel(Pmatrix& A, Vector& q, Vector& imp, Vector& impshare,
		Vector& f, short *triang,	float toler);
	friend short PSeidel(Pmatrix& A, Vector& p, Vector& v, float toler);
	friend short Seidel(Pmatrix& A, Vector& q, Vector& imp, Vector& sdc, 
		Equation& impeq, Vector& f, Vector& dd, Vector& ex, Vector& inter, 
			IVector& triang, float toler);
	};

// other prototypes
short isnum(char c);
short getfloat(float& f, FILE *fp, char *s, short& j);
short getint(short& f, FILE *fp, char *s, short& j);
#if !defined(OS2COMPILE)
short min(short a, short b);
short max(short a, short b);
float min(float a, float b);
float max(float a, float b);
#endif
short setvec(Vector& A, float z);
short setmat(Matrix& A, float z);
short pulloutcol(Vector& V, Matrix& A, short k);
short putincol(Vector& V, Matrix& A, short k);
short pulloutrow(Vector& V, Matrix& A, short k);
short putinrow(Vector& V, Matrix& A, short k);
Vector colsum(const Matrix& A);
Vector rowsum(const Matrix& A);
Vector colsum(const Pmatrix& A);
Vector rowsum(const Pmatrix& A);

short readmat(Matrix& A, char *filename);
short writemat(Matrix& A, char *filename,short width,short decimalPlaces);
short ras(Matrix& A, Vector& row, Vector& col);
Matrix ebediv (const Matrix &a, const Matrix &b);
void tap();
void outofmem(char *);
#endif

