// VAMFILE.CPP -- Methods and utilities for VAM bin file.

#include "features.h"

#include <stdio.h>  // Stick with C streams for now!
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <string.h>

#ifdef __LINUX__
#include <sys/stat.h>
#include <sys/types.h>
#include "insert.h"
#else
#include <sys\stat.h>
#include <sys\types.h>
#include <conio.h>
#endif

#include "vamfile.h"
#include "gbank.h"
#include "compbank.h"
#include "hashbank.h"
#include "constant.h"

// Uncomment this statement to use CompBank or HashBank from VamFile
//#define USECOMPBANK

// In MISCTOOL.cpp
short matchwild(char*, char*);
short breakname(char*, char*);
short breakname(char *nam, char *vnam, short &prow);
float **matrix(short nrl, short nrh, short ncl, short nch);
void free_matrix(float **m, short nrl, short nrh, short ncl);

/*==========================================================================*/
matid::matid() {
vfp=0;
loaded='n';
vptr=0;
mptr=0;
vecptr=0;
matptr=0;
pmfname=0;
}
/*==========================================================================*/
matid::~matid() {
//if(vptr)
 //  delete[] vptr;
//if(mptr)
//   delete[] mptr;
if(pmfname)
	delete[] pmfname;
}
/*==========================================================================*/
char * VamFile::Setbnktitle(const char *s)  {

bnktitle = new char[strlen(s)+1];
strcpy(bnktitle,s);
strcpy(fd.runname,bnktitle);
return bnktitle;
}

/*==========================================================================*/
VamFile::VamFile(const char *rootname) : DataBank()  {

char *cp,fname[120];
short i,err,PmatNo;
FILE *fpop;

MsgObject=new GMessage;
/* Open .vam file and initialize. */
strcpy(fname,"VamFile");
classname=new char[strlen(fname)+1];
strcpy(classname,fname);
// File can have a different extension than .vam.  If the argument passed
//   here has no extension, however, it is assumed to be .vam
strcpy(fname,rootname);
if((cp=strchr(fname,'.'))==NULL)
	strcat(fname,".vam");
bnkname=new char[strlen(fname)+1];
strcpy(bnkname,fname);

if((pbank=fopen(bnkname,"r+b"))==NULL ) {
	printf(MSGBK(24),bnkname);
	banktype='x';
	goto failure;
	}
read_filedesc();

strcpy(fname,fd.runname);
bnktitle=new char[strlen(fname)+1];
strcpy(bnktitle,fname);

// initialize some miscellaneous variables:
strcpy(loadvec,"");
lvchanged='n';
loadview='x';
loaded='n';
lvn=0;
loadline=0;
lveclines=0;  err=err;
GOpened = 0;

// Initialize some info for packed matrices:
pmh = new pmfid[fd.pmcount];
for(PmatNo=0;PmatNo<fd.pmcount;PmatNo++) {
	if((fpop = fopen(pmfilenames[PmatNo], "rb+"))==NULL){
		printf("Cannot open %s with Packed Matrix.\n", pmfilenames[PmatNo]);
		PmLoaded[PmatNo] = 0;
		continue;
		}
	// It's messier to read each element separately, but portable. (DSM-2/9/98)
	fread(&pmh[PmatNo].name,sizeof(char),80,fpop);
	fread(&pmh[PmatNo].startyear,sizeof(short),1,fpop);
	fread(&pmh[PmatNo].endyear,sizeof(short),1,fpop);
	fread(&pmh[PmatNo].np,sizeof(short),1,fpop);
	fread(&pmh[PmatNo].nr,sizeof(short),1,fpop);
	fread(&pmh[PmatNo].nc,sizeof(short),1,fpop);
	fread(&pmh[PmatNo].fr,sizeof(short),1,fpop);
	fread(&pmh[PmatNo].fc,sizeof(short),1,fpop);
	fread(&pmh[PmatNo].lr,sizeof(short),1,fpop);
	fread(&pmh[PmatNo].lc,sizeof(short),1,fpop);
	fread(&pmh[PmatNo].nelem,sizeof(long),1,fpop);
	fread(&pmh[PmatNo].BytesPerYear,sizeof(long),1,fpop);
	fread(&pmh[PmatNo].loaded,sizeof(char),1,fpop);
	// skip reading of the two pointers.
	fseek(fpop, (sizeof(short*)+sizeof(short**)),SEEK_CUR);
	fread(&pmh[PmatNo].FirstMat,sizeof(long),1,fpop);
	PmLoaded[PmatNo] = 1;
	fclose(fpop);
	}

banktype='v';
//printf("    File description of %s read.\n",bnkname);
opened=1;
// Here we try to open the associated G bank file...
strcpy(fname,rootname);
if( (cp=strchr(fname,'.'))!=NULL)
	*cp=0;
err=TryVamBank(fname);

// Calculate number of series.
nseries=0;
for(i=0;i<fd.nvecs;i++) 
	if(columns[i]==1)
		nseries += sizes[i];
if(err>0)
	nseries += VamBank->Getnseries();
stper = 1;
return;
failure:
opened=-1;
}
/*==========================================================================*/
// The next two functions essentially split up the functions of the 
//   previous constructor.  The first is a default constructor, which
//   can be used even when the VamFile is set up as a global variable.
//   The OpenVam function is like the Gbank Assignbank() function, that
//   can then be used dynamically to initialize the global VamFile.

VamFile::VamFile(void){
char fname[120];
short i,err;

strcpy(fname,"VamFile");
classname=new char[strlen(fname)+1];
strcpy(classname,fname);
strcpy(loadvec,"");
lvchanged='n';
loadview='x';
loaded='n';
lvn=0;
loadline=0;
lveclines=0;  err=err;
GOpened = 0;
banktype='v';
MsgObject=new GMessage;
opened=-1;
return;
}
/*==========================================================================*/
short VamFile::OpenVam(const char *rootname, int OpenGBank){
char *cp,fname[120],root[120];
short i,err;

strcpy(root,rootname);
strcpy(fname,rootname);
if((cp=strchr(root,'.'))==NULL)
	strcat(fname,".vam");
else *cp = 0; // strip root for later call to TryVamBank.
bnkname=new char[strlen(fname)+1];
strcpy(bnkname,fname);

if((pbank=fopen(bnkname,"r+b"))==NULL ) {
	printf(MSGBK(24),bnkname);
	banktype='x';
	goto failure;
	}
read_filedesc();
strcpy(fname,fd.runname);
bnktitle=new char[strlen(fname)+1];
strcpy(bnktitle,fname);

opened=1;
// Here we try to open the associated G bank file...
if(OpenGBank)
	err=TryVamBank(root);

// Calculate number of series.
nseries=0;
for(i=0;i<fd.nvecs;i++)
	if(columns[i]==1)
		nseries += sizes[i];
if(GOpened)
	nseries += VamBank->Getnseries();
stper = 1; err=err;
return 0;
failure:
opened=-1;
return (-1);
}
/*==========================================================================*/
// TryVamBank: return 1 for hbk, 2 for cbk, 3 for bnk, -1 for failure.
short VamFile::TryVamBank(const char *whatsitsname)  {
struct stat buffer;
char tryname[100];
short IsNotThere;
#if defined(USECOMPBANK)
strcpy(tryname,whatsitsname);
strcat(tryname,".hbk");

IsNotThere=stat(tryname,&buffer);
if(IsNotThere) {
	VamBank = new HashBank(whatsitsname);
	GOpened = VamBank->IsOpen();
	if(GOpened) return(1);
	}
strcpy(tryname,whatsitsname);
strcat(tryname,".cbk");
IsNotThere=stat(tryname,&buffer);
if(IsNotThere) {
	VamBank = new CompBank(whatsitsname);
	GOpened = VamBank->IsOpen();
	if(GOpened) return(2);
	}
#endif
strcpy(tryname,whatsitsname);
strcat(tryname,".bnk");
IsNotThere=stat(tryname,&buffer);
if(!IsNotThere) {
	VamBank = new GBank(whatsitsname);
	GOpened = VamBank->IsOpen();
	if(GOpened) 
		return(3);
	}
return(ERR);
}
/*==========================================================================*/
VamFile::~VamFile() { // destructor
short i;
//printf("Vamfile destructor called.\n");
write_filedesc();  // If new runtitle, etc, write it out.
if(MsgObject)
	delete MsgObject;
if(classname)
	delete[] classname;
if(bnkname)
	delete[] bnkname;
if(bnktitle)
	delete[] bnktitle;
if(vindx)
	delete[] vindx;
if(sizes)
	delete[] sizes;
if(csizes)
	delete[] csizes;
if(columns)
	delete[] columns;
if(rows)
	delete[] rows;
if(lags)
	delete[] lags;
if(vnames)
	delete[] vnames;
for(i=0;i<fd.nvecs;i++) {
	if(rowstb[i])
		delete[] rowstb[i];
	if(colstb[i])
		delete[] colstb[i];
	}
if(rowstb)
	delete[] rowstb;
if(colstb)
	delete[] colstb;
if(types)
	delete[] types;
if(fd.pmcount>0) {
	for(i=0;i<fd.pmcount;i++)
		if(pmfilenames[i])
			delete[] pmfilenames[i];
	delete[] pmfilenames;
	}
if(GOpened)
	delete VamBank;
if(pmh)
	delete[] pmh;
//fclose(pbank);
}
/*==========================================================================*/
/* read_filedesc -- Read file description information in VAM.BIN file. 
	This data is now in VAM.INC.  May be good to create a data structure
	of this stuff: fd,vindx,sizes,columns,lags,vnames,rowstb,colstb */

short VamFile::read_filedesc(void) {

short i;

/* read VAM file description... */
fread(&fd.startdate,sizeof(float),1,pbank);
fread(&fd.enddate,sizeof(float),1,pbank);
fread(&fd.nnpy,sizeof(long),1,pbank);
fread(&fd.startpsn,sizeof(long),1,pbank);
fread(&fd.nvecs,sizeof(short),1,pbank);
fread(&fd.ncivnames,sizeof(unsigned short),1,pbank);
fread(&fd.pmcount,sizeof(short),1,pbank);
if(fd.pmcount<0)
	fd.pmcount=0;
fread(fd.runname,sizeof(char),80,pbank);
if(fd.pmcount>0) {
	pmfilenames=new char *[fd.pmcount];
	for(i=0;i<fd.pmcount;i++)  {
		pmfilenames[i]=new char[32];
		fread(pmfilenames[i],32,1,pbank);
		}
	}
// This could more easily be done as an array of structures:
nvecs=fd.nvecs;
vindx=new unsigned short[nvecs+1];
fread(vindx,sizeof(unsigned short),nvecs+1,pbank);
sizes=new long[nvecs];
fread(sizes,sizeof(long),nvecs,pbank);
rows=new unsigned short[nvecs];
fread(rows,sizeof(unsigned short),nvecs,pbank);
columns=new unsigned short[nvecs];
fread(columns,sizeof(unsigned short),nvecs,pbank);
lags=new short[nvecs];
fread(lags, sizeof(short),nvecs,pbank);
types=new char[nvecs];
fread(types,sizeof(char),nvecs,pbank);
vnames=new char[vindx[nvecs]];
fread(vnames,1,vindx[nvecs],pbank);

rowstb = new char*[nvecs];
colstb = new char*[nvecs];
for(i=0;i<nvecs;i++) {
	rowstb[i]=new char[30];
	fread(rowstb[i],1,30,pbank);
	}
for(i=0;i<nvecs;i++) {
	colstb[i]=new char[30];
	fread(colstb[i],1,30,pbank);
	}
nobs = np = (short)(fd.enddate - fd.startdate)+1;
jahra = (short)fd.startdate;
jahra=(jahra<1900)?jahra+1900:jahra;
cumsize();
return 0;
}
/*==========================================================================*/
// CheckHeader -- Print out header info to the screen, in case we need
//    to debug reading.
short VamFile::CheckHeader(int HowMany) {
int c;
cprintf("Summary of Header contents of Vam file %s.\n\r",bnkname);
cprintf("------------------------------------------\n\r");
cprintf(" Title: %s\r\n",bnktitle);
cprintf("Starting Date %4.0f.  Ending Date %4.0f\r\n",fd.startdate,fd.enddate);
cprintf("nnpy: %9ld  startpsn: %9ld  nvecs: %5d  ncivnames: %5d\r\n",
	fd.nnpy,fd.startpsn,fd.nvecs,fd.ncivnames);
cprintf("A look at the first few vectors or matrices.\r\n\r\n");
cprintf("Name    rows cols  lags  rowstb       colstb\n\r");
cprintf("============================================\n\r");
short numprint = fd.nvecs<HowMany?fd.nvecs:HowMany;
for(short i=0;i<numprint;i++) {
	cprintf("%8s%4d %4d  %4d  %10s %10s\r\n",&vnames[vindx[i]],rows[i],columns[i],
		lags[i],rowstb[i],colstb[i]);
	if(!((i+1)%23)) 
		c = getch();
		//tap();
	}
cprintf("\r\n"); 

return 0;
}
/* =============================================================== */
short VamFile::write_filedesc(void){
	short i;
	/* write VAM file description... */
	if(opened==-1) return -1;
	fseek(pbank,0,0);
	fwrite(&fd.startdate,sizeof(float),1,pbank);
	fwrite(&fd.enddate,sizeof(float),1,pbank);
	fwrite(&fd.nnpy,sizeof(long),1,pbank);
	fwrite(&fd.startpsn,sizeof(long),1,pbank);
	fwrite(&fd.nvecs,sizeof(short),1,pbank);
	fwrite(&fd.ncivnames,sizeof(unsigned short),1,pbank);
	fwrite(&fd.pmcount,sizeof(short),1,pbank);
	if(fd.pmcount<0)
		fd.pmcount=0;
	fwrite(fd.runname,sizeof(char),80,pbank);
	if(fd.pmcount>0) {
		for(i=0;i<fd.pmcount;i++)  {
			fwrite(pmfilenames[i],32,1,pbank);
			}
		}
	return 0;
}
/* =============================================================== */
/* cumsize -- Calculate cumulative sizes*/
short VamFile::cumsize(void) {
/*Cumulate sizes */
short i;
csizes=new long[nvecs+1];
csizes[0] = 0;
for(i = 0; i < nvecs; i++){
	// space for packed matrices are not assigned,  qyu
	if(types[i]!='p')
		csizes[i+1] = csizes[i]+sizes[i];
	else csizes[i+1] = csizes[i];
	}
return(OK);
}
/* =============================================================== */
/* findname  -- find a name in the list of vector names.  This assumes
	that either read_filedesc, or configure has been called. 
	*/
short VamFile::findname(char *name) {
short i;
for(i = 0; i < nvecs; i++)
	if((strcmp(name,&vnames[vindx[i]])) == 0) return(i);
//printf("There is no vector by the name of %s.\n",name);
return(ERR);
}
/*==========================================================================*/
short VamFile::GetPmatNo(char *pmname) {
short i, pmatcnt=0;
for(i = 0;i < nvecs; i++) {
	if((strcmp(pmname,&vnames[vindx[i]])) == 0 && types[i] == 'p')
		return(pmatcnt);
	if(types[i]=='p')
		pmatcnt++;
	}
return(ERR);
}
/*==========================================================================*/
/* GETSER - Gets a vam series for a single sector within the vam.bin
	opened on fpbank. */

short VamFile::getser(char *name, float *ss,short SeriesLength) {

short sec,row,i,n,lvn,err,IsMatrix=0;
long offset, psn;
char vamitem[20];

/* If this is a vector, sec will be the vector index.  If a matrix,
	row is the row number, and sec is the column number. */

sec=breakname(name,vamitem,row);

if(sec<=0) 
	goto TryGbank;
lvn=findname(vamitem);
if(lvn<0 || lvn>nvecs-1) 
	goto TryGbank;
if(columns[lvn]>1)
	IsMatrix = 1;
if(IsMatrix && row <= 0) {
	printf(" Error in matrix expression: %s.\n",name);
	goto Error;
	}

// Check if the vector or matrix is loaded or not.

if(loaded == 'y' && strcmp(vamitem,loadvec)==0 && loadline==row) {
	for(n=0;n<np;n++)  {
		ss[n] = nmvec[n+1][sec];
		}
	}
else {
	if(!IsMatrix) {
		for(n=0;n<np;n++) {
		psn=(long)fd.startpsn+(long)fd.nnpy*(long)n*sizeof(float)+
			(long)csizes[lvn]*sizeof(float)+
			(long)(sec-1)*sizeof(float);
		if((err=fseek(pbank,psn,SEEK_SET)) != 0) {
			printf("Seek in Vam file failed.\n");
			goto TryGbank;
			}
		if((err=fread(&ss[n],sizeof(float),1,pbank))!=1) {
			printf("fread in vam file failed.\n");
			goto TryGbank;
			}
		}
	}
	else {  // matrix
		if((err=loadcmd(vamitem,'r',row)) < 0) {
			printf("Unable to do load of %s.\n",vamitem);
			goto Error;
			}
		for(n=0;n<np;n++)  {
			ss[n] = nmvec[n+1][sec];
			}
		}
	}
err=err;
return(0);
TryGbank:
if(!GOpened ) 
	goto Error;
err=VamBank->getser(name,ss);
if(err<0) goto Error;
SeriesLength=SeriesLength;
return(0);
Error:
printf("%s not found.\n",name);
return(ERR);
}
/*==========================================================================*/
/* GetVecElement - Gets a vam vector element for a single year.  Sometimes
	this is all that is needed. */
short VamFile::GetVecElement(char *name, short yr, float& f) {

short sec,i,n,lvn,err;
long offset, psn;
char vamitem[20];
sec=breakname(name,vamitem);
if(sec<=0) return(-1);
lvn=findname(vamitem);
if(lvn<0 || lvn>nvecs-1) 
	return (-1);
if(columns[lvn]!=1) {
	printf(" GetVecElement only works on vectors, sorry!\n");
	return(-1);
	}
if(yr<25)
	yr += 2000;
else if(yr < 1900)
	yr += 1900;
n = yr - (short)fd.startdate;
psn=(long)fd.startpsn+(long)fd.nnpy*(long)n*sizeof(float)+
		csizes[lvn]*sizeof(float)+
		(long)(sec-1)*sizeof(float);    /* Check on this. */
if((err=fseek(pbank,psn,SEEK_SET)) != 0) {
	printf("Seek in vam file failed.\n");
	return(-1);
	}
if((err=fread(&f,sizeof(float),1,pbank))!=1) {
	printf("fread in vam file failed.\n");
	return(-1);
	}
err=err;
return(0);
Error:
printf("%s not found.\n",name);
return(ERR);
}
/*==========================================================================*/
// This overloaded version takes name of vector and sector element separately.
short VamFile::GetVecElement(char *vname, short sec, short yr, float& f){
short i,n,lvn,err;

long offset, psn;

if(sec<=0) return(-1);
lvn=findname(vname);
if(lvn<0 || lvn>nvecs-1) 
	return (-1);
if(columns[lvn]!=1) {
	printf(" GetVecElement only works on vectors, sorry!\n");
	return(-1);
	}
if(yr<25)
	yr += 2000;
else if(yr < 1900)
	yr += 1900;
n = yr - (short)fd.startdate;
psn=(long)fd.startpsn+(long)fd.nnpy*(long)n*sizeof(float)+
		csizes[lvn]*sizeof(float)+
		(long)(sec-1)*sizeof(float);    /* Check on this. */
if((err=fseek(pbank,psn,SEEK_SET)) != 0) {
	printf("Seek in vam file failed.\n");
	return(-1);
	}
if((err=fread(&f,sizeof(float),1,pbank))!=1) {
	printf("fread in vam file failed.\n");
	return(-1);
	}
err=err;
return(0);
Error:
printf("%s not found.\n",vname);
return(ERR);

}

/*==========================================================================*/
/* GetMatElement - Gets a vam matrix element for a single year.  
	Updated 4/15/98 to get element of Packed matrices also. 
	
*/
short VamFile::GetMatElement(char *name, short row, short col, 
	short yr, float& f) {

short i,n,lvn,err,ncols,PmatNo;
long offset, psn;
float *p=0;
char vamitem[20],c;
lvn=findname(name);
if(lvn<0 || lvn>=nvecs)
	return (-1);
if((ncols=columns[lvn])==1) {
	printf(" GetMatElement only works on matrices, sorry!\n");
	return(-1);
	}
if(yr<25)
	yr += 2000;
else if(yr < 1900)
	yr += 1900;
n = yr - fd.startdate;
c = Gettype(lvn);
if(c=='p' || c=='P') { // packed matrix
	short nc,nr;
	PmatNo = GetPmatNo(name);
	nc = pmh[PmatNo].nc;
	nr = pmh[PmatNo].nr;
	p = new float[nc];
	readarow(PmatNo,yr,row,p);
	f = p[col-1];
	delete p;
	}
else { // regular matrix
	psn=(long)fd.startpsn+(long)fd.nnpy*(long)n*sizeof(float)+
			csizes[lvn]*sizeof(float)+
			(long)(row-1)*ncols*sizeof(float) +
			(long)(col-1)*sizeof(float);

	if((err=fseek(pbank,psn,SEEK_SET)) != 0) {
		printf("Seek in vam file failed.\n");
		return(-1);
		}
	if((err=fread(&f,sizeof(float),1,pbank))!=1) {
		printf("fread in vam file failed.\n");
		return(-1);
		}
	}
err=err;
return(0);
Error:
printf("%s not found.\n",name);
return(ERR);
}
/*==========================================================================*/
/* ADDSER - Writes a vam series for a single sector within the VamFile
	opened on fpbank. */

short VamFile::addser(char *name, float *ss) {

short sec,i,n,lvn,err;
long offset, psn;
char vamitem[20];
sec=breakname(name,vamitem);
lvn=findname(vamitem);
if(columns[lvn]!=1) {
	printf(" getvamss only works on vectors, sorry!\n");
	return(-1);
	}
for(n=0;n<np;n++) {
	psn=(long)fd.startpsn+(long)fd.nnpy*(long)n*sizeof(float)+
		 csizes[lvn]*sizeof(float)+
		 (long)(sec-1)*sizeof(float);    /* Check on this. */
	if((err=fseek(pbank,psn,SEEK_SET)) != 0) {
	  printf("Seek failed.  Tap a key.\n");
	  getch();
	  }
	if((err=fwrite(&ss[n],sizeof(float),1,pbank))!=1) {
		printf("fwrite failed.  Tap a key.\n");
		getch();
		}
	}
err=err;
return(0);
}
/*==========================================================================*/
/* loadcmd() -- load all values for some vector ****************************
	loadview is defined as follows:
	If the loaded object is a vector, loadview is x.
	If the loaded object is a matrix, loadview is:
		y       if the whole matrix is loaded for one year
		r       if values of a row are loaded for all years
		c   if values of a column are loaded for all years.
*/
// Adopted for use in this class 3/10/93 - DSM.
short VamFile::loadcmd(char *s,char view, short line)  /*dsm revised*/
{
	long n,i,j,err,offset,nrows,ncols;
	size_t lvecsize;
	short pmcnt,col;
	long psn,siz,lrat;
	float *row;

	if(loaded == 'y' && strcmp(s,loadvec) == 0 && view == loadview && 
		line == loadline) return(OK); /*this vector already loaded*/
	if(loaded == 'y' && lvchanged == 'y') {
		storecmd(loadvec);
		lvchanged = 'n';
		}
	if(loaded == 'y') { // Here when lv was not changed, but we want a new one.
		if(loadview != 'y') free_matrix(nmvec,1,np,1);
		else free_matrix(nmvec,1,lveclines,1);
		loaded = 'n';
		}

	if((i = findname(s)) == ERR) return(ERR);
	lvn = i;

	// Figure out lvecsize, the size in bytes of the space to be malloced
	//   for this vector, and other variables.
	offset = 0;
	if(view == 'x') { // vectors
		lveclines = sizes[lvn];
		if((nmvec = matrix(1,np,1,lveclines))==0) outofmem(s);
		for(i=1;i<=np;i++)
	 for(j=1;j<=lveclines;j++)
		 nmvec[i][j]=0.0;
		}
	else if(view == 'c'){ // column of a matrix
		siz = sizes[lvn];
		col = columns[lvn];
		lrat = siz/col;
		lveclines = (short)(lrat);
//              lveclines = (short)(sizes[lvn]/columns[lvn]);
		if((nmvec = matrix(1,np,1,lveclines))==0) outofmem(s);
		}
	else if (view == 'r') { // row of a matrix
		offset = (line-1)*columns[lvn]*4;
		lveclines =columns[lvn];
		if((nmvec = matrix(1,np,1,lveclines))==0) outofmem(s);
		}
	else if(view == 'y') {
		ncols = columns[lvn];
		siz = sizes[lvn];
		lrat = siz/ncols;
		lveclines = (short)(lrat);
//              lveclines = (short)sizes[lvn]/(short)columns[lvn];
		if(sizes[lvn]>30000){
			printf("Error: Matrix size too large to load\n");
			return ERR;
			}
		if((nmvec = matrix(1,lveclines,1,ncols))==0) outofmem(s);
		}

	if(types[lvn]=='p'){ // read packed matrix file
		pmcnt = lags[lvn];
		if(pmloadcmd(nmvec,pmcnt,view,line,1,1)==ERR){
			printf("Pmload failed\n");
			return(ERR);
			}
		goto end_read;
		}

	//  read the data into it.
	if(view == 'x' || view == 'r'){
		for(n = 1; n <= np; n++){
			psn = (long)fd.startpsn + (long)fd.nnpy*(n-1)*4 + 
				(long)csizes[lvn]*4 + (long)offset;
			if((err = fseek(pbank,psn,0)) != 0){
				printf("Fseek failed in loadcmd. Tap.\n");
				if(getch() == ESC) exit(1);
				}
			if((err = fread(&nmvec[n][1], 4, lveclines, pbank)) != lveclines){
				printf("Fread failed in loadcmd. Tap.\n");
				if(getch() == ESC) exit(1);
				}
			if(err != lveclines){
				printf("Cannot read all of vector.\n");
				return(ERR);
				}
			}
		}
	else if(view == 'c'){
		ncols = columns[lvn];
		nrows = (long)sizes[lvn]/(long)ncols;
		for(n = 1; n <= np; n++){
			for(i = 1; i <= lveclines;i++){
				psn = (long)fd.startpsn + (long)fd.nnpy*(n-1)*4L + 
					(long)csizes[lvn]*4+
					( (long)(i-1) * (long)columns[lvn]+(long)line-1L)*4L;
				err = fseek(pbank,psn,0);
				err = fread(&nmvec[n][i],4,1,pbank);
				if(err != 1){
					printf("Cannot read all of vector.\n");
					return(ERR);
					}
				}
			}
		}
	else if(view == 'y'){
		n = line - (short)fd.startdate;
		psn = (long)fd.startpsn + (long)fd.nnpy * (long)n * 4L + 
			(long)csizes[lvn] * 4L;
		err = fseek(pbank,psn,0);
		ncols = columns[lvn];
		siz = sizes[lvn];
		lrat = siz/ncols;
		nrows = (short)(lrat);
//              nrows = (short)sizes[lvn]/(short)columns[lvn];
		for(i=1;i<=nrows;i++)
			err = fread(&nmvec[i][1],4,ncols,pbank);
		}
end_read:
	loadview = view;
	loadline = line;
	loaded = 'y';
	lvchanged = 'n';
	strcpy(loadvec,s);
	return(OK);
	}

/*=========================================================================*/
/* storecmd -- store the loaded vector ***************************************/

short VamFile::storecmd(char *name)  {

// lvn = loaded vector number
long i,j,n,nrows,ncols,offset;
long psn;
float *row;

lvn = findname(name);
if(lvn <0) {
	printf("Cannot store %s.  Illegal Name\n",name);
	return(-1);
	}

if(loaded != 'y'){
	printf("No vector loaded, so none stored.\n");
	return(ERR);
	}

if(loadview != 'x'){ // The object loaded is a matrix
	ncols = columns[lvn];
	nrows = (long)sizes[lvn]/(long)ncols;
	}
else ncols = sizes[lvn]; // The object loaded is a vector

if(loadview == 'x') offset = 0;
else if (loadview == 'r') {
	offset = (loadline-1)*columns[lvn]*4;
	}
if(loadview == 'x' || loadview == 'r'){
	if(types[lvn]=='p') {
		short std = (short)fd.startdate;
		for(n=0; n<np; n++)
			writearow(pmfilenames[lags[lvn]], std+n, loadline, &nmvec[n+1][1]);
		}
	else for(n = 1; n <= np; n++){
		psn = (long)fd.startpsn + (long)fd.nnpy*(long)(n-1)*4L + 
			(long)csizes[lvn]*4 + (long)offset;
		fseek(pbank,psn,0);
		// fwrite(lvec+n*ncols,4,ncols,pbank);
		fwrite(&nmvec[n][1],4,ncols,pbank);
		}
	free_matrix(nmvec,1,np,1);
	}
else if(loadview == 'y'){
	n = loadline - (long)fd.startdate;
	if(types[lvn]=='p'){
		for(i=1; i<= nrows; i++)
			writearow(pmfilenames[lags[lvn]], loadline, i, &nmvec[i][1]);
		}
	else {
		fseek(pbank,fd.startpsn+fd.nnpy*n*4+csizes[lvn]*4,0);
		// fwrite(lvec,4,sizes[lvn],pbank);
		for(i = 1; i <= nrows;i++)
			fwrite(&nmvec[i][1],4,ncols,pbank);
		}
	free_matrix(nmvec,1,lveclines,1);
	}
else if(loadview == 'c'){
	if(types[lvn]=='p'){
		printf("Cannot store a column of a packed matrix at this moment\n");
		printf("Could you store a row instead?\n");
		}
	else {
		cprintf("\r\nStoring a column ");
		for(n = 1; n <= np; n++){
			if(n % 5 == 0) cprintf("."); 
			for(i = 0; i < nrows;i++){
				psn = (long)fd.startpsn + (long)fd.nnpy * (long)(n-1) * 4L +
					(long)csizes[lvn] * 4L +
					(i * (long)columns[lvn] + (long)loadline - 1L )*4L;
				fseek(pbank,psn,0);
				//fwrite(lvec+(n*nrows)+i,4,1,pbank);
				fwrite(&nmvec[n][i-1],4,1,pbank);
				}
			}
		}
	free_matrix(nmvec,1,np,1);
	cprintf("\r\n"); 
	}

loaded = 'n';
lvn = -1;
return(OK);
}
/*=========================================================================*/
/* unloadcmd() 
	release space allocated to nmvec without storing the loaded data. */
short VamFile::unloadcmd(void)
{
	long i,j,n,nrows,ncols,offset;
	unsigned long psn;
	float *row;

	if(loaded != 'y') return(OK);
	if(loadview == 'y')
		free_matrix(nmvec,1,lveclines,1);
	else free_matrix(nmvec,1,np,1);

	loaded = 'n';
	lvchanged = 'n';
	lvn = -1;
	return(OK);
	}
/*=========================================================================*/
/* writearow() - write a row of a packed matrix for year yr to 
	file for packed matrix corresponding to PmatNo.
	p is a vector of full size, the number of columns of the 
		matrix.  This routine writes out only the non-zero elements.
	*/
short VamFile::writearow(short PmatNo, short yr, short row, float *p) {
	writearow(pmfilenames[PmatNo],yr,row,p);
	}
short VamFile::writearow(char* pmfname, short yr, short row, float *p)
{
	FILE *fin;
	pmfid pmh;
	long rowpos, ppos, nnz, nz=0;
	short *prnc;
	float *pp;
	char wheretxt[60];
	short i, j, j1, k;

	printf(wheretxt,"Pmatrix, writing");
	if((fin=fopen(pmfname, "rb+"))==0) 
		return(ERR);
	fread(&pmh, sizeof(pmh), 1, fin);
	j=pmh.nr>pmh.nc ? pmh.nr : pmh.nc;  // pnrc first store nonzeros in each row
									// then strore column number in that "row"
	if((prnc = new short[j])==0) outofmem(wheretxt);
	fread(prnc, 2, row, fin);
	for(i=0; i<row-1; i++)
		nz +=prnc[i];
	nnz = prnc[row-1];
	rowpos = sizeof(pmh) + 2*pmh.nr + 2 * nz;
	ppos = sizeof(pmh)+2*pmh.nr+2*pmh.nelem+4*(yr-pmh.startyear) * pmh.nelem;
	ppos = ppos + 4*nz;
	fseek(fin, rowpos, SEEK_SET);
	fread(prnc, 2, nnz, fin);
	if((pp=new float[nnz])==0) outofmem(wheretxt);
	for(j1=0, k=pmh.fc;k<=pmh.lc; k++ )
		if(k==prnc[j1]) 
			pp[j1++] = p[k-1];
	fseek(fin, ppos, SEEK_SET);
	if(fwrite(pp, 4, nnz, fin)!=nnz){
		delete prnc;
		delete pp;
		fclose(fin);
		return(ERR);
		}
	delete prnc;
	delete pp;
	fclose(fin);
	return(0);
	}
/*=========================================================================*/
/* readarow()
	Read a row of a packed matrix for year yr to file pmfname.
	*/
short VamFile::readarow(short PmatNo, short yr, short row, float *p) {
	readarow(pmfilenames[PmatNo],yr,row,p);
	}
short VamFile::readarow(char *pmfname, short yr, short row, float *p)
{
	FILE *fin;
	pmfid pmh1;
	long rowpos, ppos, nnz, nz=0;
	short *prnc;
	float *pp;
	char wheretxt[60];
	short i, j, j1, k;

	sprintf(wheretxt,"Pmatrix, writing");
	if((fin=fopen(pmfname, "rb+"))==0) 
		return(ERR);
	fseek(fin,0L,SEEK_SET);
	fread(&pmh1, sizeof(pmh1), 1, fin);
	j=pmh1.nr>pmh1.nc ? pmh1.nr : pmh1.nc;  // pnrc first store nonzeros in each row
									// then strore column number in that "row"
	if((prnc = new short[j])==0) 
		outofmem(wheretxt);
	fread(prnc, 2, row, fin);
	for(i=0; i<row-1; i++)
		nz +=prnc[i];
	nnz = prnc[row-1];
	rowpos = sizeof(pmh1) + 2*pmh1.nr + 2 * nz;
	ppos = sizeof(pmh1)+2*pmh1.nr+2*pmh1.nelem+
		4*(yr-pmh1.startyear) * pmh1.nelem;
	ppos = ppos + 4*nz;
	fseek(fin, rowpos, SEEK_SET);
	fread(prnc, 2, nnz, fin);
	if((pp=new float[nnz])==0) 
		outofmem(wheretxt);
	fseek(fin, ppos, SEEK_SET);
	if(fread(pp, 4, nnz, fin)!=nnz){
		delete prnc;
		delete pp;
		fclose(fin);
		return(ERR);
		}
	for(j1=0, k=pmh1.fc;k<=pmh1.lc; k++ ) {
		p[k-1] = 0;
		if(k==prnc[j1]) 
			p[k-1] = pp[j1++];
		}
	delete prnc;
	delete pp;
	fclose(fin);
	return(0);
	}

/*=========================================================================*/
/*  pmloadcmd will load data from the file specified in pmfilenames[pmcnt]
	 into temp
	 */
short VamFile::pmloadcmd(float** temp,short pmcnt,char view,short line,short fr,short fc)
{
	FILE *fp;
	pmfid pmh1;
	short *row, *col;
	long *offset;
	float *tempcol;
	short pnr, pnc;
	short rno, cno;
	long i, j, k, k1, r, c;
	short lr,lc;
	long rncpos, colpos, ppos, cpos, ll, nz;

	if((fp=fopen(pmfilenames[pmcnt], "rb"))==0){
		printf("Error: Can not open file %s, load command failed.\n",pmfilenames[pmcnt]);
		return ERR;
		}
	fread(&pmh1, sizeof(pmh1), 1, fp);
	pnr = pmh1.nr;
	pnc = pmh1.nc;
	if(view=='r'){
		c=pnc;
		r=pmh1.endyear-pmh1.startyear+1;
		if(line<pmh1.fr||line>pmh1.lr){
			printf("Error: Invalid column number: %d!\n", line);
			return ERR;
			}
		}
	else if(view=='c'){
		c=pnr;
		r=pmh1.endyear-pmh1.startyear+1;
		if(line<pmh1.fc||line>pmh1.lc){
			printf("Error: Invalid row number: %d!\n", line);
			return ERR;
			}
		}
	else if(view=='y'){
		r=pnr;
		c=pnc;
		if(line<pmh1.startyear||line>pmh1.endyear){
			printf("Error: Invalid year : %d!\n", line);
			return ERR;
			}
		}
	else {
		printf("Error: Invalid view.\n");
		return ERR;
		}

	// Load data begins
	lr=fr+r-1;
	lc=fc+c-1;
	for(i=fr;i<=lr; i++)
		for(j=fc;j<=lc;j++)
			temp[i][j]=0.0;
	rncpos = sizeof(pmh1);
	colpos = rncpos + 2 * pnr;
	ppos = colpos + 2 * pmh1.nelem; 
	if((row = new short [pnr])==0||(col = new short [pnc])==0
		||(tempcol=new float[pnc])==0){
			outofmem(" pmload");
			}
	fseek(fp, rncpos, SEEK_SET);
	fread(row, 2, pnr, fp);

	if(view=='r'){
		// skip first (line-pmh1.fr) rows
		k1 = line - pmh1.fr;
		for(j=0, nz=0; j< k1; j++)
			nz+=row[j];
		pmrnzs = row[k1]; // this is for indexcmd();
		ppos += 4 * nz;
		colpos += 2 * nz;
		fseek(fp, colpos, SEEK_SET);
		fread(col, 2, row[k1], fp);
		for(i=fr; i<=lr; i++){
			fseek(fp, ppos, SEEK_SET);
			fread(tempcol, 4, row[k1], fp);
			for(j=0; j<row[k1]; j++){
				cno = col[j]-pmh1.fc+fc;
				temp[i][cno] = tempcol[j];
				}
			ppos+=4*pmh1.nelem;
			}
		}

	if(view=='c'){
		if((offset=new long[c])==0) 
			outofmem(" pmload");
		/*      If offset[i] is nozero, it is the offset of the first matrix whose 
			row i column "line" is nonzero so we need to read the number;
			col[i] is zero if we do not read this row.
			If view = 'c', c is the number of rows of the packed matrix.
			K stores accumulated number of nozeros in all previous rows;
			*/
		for(i=0, k=0; i<c; i++){ 
			offset[i] = 0;
			fread(col, 2, row[i], fp);
			for(j=0; j<row[i]; j++){
				if(col[j]==line) {
					offset[i]= ppos + 4 * (k + j);
					break;
					}
				}
			k+=row[i];
			}
		for(i=fr; i<=lr; i++)
			for(j=0; j<c; j++){ 
				ll = 4 * pmh1.nelem * (i-fr);
				if(offset[j]!=0) {
					cpos = ll + offset[j];
					cno = fc + j;
					fseek(fp, cpos, SEEK_SET);
					fread(&temp[i][cno], 4, 1, fp);
					}
			}
		delete offset;
		}

	if(view=='y'){
		ppos = ppos + (line - pmh1.startyear) * pmh1.nelem * 4;
		for(i=0; i<pnr; i++){
			fseek(fp, colpos, SEEK_SET);
			fread(col, 2, row[i], fp);
			colpos+=2*row[i];
			fseek(fp, ppos, SEEK_SET);
			fread(tempcol, 4, row[i], fp);
			ppos+=4*row[i];
			rno=i+fr;
			for(j=0; j<row[i]; j++){
				cno=col[j]+fc-pmh1.fc;
				temp[rno][cno]=tempcol[j];
				}
			}
		}
	delete tempcol;
	delete col;
	delete row;
	fclose(fp);
	return OK;
	}
/* ==================================================================== */
// ADDCONCEPT -- add an entire concept from the current VamFile to the
//  gbank referenced as *gb.

#include "gbank.h"

short VamFile::addconcept(char *concept, GBank *gb,
	  short startdate, short enddate)  {
float *zz=new float[NOMAX];
float *bb=new float[NOMAX];
short i,j,m,n,found,iblk,jser,err,nsec,isec,dby,gby,shift,ia,iz;
char gname[32];

lvn=findname(concept);
if(lvn==ERR) {
	printf("No such concept as %s.\n",concept);
	return(ERR);
	}
nsec = rows[lvn];

dby=GetBaseYr(); 
gby=gb->GetBaseYr();
gby=(gby<1900)?gby+1900:gby;

if(startdate<dby || startdate < gby ) {
	printf("Startdate %4d is too early for either Vamfile (%4d) or gbank(%4d)\n",
			 startdate, dby, gby);
	return(-1);
	}
shift=dby-gby;   // vamfile base yr minus g base yr.
ia=startdate-dby; // we're assuming annual for vamfile!
iz=enddate-dby;

err=loadcmd(concept,'x',0);
if(err==ERR) 
	return(ERR);

for(i=1;i<=nsec;i++) {
	for(j=0;j<nobs;j++)
		zz[j] = nmvec[j+1][i];
	for(j=0;j<nobs;j++) 
		bb[j]=0.000001;  // missing data value
	for(j=ia;j<=iz;j++)
		bb[j+shift]=zz[j];
	sprintf(gname,"%s%d",concept,i);
	err=gb->addser(gname,bb,'n'); // 'n' for don't write index
	}
return 0;
}
//=======================================================================
short VamFile::Announce(short bankno) {
// Give announcement that is normally given by bankcmd in G.
cprintf(MSGBK(19));
cprintf(MSGBK(3),bnktitle);
cprintf(MSGBK(66),jahra);
cprintf(MSGBK(5),nobs);
cprintf(MSGBK(6),nseries);
if(bankno!=0) 
	cprintf(MSGBK(65),'a'+bankno);
return 0;
}
//=======================================================================
short VamFile::ListNames(FILE* out, char* mask, short wide) {
short i,j,c,matchcount,toscreen,ObsPerLine,n;
char *cp,s[40];

if(wide)
	ObsPerLine=5;
else ObsPerLine=1;
if(out==stdout)
	toscreen=TRUE;
i = matchcount = 0;
if(nseries > 100 && toscreen) {
	#if !defined(OS2COMPILE)
	clrscr();
	gotoxy(1,1);
	#endif
	}
for(i=0;i<fd.nvecs;i++) {
	if(columns[i] != 1) continue;
	n=rows[i];
	for(j=0;j<n;j++) {
		sprintf(s,"%s%d",&vnames[vindx[i]],j+1);
		if(matchwild(s,mask)) {
			if(matchcount && matchcount % ObsPerLine == 0) {
				if(toscreen) cprintf("\r\n");
				else fprintf(out,"\n");
				}
			if( (toscreen) && matchcount && matchcount % (ObsPerLine*24) == 0){
				cprintf("\r");
				cprintf(MSGLSC(11));
				c = getch();
				if(c == ESC) break;
				#if !defined(OS2COMPILE)
				clrscr();
				gotoxy(1,1);
				#endif
			}
			if(toscreen) cprintf("%-15s",s);
			else fprintf(out,"%-15s",s);
			matchcount++;
			}
		}
	}
ObsPerLine=ObsPerLine;
return(OK);
}
