// VAMFILE.H -- Class declarations for vam file objects.

#if !defined(VAMFILE_H)
#define VAMFILE_H 1
#include "databank.h"
#include "gbank.h"
#include "matrix.h"

#define MAXMAT  350
class VamFileDesc {
 public:
	float startdate;
	float enddate;
	long nnpy;     // number of numbers per year
	long startpsn; // position in output file of first element of first vector
				 //    in first year.
	short nvecs;     // number of vectors
	unsigned short ncivnames; //number of characters in vector names
	short pmcount;  // packed matrix count
	char runname[80];
	};
#if !defined (MATID)
# define MATID 1
// matid is used in Interdyme.  See InitMat(), also see Interdyme-based
//   Vector and Matrix constructors.
class VamFile;  // forward reference
class matid { 
 public:
	char name[20];
	VamFile *vfp;   // Vamfile pointer associated with this matid.
	short rows, cols, lags;
	char type; /*	'm' for matrix, 'p' for packed matrix, 'v' for vector 
					with no lags, 'l' for vector with lagged values. */
	char writeflag; /* 'n' is this object is NOT to be written. 
		(There is no need to write objects that are not changed by the run.)*/

	char readflag; /* If readflag == 'n', this object is never to be read. 
		(It is in VAM but not yet in dyme.) If readflag == 'i', this object
		is to be read in the initial year of the run but not thereafter.  Use
		for vectors where last year's value is a good starting value for this
		year.  If readflag == 'y', the Vam object is to be read every year. 
		If readflag == 'b', the vector or matrix is read up through the year
		specified by lastdata, the next element in the structure. If readflag
		is 'a', then read when there is available data.  (This is determined
		by LastDat below.  */
	char loaded;
	short LastDat; // the last year in which actual data is available.
	long offset; // Position (in bytes) of first byte of the matrix in each year.
	long NumElements; // Number of elements in this object in vam file
	float *vptr;
	float **mptr;
	Vector *vecptr;
	Matrix *matptr;
	Pmatrix *pmatptr;
	char *pmfname;
// Functions here:
	matid();
	~matid(); // free up vf,vptr,mptr,vecptr,matptr,pmatprt,pmfname here!?
	};
#endif

class VamFile : public DataBank {

protected:
	VamFileDesc fd;
	unsigned short *vindx;
	long *sizes, *csizes;
	unsigned short nvecs,*columns,*rows;
	short *lags;
	char *vnames;         /* up to NCMAX ? */
	char **rowstb;        /* up to NVECMAX by 30? */
	char **colstb;
	char *types;
public:
	enum {MAXPMN = 30}; // max number of packed matrices.
	short lvn, loadline,lveclines;
	char action,loaded,loadview,lvchanged,loadvec[20];
	float **nmvec; // this is declared to start at [1][1]
	pmfid *pmh;   // MAXPMN pmh's will be dymamically defined
	short PmLoaded[MAXPMN];
	long pmrnzs;
	char **pmfilenames; // for use with packed matrices
	short np;
	// Pointer to G databank, whatever type.
	DataBank *VamBank;
	short GOpened;

	short read_filedesc(void);  // reads in some of the above
	short write_filedesc(void);

public:
//	enum {ESC=27, TRUE=1, FALSE=0};
	// constructor
	VamFile(const char *); // Open file, initialize members.
	VamFile(void);	 // "default" constructor
	short OpenVam(const char *, int OpenGBank=1); // Open Vam that was declared with default constr.
	short TryVamBank(const char *);  // Try to open associated G bank.
	short CheckHeader(int Howmany=10);

	// destructor
	~VamFile();

	// These functions were added to provide an interface to the file
	//  	description information in the VamFile (for example, used by
	//		InitMat() and other functions.
	char* Setbnktitle(const char *title);
	short Getnvecs(void) { return fd.nvecs;}
	float Getstartdate(void) {return fd.startdate;}
	float Getenddate(void)  {return fd.enddate;}
	long Getstartpsn(void) {return fd.startpsn;}
	long Getnnpy(void)  { return fd.nnpy;}
	long Getsizes(short lvn) {return( (lvn>nvecs-1)?-1:sizes[lvn]);}
	long Getoffset(short lvn) {return ( (lvn>nvecs-1)? -1 : csizes[lvn]);}
	char *Getname(short lvn) {return( (lvn>nvecs-1)? (char*)NULL : &vnames[vindx[lvn]]);}
	short Getnrows(short lvn)  {return( (lvn>nvecs-1)? -1 : (short)rows[lvn]); }
	short Getncols(short lvn)  {return( (lvn>nvecs-1)? -1 : (short)columns[lvn]); }
	short Getnlags(short lvn)  {return( (lvn>nvecs-1)? -1 : lags[lvn]);}
	short Getlveclines(void) {return lveclines;}
	char Gettype(short lvn)  {return( (lvn>nvecs-1)? 0 : types[lvn]);}
	FILE* Getfp(void) {return( (IsOpen())?pbank:NULL);}
	char *Getrunname(void) {return fd.runname;}
	char* Setrunname(char* s) {
		strncpy(fd.runname,s,79);
		write_filedesc();
		return fd.runname;}


	// Put and Get
	short addser(char *nam, float *f);
	short getser(char *nam, float *ss,short SeriesLength=-1);
	short GetVecElement(char *name, short yr, float& f);
   short GetVecElement(char *vname, short sec, short yr, float& f);
	short GetMatElement(char *name, short row, short col, short yr, float &f);
	short ListNames(FILE*,char*,short wide=1);
	short Announce(short bankno);
	// Write or read a row of packed matrix:
	short writearow(short PmatNo, short yr, short row, float *p);
	short readarow(short PmatNo, short yr, short row, float *p);
	short writearow(char* pmfname, short yr, short row, float *p);
	short readarow(char* pmfname, short yr, short row, float *p);

	// Other functions.
	short cumsize(void);
	short findname(char *name);
	short GetPmatNo(char *pmname);

	short loadcmd(char *s,char view, short line);
	short loadcmd(short vno,char view, short line) //qyu  7/8/93
		{return(loadcmd(&vnames[vindx[vno]], view, line));}
	short unloadcmd(void);
	short storecmd(char *vname);
	short pmloadcmd(float** temp,short pmcnt,char view,short line,short fr,short fc);

	// Special for Haver Analytics FAME format: (IBM project)
	short dumpall(FILE *dmp,short startyr,short lastyr,short decs);
	short addconcept(char *concept, GBank *gb,
	  short startdate, short enddate);

	// Planned link between VamFile and vorm classes...
	//friend short Vector::Load(short yr, VamFile& vf);
	//friend short Matrix::Load(short yr, VamFile& vf);
	//friend short Pmatrix::Load(short yr, VamFile& vf);

};

#endif // end VAMFILE_H class definitions


