extern char *Version;    // Interdyme Version
char *SlimDymeDate="Optimizing April 2004";
char *ModelTitle="TIDY - Thailand Interindustry Dynamic Model [26x26]";
/*
  MODEL.CPP -- This file consists of the loop function, plus any
  equation functions that comprise the part of the model written
  by the Interdyme user.

  This file, and the CONFIG.CPP file, should be the only two .CPP
  files that the model builder needs to work on.
*/
#include "dymesys.h"  // All includes for general Interdyme "infrastructure".
#if defined(INCL_MACRO)
#include "tseries.inc" // Re-written each time by IDBUILD.
#include "heart.h"
#endif
#include "user.h"     // All global variables for user model.

// Proto type for Seidel used in TINY
short Seidel(Matrix& A, Vector& q, Vector& f, IVector& triang,
	float toler);

/* Prototypes for other versions of Seidel
short Seidel(Matrix& A, Vector& q, Vector& imp, Vector& sdc,
        Equation& imports, Vector& f, short *triang,    float toler);
short PSeidel(Matrix& A, Vector& p, Vector& v, short *triang,
        float toler);
// With import share matrix:
short Seidel(Matrix& A, Matrix& MS, Vector& q, Vector& f, short *triang,
        float toler);
short PSeidel(Matrix& A, Matrix& MS, Vector& p, Vector& v, short *triang,
        float toler);

short Seidel(Matrix& A, Vector& q, Vector& imp, Vector& sdc,
        Equation& imports, Vector& f, IVector& triang, float toler);
short PSeidel(Matrix& A, Vector& p,  Vector& v, IVector& triang, float toler);

*/


/* ====================================================================*/
/* loop()is the start-up routine of the forecasting model.  It resizes the
   matrices and vectors of the VAM file which which are declared globally
   in user.h.  Similarly, resizes the equation objects that have been
   estimated in G and globally declared in user.h. Finally, it either
   calls spin() to cycle through the years, calling the various functions
   that do the forecasting, or it calls maxsolve to optimize through
   many calls to spin().
*/

const int NSEC=8; // Number of sectors in I-O table, most vectors.

void spin();

void loop(void)
{

    /* Resizing of vectors and matrices from the VAM file is done here.
       These vectors and matrices are declared globally in user.h, so they
       can be accessed by any function anywhere in the program.
       This declaration, however, cannot allocate space for them.  Essentially,
       they all have 0 rows and 0 columns.  Giving them real sizes and
       allocating space for them is done here by calls to each object's
       "resize" function, denoted by just ".r".

     The format is:
    <Dymename>.r("<vamname>",<readflag>,<writeflag>)

        where <Dymename> is what the object will be called in this program.
                <vamname>  is what it is called in VAM.
                <readflag> is 'i' if item is to be read only in the initial year.
                            'n' if it is NEVER to be read.
                            'y' otherwise.
                <writeflag> is 'n' if the item is not to be written.
    Normally, <Dymename> is the same as <vamname>.
    Default values of readflag = 'y' and writeflag = 'y' are supplied
    if they are omitted in the declaration.  Vectors with lagged values
    should be declared both as a Vector and as a Matrix, with the
    Vector definition coming first.

    Interdyme provides for special treatment of large, sparse matrices.
    The non-zero elements of the matrices may be store in a special form
    called "packed matrices."  The packed matrices do not go into the VAM
    file; rather, each has its own file.  Here is an example.
       Example of Packed Matrix (Pmatrix):
       Pmatrix B("bmv",'a','y', BmvFileName);

    Example of vectors with lagged values.  The resizing as a Matrix
    cannot precede the resizing as a Vector.  Do not use the same name
    or one differing only by capitalization for both the Vector and Matrix.

     Outlag.r("q",'i','n'), Priceslag.r("pq",'i','n');

    Second vam file:
    If you were reading some vectors out of a second Vam file,
    Vector declarations could be as follows:

    Vector fd(vf[1],"fd",'y','n'), va(vf[1],"va",'y','n');
        or:
    VamFile& vfScratch=vf[2];
    Vector fd(vfScratch,"fd",'y','n'), va(vfScratch,"va",'y','n');
    */

    // Vectors

    out.r("out");pce.r("pce");gov.r("gov");
    inv.r("inv");ex.r("ex");im.r("im");fd.r("fd");dep.r("dep");
    depc.r("depc"),lab.r("lab");labc.r("labc");cap.r("cap");
    capc.r("capc"),ind.r("ind");indc.r("indc");pcec.r("pcec");
    invc.r("invc"),govc.r("govc");exc.r("exc");imc.r("imc");
    x.r("x");
    y.r("y");

    // Matrices
    AM.r("AM");

    #ifdef INCL_IVECTOR
    // The triang vector specifies the order of solving equations in Seidel.
    // Example of how to use IVector class for all triangulation.
    triang.resize(NSEC);
    triang.ReadA("triang.iv");   // Read the triangularization vector
    #endif

    // Resize and read the vecor for spreading the PCE discrepancy.
    PCESpread.resize(NSEC);
    PCESpread.ReadA("PCESpread.dat");

    if(MaxFlag == 'n')
    	spin();
    else{
    	maxsolve();
    	MaxFlag = 'n';
    	spin();
    	}
   #if defined(INCL_MACRO)
       storets(); // Store the macro variable time series
   #endif

   printf("\nThe run has finished.  Use G or Compare to view results.\n");
   tap();
   }

// The spin() function sends the model through one cycle of years.

void spin(){
   // General start of the spin() function:
    #if defined(INCL_MACRO)
    tserin(); //load macrovariables
    #endif
    // Set tolerance for Seidel process
    toler = 1.;
    float oldpcetot,oldinvtot,pcedif,invdif;
    float pcesum, pcediscrep;
	int Iteration;
    // Iteration loop period-by-period:
    // ===============================
    for (t = godate; t<= stopdate; t++) {
 
       if (MaxFlag == 'n') printf("%d ",t);
       // Load all vectors and matrices.
       load(t);
	uptseries();

  	// Start of code particular to TINY:
	Iteration = 0;
	// The loop for convergence on pcetot and invtot.
    while(Iteration < 20){
    	Iteration++;
	    oldinvtot = invtot[t]; oldpcetot = pcetot[t];	
	    if(t>= MacEqStartDate)
	    // pce = pcetot[t]*pcec; Removed
	    // Compute the estimated PCE equations
	    pcef(pce);
	    // Sum up the calculated PCE elements
 	    pcesum = pce.sum();
	    pcediscrep = pcetot[t] - pcesum;
	    // printf("\npcediscrep = %10.2f\n",pcediscrep);
	    // Spread the discrepancy by the proportions of the PCESpread vector.
	    pce = pce + pcediscrep*PCESpread;
           
	    invtotf();
	    inv = invtot[t]*invc;
	    gov = govtot[t]*govc;
	    im  = imtot[t]*imc;
	    ex  = extot[t]*exc;
	    fd = pce + gov + inv + ex + im;
	    Seidel(AM, out, fd, triang, toler);
	    dep = ebemul(depc,out);  lab = ebemul(labc,out);
    	    cap = ebemul(capc,out);  ind = ebemul(indc,out);
	    // The Accountant for Tiny
	    gdp[t] = fd.sum();
	    if(t>MacEqStartDate){
	        deprec[t] = dep.sum(); labinc[t] = lab.sum();
	        capinc[t] = cap.sum(); indtax[t] = ind.sum();
        	accountf();
	        }
		invdif = fabs(invtot[t] - oldinvtot);
		pcedif = fabs(pcetot[t]- oldpcetot);
		printf("Iter %2d pce = %7.1f pcedif = %6.2f invdif = %6.2f\n",Iteration,
		     pcetot[t],pcedif,invdif);
		if(invdif < .5 && pcedif < .5) break;
		}
    // End of code particular to TINY
	// Here when both Investment and PCE have converged
	// Standard end of the spin() function:
	if(MaxFlag == 'y')
            shiftback(t);
       else{
           // Store the values of vectors and matrices for this period.
           store(t);
           printf("\n");
           }
	}
    }

